:- module(basics,	[
	        bagof0/3, 
		bip/1,
                peel_meta/2,
		mgu/3,
		setof0/3,
		sub_apply/3,
		sub_constructor/3,
		sub_domain/2,
	        sub_inverse/2,
		sub_mapsto/3,
		sub_restrict/3,
		subs_restrict/3,
		tagterm/3,
	        vars/2, 
		var_occ/2,
		intern/3,
		variable/1,
		variable_id/2,
		variable_ids/2,
		variable_max_no/1,
		variable_id_tagging/2,
                nameListToSub/3
	]).


:- use_module(ordsets).



% bip predicate

bip(abolish(_)).
bip(assert(_)).
bip(retract(_)).
bip(erase(_)).
bip(recorda(_,_,_)).
bip(recorded(_,_,_)).
bip(format(_,_)).
bip(format(_,_,_)).


bip(abort).
bip(put(_Chr)).
bip(length(_List,_Num)).
bip(name(_Atom,_ASCIIs)).
bip(sort(_Atom,_ASCIIs)).
bip(keysort(_Atom,_ASCIIs)).
bip(statistics(_Key,_Stats)).



bip(repeat).  %done for poly/mono type/groundness
bip(listing).
bip(listing(_)).
bip(float(_)).
bip(string(_)).
bip(compound(_)).
bip(ground(_)).
bip(portray_clause(_)).
bip(\=(_,_)).
bip(=@=(_,_)).
bip(compare(_,_,_)).
bip(!).                     
bip(functor(_,_,_)).
bip(arg(_,_,_)).
bip(=..(_,_)).

                            % done for all analyses
bip(display(_TERM)).      % built-in predicates for input/output
bip(nl).                  % they do not change program state
bip(print(_TERM)).
bip(read(_TERM)).         % except read that changes program state
bip(tab(_Natural)).
bip(write(_TERM)).
bip(writeq(_TERM)).
bip(atom(_TERM)).         % built-in predicates for test
bip(atomic(_TERM)).       % they do not change program state
bip(integer(_TERM)).
bip(number(_TERM)).
bip(nonvar(_TERM)).
bip(var(_TERM)).
bip(=:=(_NUMBER1,_NUMBER2)).
bip(=\=(_NUMBER1,_NUMBER2)).
bip(<(_NUMBER1,_NUMBER2)).
bip(>(_NUMBER1,_NUMBER2)).
bip(=<(_NUMBER1,_NUMBER2)).
bip(>=(_NUMBER1,_NUMBER2)).
bip(@<(_TERM1,_TERM2)).
bip(@>(_TERM1,_TERM2)).
bip(@=<(_TERM1,_TERM2)).
bip(@>=(_TERM1,_TERM2)).
bip(=(_Left,_Right)).
bip(==(_Left,_Right)).
bip(\==(_Left,_Right)).
bip(true).                  % flow-control
bip(false).
bip(fail).
bip(is(_TERM,_NUMBER)).     % may change program state


setof0(X,Goal,SX) :- setof(X,Goal,SX1) -> SX=SX1; SX=[].
bagof0(X,Goal,SX) :- bagof(X,Goal,SX1) -> SX=SX1; SX=[].

mgu(X,Y,Sub) :-
	variable(X),
	!,
	( 
		X==Y
	->
		Sub=[]
	;
		\+ occur(X,Y)
	->
		Sub = [X/Y]
	).
mgu(Y,X,Sub) :-
	variable(X),
	!,
	( 
		X==Y
	->
		Sub=[]
	;
		\+ occur(X,Y)
	->
		Sub = [X/Y]
	).

mgu(X,Y,Sub) :-
	!,
	X=..[F|Args1],
	Y=..[F|Args2],
	mgu_list(Args1,Args2,Sub).

mgu_list([],[],[]). 
mgu_list([H1|T1],[H2|T2],Sub) :-
	mgu_list(T1,T2,SubT),
	sub_apply(SubT,H1,H11),
	sub_apply(SubT,H2,H21),
	mgu(H11,H21,SubH),
	sub_compose(SubT,SubH,Sub). 

% occur(+V,+T) +V occurs in term +T.

occur(X,Y) :-	% Y is a variable 
	variable(Y),
	!,
	X==Y.
occur(X,Y) :-	% Y is not a variable
	Y=..[_|Args],
	occur_in_one_term(X,Args).

occur_in_one_term(X,[H|T])  :- 
	( occur(X,H) -> true ; occur_in_one_term(X,T) ).



/* sub_compose(+Theta,+Sigma,-Tau) Tau = Theta o Sigma */

sub_compose(Theta,Sigma,Tau) :-
	instantiation_pass(Theta,Sigma,Tau_),
	merging_pass(Sigma,Tau_,Tau).

instantiation_pass([],_,[]) :-
	!.
instantiation_pass([V/Ti|Theta],Sigma,[V/To|Tau]) :-
	sub_apply(Sigma,Ti,To),
	instantiation_pass(Theta,Sigma,Tau).

merging_pass(Sigma,Theta,Tau) :-
	sub_domain(Theta,VL),
	merging_pass_1(Sigma,VL,Theta,Tau).

merging_pass_1([],_,Theta,Theta).
merging_pass_1([V/T|Sigma],VL,Theta,Tau) :-
	( 
		ord_subset([V],VL)
	->
		merging_pass_1(Sigma,VL,Theta,Tau)
	;
		merging_pass_1(Sigma,VL,Theta,Tau1),
		Tau = [V/T|Tau1]
	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

/* 

vars(+TERM,-VARSET) VARSET is the set of variables in TERM
and is ordered.  

*/

vars(T,Bag)	:-
    vars(T,[],Bag).
vars(X,Bag,NewBag) :-
    variable(X),
    !,
    ord_add_element(Bag,X,NewBag).
vars(X,Bag,NewBag) :-	
%    \+ variable(X),
    !,
    X=..[_|Args],
    vars_all(Args,Bag,NewBag).

vars_all([],Bag,Bag).
vars_all([Arg1|Args2],Bag,NewBag) :-	
    vars_all(Args2,Bag,NewBag1),
    vars(Arg1,NewBag1,NewBag).


/*var_occ(+TERM,-VAROCC) VAROCC  is  an  ordered list  of pairs  of  a
variable  and its number of occurrences in TERM. Each variable in TERM
corresponds to a pair in VAROCC*/

var_occ(T,Bag) :-
    var_occ(T,[],Bag).

var_occ(X,Bag,NewBag) :-
    variable(X),
    !,
    insert(Bag,X,NewBag).
var_occ(X,Bag,NewBag) :-	
%    \+ variable(X),
    !,
    X=..[_|Args],
    var_occ_all(Args,Bag,NewBag).

var_occ_all([],Bag,Bag).
var_occ_all([Arg1|Args2],Bag,NewBag) :-	
    var_occ_all(Args2,Bag,NewBag1),
    var_occ(Arg1,NewBag1,NewBag).

insert([],Var,[(Var,1)]).
insert([(Var1,Occ1)|Bag],Var,[(Var1,NewOcc1)|Bag]) :-
    	Var == Var1,
	!,
	NewOcc1 is Occ1+1.
insert([(Var1,Occ1)|Bag],Var,[(Var,1),(Var1,Occ1)|Bag]) :-
    	Var @< Var1,
	!.
insert([(Var1,Occ1)|Bag],Var,[(Var1,Occ1)|NewBag]) :-
    	%Var >@ Var1,
	!,
	insert(Bag,Var,NewBag).


/* tagterm(+In,+Shift,-Out) the term Out is the result of tagging each
    variable in the term In
*/


tagvar('$VAR'(N1),Shift,'$VAR'(N2))	:-
	N2 is N1 + Shift.                  

tagterm(In,Shift,Out)	:-
	variable(In),
	!,
	tagvar(In,Shift,Out).
tagterm(In,Shift,Out)	:-
%	 \+ variable(In),
	In =.. [Func|InArgs],
	tagtermall(InArgs,Shift,OutArgs),
	Out =.. [Func|OutArgs].

tagtermall([],_,[]). 
tagtermall([In|Ins],Shift,[Out|Outs])	:-
	tagterm(In,Shift,Out),
	tagtermall(Ins,Shift,Outs).


/*

sub_mapsto(+Var,+Sub,-Val) If Var is in the domain of Sub
then sub_mapsto(Var,Sub,Val) succeeds and Val is the value
assigned to Var by Sub. Otherwise sub_mapsto(Var,Sub,Val)
fails.

*/

sub_mapsto(Var,[Var1/Val1|Sub],Val) :-
	( Var==Var1 -> Val=Val1 ; sub_mapsto(Var,Sub,Val) ).

/*

sub_apply(+Sub,+TermIn,-TermO) TermO=Sub(TermIn). It
unravals the input term TermIn in the top-down manner and
constructs the output term TermO in the bottom-up manner.

*/

sub_apply(Sub,TermIn,TermO) :-
	(    
		variable(TermIn)
	->
		( sub_mapsto(TermIn,Sub,TermO) -> true; TermO=TermIn )
	; 
		TermIn =.. [F|ArgsIn],
		sub_apply_all(ArgsIn,Sub,ArgsO),
		TermO =.. [F|ArgsO]         
	).

sub_apply_all([],_,[]).
sub_apply_all([ArgIn|ArgsIn],Sub,[ArgO|ArgsO]) :-
	sub_apply(Sub,ArgIn,ArgO),
	sub_apply_all(ArgsIn,Sub,ArgsO).


/*

sub_domain(+Sub,-Vars) Vars is the domain of Sub in the form
of ordered list

*/

sub_domain([],[]).
sub_domain([V/_|Sub1],D) :-
	sub_domain(Sub1,D1),
	ord_add_element(D1,V,D).


/*

sub_restrict(+SubIn, +Vars,-SubO) SubO = SubIn | Vars 

*/

sub_restrict(SubIn,Vars,SubO) :-
	sort(Vars,Vars1),
	sub_restrict_1(SubIn,Vars1,SubO).

sub_restrict_1([],_,[]).
sub_restrict_1([Var/Val|SubIn],Vars,[Var/Val|SubO]) :-
	ord_subset([Var],Vars),
	!,
	sub_restrict_1(SubIn,Vars,SubO).
sub_restrict_1([_Var/_Val|SubIn],Vars,SubO) :-
%	\+ ord_subset([_Var],Vars),
	sub_restrict_1(SubIn,Vars,SubO).

/*

subs_restrict(+SubsIn,+Vars,-SubsO)

*/

subs_restrict(SubsIn,Vars,SubsO) :-
	sort(Vars,Vars1),
	subs_restrict_1(SubsIn,Vars1,SubsO).

subs_restrict_1([],_,[]).
subs_restrict_1([Si|SiL],Vars,[So|SoL]) :-
	sub_restrict_1(Si,Vars,So),
	subs_restrict_1(SiL,Vars,SoL).


/*

sub_inverse(+SubIn,-SubO) SubO is the inverse of SubIn

*/

sub_inverse([],[]).
sub_inverse([Var/Val|SubIn],SubO) :-
	sub_inverse(SubIn,SubM),
	ord_add_element(SubM,Val/Var,SubO).   
    
/*

sub_constructor(+Vars,+Vals,-Sub)

*/

sub_constructor(Vars,Vals,Sub) :-
    	sub_constructor_1(Vars,Vals,SubM),
	sort(SubM,Sub).

sub_constructor_1([],[],[]).
sub_constructor_1([Var|Vars],[Val|Vals],[Var/Val|Sub]) :-
    	sub_constructor_1(Vars,Vals,Sub).



/*

In meta logic programming, there are two approaches for program
representation: ground representation and non-ground
representation. Variables in object programs are represented by
ground terms in meta programs in the ground representation
approach while they are represented by variables in meta programs
in the non-ground representation approach.

We have choosen the ground representation approach. The
representation scheme is in the following.

A variable is represented by a ground term of the form of
'$VAR'(NATURAL_NUMBER) where NATURAL_NUMBER is a nutural number
and we assume that object programs do not use '$VAR' as a function
symbol or a predicate symbol. Different variables in a clause are
identified with different natural numbers. Since a variable in a
clause is universally quantified, it is appropriate to identify
different variables with the same natural number so long as they
appear in different clauses and to identify the same variable in
different clauses with different variables. We identify variables
in a clause with natural numbers which are powers of 2, that is,
0x0001, 0x0002, 0x0004 and so on. The reason for the choice of the
natural numbers will be become clear later.

Terms, atoms, conjunctions, disjunctions, implications (facts are
implications whose bodies are the atom _true_) are represented as
usual.

A clause is represented in the form of clause_rep(IMPLICATION_REP,
SYMBOL_TABLE). where IMPLICATION_REP is the representation of the
implication of the clause while SYMBOL_TABLE is a bijection
between variable identifiers variables and their external
names. It is a list of pairs of the form
(NATURAL_NUMBER,EXTERNAL_NAME) where EXTERNAL_NAME is a string.

Clauses are only renamed once. Therefore, renaming can be replaced by
tagging. The analysis system assume that there are at most 999 variables in
any clause of an object program. We believe from our own experience that
this is a practical assumption since any logical statement with more than a
dozen universally quantified variables will be very difficult to
understand. This assumption has only to do with tagging and replacing 999
with a high number will only require a new tagging procedure.

*/

:- use_module(library(lists)). 

variable_max_no(64).


variable('$VAR'(_)).

variable_id('$VAR'(Id),Id).

variable_id_tagging(Id,Id1) :-
    variable_max_no(N),
    Id1 is Id+N.

variable_ids([],[]).
variable_ids([V|Vs],[Id|Ids]) :-
    variable_id(V,Id),
    variable_ids(Vs,Ids).




/* 

intern(+In,-Out, -Map) -Out and +In represents the same term or the
same atom object. Out is the meta-level representation while +In is
the object-level representation. -Map is the bijection between object
variables with thier identifiers.

*/

look_for_intern(EXT, [INT1/EXT1|Map], INT)	:-
	(
		EXT == EXT1, INT = INT1
	;
		look_for_intern(EXT, Map, INT)
	).




intern(Obj,Meta,Map)	:-
	intern( Obj, Meta, 1, _, [], Map ).

intern(Obj, Meta, NatIn, NatO, MapIn, MapO)	:-
	var(Obj),
	!,
	(
		look_for_intern(Obj, MapIn, Meta)
	->
		NatO = NatIn,
		MapO = MapIn
	;
		Meta = '$VAR'(NatIn),
		NatO is NatIn + 1,
		MapO = ['$VAR'(NatIn)/Obj|MapIn]
	).

intern(Obj, Meta, NatIn, NatO, MapIn, MapO)	:-
%	\+ var(Obj),
	!,
	Obj =.. [Func|ObjArgs],
	intern_all(ObjArgs, MetaArgs, NatIn, NatO, MapIn, MapO),
	Meta =.. [Func|MetaArgs].

intern_all([], [], NatIn, NatIn, Map, Map).
intern_all([Obj|Objs], [Meta|Metas], NatIn, NatO, MapIn, MapO)	:-
	intern(Obj, Meta, NatIn, NatM, MapIn, MapM),
	intern_all(Objs, Metas, NatM, NatO, MapM, MapO).


/*nameListToSub(+VL,+NL,-SUB) -SUB  is  a substitution over  domain VL
with variables assigend  their names as defined  in NL or an anonymous
names if undefined in NL.*/

nameListToSub(VL,NL,Sub) :-
    nameListToSub_1(NL,Sub1),
    sub_domain(Sub1,VL1),
    ord_subtract(VL,VL1,VL2),
    anonymous_sub(VL2,Sub2),
    ord_union(Sub1,Sub2,Sub). 

nameListToSub_1([],[]).
nameListToSub_1([Name=Var|NL1],Sub) :-
    nameListToSub_1(NL1,Sub1),
    ord_add_element(Sub1,Var/Name,Sub).

anonymous_sub([],[]).
anonymous_sub([Var|Vars],[Var/'_'|Sub]) :-
    anonymous_sub(Vars,Sub).


peel_meta(once(A),B) :-
    !,
    peel_meta(A,B).
peel_meta(call(A),B) :-
    !,
    peel_meta(A,B).
peel_meta(A,B) :-
    A =..[call,G|Args],
    \+ variable(G),
    !,
    G =..[Func|Args0],
    append(Args0,Args,ArgsNew),
    C =.. [Func|ArgsNew],
    peel_meta(C,B).
peel_meta(A,B) :-
    A =..[apply,G|Args],
    \+ variable(G),
    !,
    G =..[Func|Args0],
    append(Args0,Args,ArgsNew),
    C =.. [Func|ArgsNew],
    peel_meta(C,B).

peel_meta(A,A). % the catch-all case.


%%%%%%%%%%%%%%%%%%%%%%%%%%



