:- module(domain, [ 
		asub_eq_constraints/3,
                asub_init/1,
		asub_list_print/2,
		asub_unify_entry/6,
		asub_unify_exit/7,
		asub_bipai/4,
		asub_compile/3,
		asub_bot/3,
		asub_ident/3,
		asub_join/4,
		asub_leq/3,
		asub_internalize/3,
		asub_externalize/3,
		asub_bw_compile/3,
		asub_bip_has_cond/3,
		asub_top/3,
		asub_meet/4,
		asub_bw_unify_entry/6,
		asub_bw_unify_exit/7,
		asub_bw_bipai/4
	       ]).


:- discontiguous
                asub_eq_constraints/3,
                asub_init/1,
		asub_list_print/2,
		asub_unify_entry/6,
		asub_unify_exit/7,
		asub_bipai/4,
		asub_compile/3,
		asub_bot/3,
		asub_ident/3,
		asub_join/4,
		asub_leq/3,
		asub_internalize/3,
		asub_externalize/3,
		asub_bw_compile/3,
		asub_bip_has_cond/3,
		asub_top/3,
		asub_meet/4,
		asub_bw_unify_entry/6,
		asub_bw_unify_exit/7,
		asub_bw_bipai/4.


% rtype

%:- use_module(rtype_or).

:- use_module(rtype).
%:- use_module(rtype_ntb).


asub_list_print(rtype,ASubList):-
    rtype_list_print(ASubList).

asub_init(rtype) :- 
    rtype_init.

asub_eq_constraints(rtype,Sub,F) :- 
    rtype_eq_constraints(Sub,F).

asub_unify_entry(rtype,Sub,Sigma,NVarR,NVar,Zeta) :- 
    rtype_unify_entry(Sub,Sigma,NVarR,NVar,Zeta).

asub_unify_exit(rtype,Sub,Sigma,_NVarR,Theta,NVar,Zeta) :-
    rtype_unify_exit(Sub,Sigma,_NVarR,Theta,NVar,Zeta).

asub_bipai(rtype,A,Theta,Zeta) :- 
    rtype_bipai(A,Theta,Zeta).

asub_compile(rtype,Call,Compiled) :- 
    rtype_compile(Call,Compiled).

asub_bot(rtype,NVar,Bot) :- 
    rtype_bot(NVar,Bot).

asub_ident(rtype,NVar,Id) :- 
    rtype_ident(NVar,Id).

asub_join(rtype,Zeta,Eta,Beta) :- 
    rtype_join(Zeta,Eta,Beta).

asub_leq(rtype,F1,F2) :- 
    rtype_leq(F1,F2). 

asub_internalize(rtype,Extern,Intern) :- 
    rtype_internalize(Extern,Intern).

asub_externalize(rtype,Intern,Extern) :- 
    rtype_externalize(Intern,Extern).

asub_top(rtype,NVar,TOP) :- 
    rtype_top(NVar,TOP).

asub_meet(rtype,Zeta,Eta,Beta) :- 
    rtype_meet(Zeta,Eta,Beta).

asub_bip_has_cond(rtype,AIJ,Cond) :- 
    rtype_bip_has_cond(AIJ,Cond).

asub_bw_unify_exit(rtype,CE,Answer,NVarR,DemandRt,NVar,DemandLt) :-
    rtype_bw_unify_exit(CE,Answer,NVarR,DemandRt,NVar,DemandLt).

asub_bw_unify_entry(rtype,CE,Sigma,NVarR,NVar,Eta) :-
    rtype_bw_unify_entry(CE,Sigma,NVarR,NVar,Eta).

asub_bw_compile(rtype,Call,Compiled) :- 
    rtype_bw_compile(Call,Compiled).

asub_bw_bipai(rtype,A,Theta,Zeta) :-
    rtype_bw_bipai(A,Theta,Zeta).




% POS-based groundness analysis 

:- use_module(pos).

asub_list_print(pos,ASubList) :- 
    pos_list_print(ASubList).

asub_init(pos) :- 
    pos_init.

asub_eq_constraints(pos,Sub,F) :- 
    pos_eq_constraints(Sub,F).

asub_unify_entry(pos,Sub,Sigma,NVarR,NVar,Zeta) :- 
    pos_unify_entry(Sub,Sigma,NVarR,NVar,Zeta).

asub_unify_exit(pos,Sub,Sigma,NVarR,Theta,NVar,Zeta) :-
    pos_unify_exit(Sub,Sigma,NVarR,Theta,NVar,Zeta).

asub_bipai(pos,A,Theta,Zeta) :- 
    pos_bipai(A,Theta,Zeta).

asub_compile(pos,Call,Compiled) :- 
    pos_compile(Call,Compiled).

asub_bot(pos,NVar,ASub) :- 
    pos_bot(NVar,ASub).

asub_ident(pos,NVar,ASub) :- 
    pos_ident(NVar,ASub).

asub_top(pos,NVar,ASub) :- 
    pos_top(NVar,ASub).

asub_join(pos,Zeta,Eta,Beta) :- 
    pos_join(Zeta,Eta,Beta).

asub_leq(pos,F1,F2) :- 
    pos_leq(F1,F2).

asub_internalize(pos,In,Out):- 
    pos_internalize(In,Out).

asub_externalize(pos,In,Out):- 
    pos_externalize(In,Out).

asub_meet(pos,Zeta,Eta,Beta) :- 
    pos_meet(Zeta,Eta,Beta).

asub_bw_compile(pos,Call,Compiled) :- 
    pos_bw_compile(Call,Compiled).

asub_bip_has_cond(pos,AIJ,Cond) :- 
    pos_bip_has_cond(AIJ,Cond).

asub_bw_unify_entry(pos,CE,Sigma,NVarR,NVar,Eta) :-
    pos_bw_unify_entry(CE,Sigma,NVarR,NVar,Eta).

asub_bw_unify_exit(pos,CE,Answer,NVarR,Success,NVar,F3) :-
    pos_bw_unify_exit(CE,Answer,NVarR,Success,NVar,F3).

asub_bw_bipai(pos,A,Theta,Zeta) :-
    pos_bw_bipai(A,Theta,Zeta).




% pair sharing analysis

:- use_module(pair_compile).

asub_list_print(pair,ASubList) :- 
    pair_list_print(ASubList).

asub_init(pair) :- 
    pair_init.

asub_eq_constraints(pair,Sub,F) :- 
    pair_eq_constraints(Sub,F).

asub_unify_entry(pair,Sub,Sigma,NVarR,NVar,Zeta) :- 
    pair_unify_entry(Sub,Sigma,NVarR,NVar,Zeta).

asub_unify_exit(pair,Sub,Sigma,NVarR,Theta,NVar,Zeta) :-
    pair_unify_exit(Sub,Sigma,NVarR,Theta,NVar,Zeta).

asub_bipai(pair,A,Theta,Zeta) :- 
    pair_bipai(A,Theta,Zeta).

asub_compile(pair,Call,Compiled) :- 
    pair_compile(Call,Compiled).

asub_bot(pair,NVar,ASub) :- 
    pair_bot(NVar,ASub).

asub_ident(pair,NVar,ASub) :- 
    pair_ident(NVar,ASub).

asub_top(pair,NVar,ASub) :- 
    pair_top(NVar,ASub).

asub_join(pair,Zeta,Eta,Beta) :- 
    pair_join(Zeta,Eta,Beta).

asub_leq(pair,F1,F2) :- 
    pair_leq(F1,F2).

asub_internalize(pair,In,Out):- 
    pair_internalize(In,Out).

asub_externalize(pair,In,Out):- 
    pair_externalize(In,Out).

asub_meet(pair,Zeta,Eta,Beta) :- 
    pair_meet(Zeta,Eta,Beta).

asub_bw_compile(pair,Call,Compiled) :- 
    pair_bw_compile(Call,Compiled).

asub_bip_has_cond(pair,AIJ,Cond) :- 
    pair_bip_has_cond(AIJ,Cond).

asub_bw_unify_entry(pair,CE,Sigma,NVarR,NVar,Eta) :-
    pair_bw_unify_entry(CE,Sigma,NVarR,NVar,Eta).

asub_bw_unify_exit(pair,CE,Answer,NVarR,Success,NVar,F3) :-
    pair_bw_unify_exit(CE,Answer,NVarR,Success,NVar,F3).

asub_bw_bipai(pair,A,Theta,Zeta) :-
    pair_bw_bipai(A,Theta,Zeta).



% monomorphis type analysis 

:- use_module(monotype).

asub_list_print(monotype,ASubList) :- 
    monotype_list_print(ASubList).

asub_init(monotype) :- 
    monotype_init.

asub_eq_constraints(monotype,Sub,F) :- 
    monotype_eq_constraints(Sub,F).

asub_unify_entry(monotype,Sub,Sigma,NVarR,NVar,Zeta) :- 
    monotype_unify_entry(Sub,Sigma,NVarR,NVar,Zeta).

asub_unify_exit(monotype,Sub,Sigma,NVarR,Theta,NVar,Zeta) :-
    monotype_unify_exit(Sub,Sigma,NVarR,Theta,NVar,Zeta).

asub_bipai(monotype,A,Theta,Zeta) :- 
    monotype_bipai(A,Theta,Zeta).

asub_compile(monotype,Call,Compiled) :- 
    monotype_compile(Call,Compiled).

asub_bot(monotype,NVar,ASub) :- 
    monotype_bot(NVar,ASub).

asub_ident(monotype,NVar,ASub) :- 
    monotype_ident(NVar,ASub).

asub_top(monotype,NVar,ASub) :- 
    monotype_top(NVar,ASub).

asub_join(monotype,Zeta,Eta,Beta) :- 
    monotype_join(Zeta,Eta,Beta).

asub_leq(monotype,F1,F2) :- 
    monotype_leq(F1,F2).

asub_internalize(monotype,In,Out):- 
    monotype_internalize(In,Out).

asub_externalize(monotype,In,Out):- 
    monotype_externalize(In,Out).

asub_meet(monotype,Zeta,Eta,Beta) :- 
    monotype_meet(Zeta,Eta,Beta).

asub_bw_compile(monotype,Call,Compiled) :- 
    monotype_bw_compile(Call,Compiled).

asub_bip_has_cond(monotype,AIJ,Cond) :- 
    monotype_bip_has_cond(AIJ,Cond).

asub_bw_unify_entry(monotype,CE,Sigma,NVarR,NVar,Eta) :-
    monotype_bw_unify_entry(CE,Sigma,NVarR,NVar,Eta).

asub_bw_unify_exit(monotype,CE,Answer,NVarR,Success,NVar,F3) :-
    monotype_bw_unify_exit(CE,Answer,NVarR,Success,NVar,F3).

asub_bw_bipai(monotype,A,Theta,Zeta) :-
    monotype_bw_bipai(A,Theta,Zeta).


/* Approximate reduced product (Codish et.al) is implemented. An
    abstract substitution for the composite analysis is represented as
    a pair consisting of abstract substitutions of respective
    component analyses and so is a compiled constraint for the
    composite analysis. 
*/

asub_list_print(rp(Alpha,Beta),ASubList) :- 
    proj(ASubList,ASubList1,ASubList2),
    asub_list_print(Alpha,ASubList1),
    asub_list_print(Beta, ASubList2).

    proj([],[],[]).
    proj([(X,Y)|L],[X|L1],[Y|L2]) :- proj(L,L1,L2).

asub_init(rp(Alpha,Beta)) :- 
    asub_init(Alpha),
    asub_init(Beta).

asub_eq_constraints(rp(Alpha,Beta),Sub,(EC1,EC2)) :- 
    asub_eq_constraints(Alpha,Sub,EC1),
    asub_eq_constraints(Beta,Sub,EC2).

asub_unify_entry(rp(Alpha,Beta),(EC1,EC2),(Sg1,Sg2),NVarR,NVar,(Zt1,Zt2)) :- 
    asub_unify_entry(Alpha,EC1,Sg1,NVarR,NVar,Zt10),
    asub_unify_entry(Beta,EC2,Sg2,NVarR,NVar,Zt20),
    asub_reduce(Alpha,Beta,Zt10,Zt20,Zt1,Zt2).

asub_unify_exit(rp(Alpha,Beta),(EC1,EC2),(Sg1,Sg2),NVarR,(Tht1,Tht2),NVar,(Zt1,Zt2)) :-
    asub_unify_exit(Alpha,EC1,Sg1,NVarR,Tht1,NVar,Zt10),
    asub_unify_exit(Beta, EC2,Sg2,NVarR,Tht2,NVar,Zt20),
    asub_reduce(Alpha,Beta,Zt10,Zt20,Zt1,Zt2).

asub_bipai(rp(Alpha,Beta),A,(Tht1,Tht2),(Zt1,Zt2)) :- 
    asub_bipai(Alpha,A,Tht1,Zt10),
    asub_bipai(Beta,A,Tht2,Zt20),
    asub_reduce(Alpha,Beta,Zt10,Zt20,Zt1,Zt2).

asub_compile(rp(Alpha,Beta),Call,(EC1,EC2)) :- 
    asub_compile(Alpha,Call,EC1),
    asub_compile(Beta, Call,EC2).

asub_bot(rp(Alpha,Beta),NVar,(ASub1,ASub2)) :- 
    asub_bot(Alpha,NVar,ASub1),
    asub_bot(Beta, NVar,ASub2).

asub_ident(rp(Alpha,Beta),NVar,(ASub1,ASub2)) :- 
    asub_ident(Alpha,NVar,ASub1),
    asub_ident(Beta, NVar,ASub2).

asub_top(rp(Alpha,Beta),NVar,(ASub1,ASub2)) :- 
    asub_top(Alpha,NVar,ASub1),
    asub_top(Beta, NVar,ASub2).

asub_join(rp(Alpha,Beta),(Sg1,Sg2),(Et1,Et2),(Zt1,Zt2)) :- 
    asub_join(Alpha,Sg1,Et1,Zt10),
    asub_join(Beta, Sg2,Et2,Zt20),
    asub_reduce(Alpha,Beta,Zt10,Zt20,Zt1,Zt2).

asub_leq(rp(Alpha,Beta),(Sg1,Sg2),(Et1,Et2)) :- 
    asub_leq(Alpha,Sg1,Et1),
    asub_leq(Beta, Sg2,Et2).

asub_internalize(rp(Alpha,Beta),(In1,In2),(Out1,Out2)):- 
    asub_internalize(Alpha,In1,Out1),
    asub_internalize(Beta, In2,Out2).

asub_externalize(rp(Alpha,Beta),(In1,In2),(Out1,Out2)):- 
    asub_externalize(Alpha,In1,Out1),
    asub_externalize(Beta, In2,Out2).

asub_meet(rp(Alpha,Beta),(Sg1,Sg2),(Et1,Et2),(Zt1,Zt2)) :- 
    asub_meet(Alpha,Sg1,Et1,Zt1),
    asub_meet(Beta, Sg2,Et2,Zt2). % reduction?

asub_bw_compile(rp(Alpha,Beta),Call,(Cpled1,Cpled2)) :- 
    asub_bw_compile(Alpha,Call,Cpled1),
    asub_bw_compile(Beta, Call,Cpled2).

asub_bip_has_cond(rp(Alpha,Beta),AIJ,(Cond1,Cond2)) :- 
    asub_bip_has_cond(Alpha,AIJ,Cond1),
    asub_bip_has_cond(Beta, AIJ,Cond2).

asub_bw_unify_entry(rp(Alpha,Beta),(CE1,CE2),(Et1,Et2),NVarR,NVar,(Zt1,Zt2)) :-
    asub_bw_unify_entry(Alpha,CE1,Et1,NVarR,NVar,Zt10),
    asub_bw_unify_entry(Beta, CE2,Et2,NVarR,NVar,Zt20),
    asub_reduce(Alpha,Beta,Zt10,Zt20,Zt1,Zt2).

asub_bw_unify_exit(rp(Alpha,Beta),(CE1,CE2),(Sg1,Sg2),NVarR,(Et1,Et2),NVar,(Zt1,Zt2)) :-
    asub_bw_unify_exit(Alpha,CE1,Sg1,NVarR,Et1,NVar,Zt10),
    asub_bw_unify_exit(Beta, CE2,Sg2,NVarR,Et2,NVar,Zt20),
    asub_reduce(Alpha,Beta,Zt10,Zt20,Zt1,Zt2).

asub_bw_bipai(rp(Alpha,Beta),A,(Tht1,Tht2),(Zt1,Zt2)) :-
    asub_bw_bipai(Alpha,A,Tht1,Zt10),
    asub_bw_bipai(Beta, A,Tht2,Zt20),
    asub_reduce(Alpha,Beta,Zt10,Zt20,Zt1,Zt2).

/* asub_reduce(+Alpha,+Beta,+In1,+In2,-Out1,-Out2) */

asub_reduce(pos,pair,F,PSIn,F,PSOut) :- 
    !,
    pruning(PSIn,F,PSOut).

asub_reduce(_,_,_,_,_,_) :- write('Reduction not supported!'),nl. 

% pruning needs to know representation of pair sharing
 
pruning([],_,[]). 
pruning([(X,Y)|PSIn],F,PSOut) :- 
    ( pos_vid_grounded(X,F); pos_vid_grounded(Y,F) ),
    !,
    pruning(PSIn,F,PSOut).
pruning([(X,Y)|PSIn],F,[(X,Y)|PSOut]) :- 
    pruning(PSIn,F,PSOut).
