/** A type expression consists of type atoms, and, or and not. 

    A sequence is either a list of type expressions with []
    denoting $\epsilon$ the empty sequence of top's. We use "lambda"
    to code $\Lambda$ the empty sequence of bot's. A conjunctive
    type $C$ in the table is represented by $\ll pos(C),neg(C)\rl$. 

*/


:- use_module(sicstus).
:- use_module(library(lists)).


:- op(910,xfy,and).
:- op(920,xfy,or).

/*
:- op(930,xfx,==>).

nat ==> 0. 
nat ==> s(nat).
even ==> 0.
even ==> s(odd).
odd  ==> s(even).
list(_) ==> nil.
list(X) ==> cons(X,list(X)).

% top ==> 0.                   % no need to add type rules for top. 
% top ==> s(top).                % tops are trapped and dealt with 
% top ==> nil.
% top ==> cons(top,top).

functors([0-0,0-nil,1-s,2-cons]).

*/

:- [primitive_type].


etype(E)     :- etype(E,[]). % EQ. 4

etype(E,Phi) :-              % EQ. 5 
        type_dnf(E,DC),
        apply_etype_conj(DC,Phi).

    apply_etype_conj(C or DC, Phi) :-
        !, 
        etype_conj(C,Phi),
        apply_etype_conj(DC,Phi).
    apply_etype_conj(C, Phi) :-
        !, 
        etype_conj(C,Phi).

etype_conj(C,Phi) :-         % EQ. 6               
    pos_neg(C,Ps0,Ns),
    ( Ps0=[] -> Ps=[top] ; Ps=Ps0 ), 
    ( 
        ( intersection(Ps,Ns,PsNs), PsNs \= []) 
    ->
        true
     ;
        implied(Ps,Ns,Phi)
    -> 
        true
     ; 
        conj_splitting(Ps,PsU0,PsP), 
        ( PsU0=[] -> PsU=[top] ; PsU=PsU0 ), 
        conj_splitting(Ns,NsU,NsP),
        ( 
            etype_conj_prim(PsP,NsP)
        -> 
            true
	 ;
            ( 
                NsU=[], PsU=[top] 
            ->
                fail % because \+ etype_conj_prim(PsP,NsP) 
	     ;
                PsU=[top] % NsU \== []
            ->  
                primitive_types(PTs),
                apply_add_and_test_conj(PTs,C,[Ps-Ns|Phi]),
                pos_main_functors_extended(PsP,PsU,Funcs),
                apply_eseq(Funcs,Ps,Ns,PsU,PsP,NsU,NsP,[Ps-Ns|Phi])
             ;  
                pos_main_functors_extended(PsP,PsU,Funcs),
                apply_eseq(Funcs,Ps,Ns,PsU,PsP,NsU,NsP,[Ps-Ns|Phi])
            )
        )
    ). 

    apply_add_and_test_conj([],_,_).
    apply_add_and_test_conj([P|Ps],C,Phi) :- 
        etype_conj(and(P,C),Phi),
        apply_add_and_test_conj(Ps,C,Phi). 

    apply_eseq([],_,_,_,_,_,_,_).
    apply_eseq([F|Fs],Ps,Ns,PsU,PsP,NsU,NsP,Phi) :-
        (  
            F = 0-Nullary
        ->  
            ( 
                 membership_conj_prim(PsP,NsP,Nullary) 
            ->   
                 bcal(PsU,NsU,F,SE),
                 eseq(SE,Phi)
            ;
                 true 
            )
        ;
            bcal(Ps,Ns,F,SE), % non Nullary function symbols 
            eseq(SE,Phi)
        ),
        !,
        apply_eseq(Fs,Ps,Ns,PsU,PsP,NsU,NsP,Phi).


      pos_main_functors_extended([],PsU,Funcs) :- 
          !,
          % PsU is assummed not empty and  treat nullaries specially 
          pos_main_functors(PsU,Funcs).

      pos_main_functors_extended([P|PsP],PsU,Funcs) :- 
          !,
          % PsU is assummed not empty and  treat nullaries specially 
          pos_main_functors(PsU,FU), 
          functors_splitting(FU,FU0,FU1),
          pos_main_functors([P|PsP],FP),
          intersection(FU1,FP, FUP1),
          append(FU0,FUP1,Funcs). 



          functors_splitting([],[],[]).
          functors_splitting([F|Fs],Muls,Nulls) :- 
            (
                 F = 0-_Name
            -> 
                 Muls = [F|Muls1],
                 functors_splitting(Fs,Muls1,Nulls)
            ;
                 Nulls= [F|Nulls1],
                 functors_splitting(Fs,Muls,Nulls1)
            ). 

   
eseq(SE,Phi) :-              % EQ. 7 
        dnf_and_or(SE,DC),
        apply_eseq_conj(DC,Phi).

    apply_eseq_conj(C or DC, Phi) :-
        !, 
        eseq_conj(C,Phi),
        apply_eseq_conj(DC,Phi).
    apply_eseq_conj(C, Phi) :-
        !, 
        eseq_conj(C,Phi).

eseq_conj(C,Phi) :-          % EQ. 8
    (
        in_conj(lambda,C)
    ->  
        true 
     ; 
        eseq_conj_1(C,Phi)
    ).

    eseq_conj_1(C,Phi) :- 
        dimension(C,Dim),   
        Dim =\= 0,	     % fail if Dim==0
        proj_fst(C,E,C1),
        ( 
            etype(E,Phi)
        -> 
            true
	 ;
            eseq_conj_1(C1,Phi)
        ).

/** proj_fst(+CSEi,-TEo, -CSEo) TEo is the conjunction of the
    first type expressions in the sequences in CSEi. CSEo is the
    result of removing the first type expression from each
    sequence in CSEi. */

proj_fst([TE|TEs] and CSE, TE and TEo, TEs and CSEo) :-
    !,
    proj_fst(CSE, TEo, CSEo).
proj_fst([TE|TEs], TE, TEs).


/** dimension(+CSE,-N) N=\|CSE\|. */

dimension(S and _, N) :-
    !, 
    length(S,N). 
dimension(S, N) :-
    length(S,N).


/** implied(+Ps,+Ns,+Phi) The emptiness of the conjunctive type
    represented by Ps and Ns is implied by that of a tabled
    conjunctive type in Phi. */

implied(Ps,Ns,Phi) :- 
    member(Ps1-Ns1,Phi),
    subset(Ps1,Ps),
    subset(Ns1,Ns),
    !.


/** pos_neg(+C,-Ps,-Ns) Ps= pos(C), and Ns=neg(C). +C is a conjunctive
    type. Ps and Ns are sorted according to the term ordering. */

pos_neg(C,Ps,Ns) :- 
    pos_neg_1(C,Ps0,Ns0),
    sort(Ps0,Ps),
    sort(Ns0,Ns).

pos_neg_1(L and C, Ps, Ns) :-
    !,
    (
        L = not(A)
    ->
        Ps = Ps0, Ns=[A|Ns0]
    ;
        Ps = [L|Ps0], Ns=Ns0 
    ),
    pos_neg_1(C, Ps0, Ns0).

pos_neg_1(L, Ps, Ns) :-
    (
        L = not(A)
    ->
        Ps = [], Ns=[A]
    ;
        Ps = [L], Ns=[] 
    ).


/* unravel_complement(+Fi,-Fo) Fi is equivalent to Fo and Fo is a
   formula where negation is applied to atomic formulae (type
   atoms or sequences of type expressions. Fi is either is type
   expression or sequence expression. */

unravel_complement(X or  Y, Xp or  Yp) :-
        !,
        unravel_complement(X, Xp),
        unravel_complement(Y, Yp).
unravel_complement(X and Y, Xp and Yp) :-
        !,
        unravel_complement(X, Xp),
        unravel_complement(Y, Yp).
unravel_complement(not(X),   NotXp)     :-
        !,
        complement_de_morgan(X, NotXp).
unravel_complement(X,X). % X is a type atom

complement_de_morgan(X and Y, DMx or  DMy) :-
        !,
        complement_de_morgan(X,DMx),
        complement_de_morgan(Y,DMy).
complement_de_morgan(X or  Y, DMx and DMy) :-
        !,
        complement_de_morgan(X,DMx),
        complement_de_morgan(Y,DMy).
complement_de_morgan(not(X), Xp) :-
        !,
        unravel_complement(X, Xp).
complement_de_morgan(top, bot) :- !. % X is a type atom.
complement_de_morgan(bot, top) :- !. % X is a type atom.
complement_de_morgan(X, not(X)). % X is a type atom.

%:-unravel_complement(not(nat and list or not int),L),write(L),nl.

/** dnf_and_or(+F,-Dnf) Dnf is the disjunctive normal form of F
    where F is formed of atoms, and's and or's */

dnf_and_or(X or Y, XYdnf) :-
        !, 
        dnf_and_or(X, Xdnf),
        dnf_and_or(Y, Ydnf),
        concate_disj_disj(Xdnf,Ydnf,XYdnf).
dnf_and_or(X and Y, XYdnf)       :-
        !, 
        dnf_and_or(X, Xdnf),
        dnf_and_or(Y, Ydnf),
        product_disj_disj(Xdnf,Ydnf,XYdnf).
dnf_and_or(X,X).                 % X is a type atom

product_disj_disj(C or Y, Z, CZYZ) :-
        !,
        product_disj_conj(Z, C, CZ),
        product_disj_disj(Y, Z, YZ),
        concate_disj_disj(CZ,YZ,CZYZ).
product_disj_disj(C, Z, CZ)            :-
        product_disj_conj(Z, C, CZ).

product_disj_conj(C1 or Y, C0, C0C1 or C0Y) :-
        !,
        concate_conj_conj(C0, C1, C0C1),
        product_disj_conj(Y,  C0, C0Y).
product_disj_conj(C1, C0, C0C1)            :-
        concate_conj_conj(C0, C1, C0C1).

% This version doesn't remove duplicate literals from a conjunct

concate_conj_conj(L and C0, C1,  L and C0C1) :- 
        !,
        concate_conj_conj(C0, C1,  C0C1).
concate_conj_conj(L, C1,  L and C1). 

/* 
concate_conj_conj(L and C0, C1,  Co) :- 
        !,
        concate_conj_conj(C0, C1,  C0C1),
        ( 
             in_conj(L,C0C1)
        -> 
             Co = C0C1
        ;
             Co = (L and C0C1)
        ).
concate_conj_conj(L, C1,  L and C1). 
*/


/** in_conj(+L,+C) L is a literal in conjunction C assumming ground
    representation of type and sequence expressions . */

in_conj(L,L and _) :-  
        !.
in_conj(L,_ and C) :-
        !,
        in_conj(L,C).
in_conj(L,L).


concate_disj_disj(L or C0, C1,  L or C0C1) :- 
        !,
        concate_disj_disj(C0, C1,  C0C1).
concate_disj_disj(L, C1,  L or C1). 


/* type_dnf(Ei,Eo) Eo is the disjunctive normal form of Ei. Ei and Eo
   are type expressions. */

type_dnf(E,DNF) :- 
        unravel_complement(E,Ec), 
        dnf_and_or(Ec,DNF).

/*
fwrite(L1 or L2) :- 
   !,
   write(L1),nl,
   fwrite(L2).

fwrite(L) :- 
   write(L),nl.
*/

/*
:- type_dnf(not(nat and list or not int) and 
           ( nat and not list or not(int and nat) ),L), 
   fwrite(L),nl.
*/

/* 
seq_dnf(SE,DNF) :- 
        dnf_and_or(SE,DNF).
*/


/* negate_disj(+Seqs,-SE) SE=push(\tneg\tor Seqs). It is assumed
   Seqs=\=[]. */

negate_disj([S1,S2|Ss], Se1 and Se2) :-
        !,
        negate_seq(S1,Se1),
        negate_disj([S2|Ss], Se2).
negate_disj([S1], Se1) :-
        !,
        negate_seq(S1,Se1).

negate_seq([],lambda).
negate_seq([E|Es],DS) :- 
        ones(Es,Ones),
        negate_seq_ne([E|Es],[],Ones,DS).

negate_seq_ne([E], Prefix, [], S)        :- 
        append(Prefix, [not(E)], S).
negate_seq_ne([E1,E2|Es], Prefix, [One|Suffix], S or DS) :-
        append(Prefix, [not(E1),One|Suffix], S),
        negate_seq_ne([E2|Es], [One|Prefix], Suffix, DS).

ones([],[]).
ones([_|Es], [top|Ones]) :- ones(Es, Ones).


/* pos_main_functors(Pos,-Funcs). Funcs = \cap_{\alpha\in Pos}
   \fcal(\alpha). It is assumed that Pos =\= []. */ 

pos_main_functors([T1,T2|Ts], Funcs) :- 
    !,
    pos_main_functors([T2|Ts],Funcs2),
    type_main_functors(T1,Funcs1),
    intersection(Funcs2,Funcs1,Funcs).

pos_main_functors([T1], Funcs1) :- 
    type_main_functors(T1,Funcs1).

/* type_main_functors(+Type,-Funcs) Set of main functors of terms
   of type Type. */

type_main_functors(Type, Funcs) :-
    (
       Type = top 
    -> 
       %functors(Funcs) % efficiency barrier
       typedef_functors(Funcs) % assumming Funcs is the set of all functors
       %( setof(Func,Typ^type_main_func(Typ,Func),Funcs) -> true ; Funcs=[] )
       % 
    ;
       ( setof(Func,type_main_func(Type,Func),Funcs) -> true ; Funcs = []  )
    ).

type_main_func(Type,Arity-Name) :-  % Type \== top,
    %Type ==> Body, 
    %functor(Body,Name, Arity).
    typedef(Name/Arity,_,Type).

/** bcal(+Ps,+Ns,+Func,-SeqExp) SeqExp is the sequence expression
    for the conjunctive type $C$ with $Ps=pos(C)$ and $Ns=neg(C)$
    and the functor Func. +Ps is a non-empty set of type atoms
    whilst +Ns is a set of type atoms. */

bcal(Ps,Ns,Func,SeqExp) :- 
    bcal_pos(Ps,Func,SeqExpPos),        
    bcal_neg(Ns,Func,SeqExpPos,SeqExp). 

bcal_pos([T], F, SE) :- 
    arg_type_seqs(T, F, SL),
    bigOrOfSeqs(SL, SE).

bcal_pos([T1,T2|Ts], F, SE1 and SE2) :- 
    arg_type_seqs(T1, F, SL),
    bigOrOfSeqs(SL, SE1),
    bcal_pos([T2|Ts], F, SE2).

bcal_neg([], _, SE, SE).
bcal_neg([T|Ts], F, SE0, SE) :- 
    arg_type_seqs(T, F, SL),
    (
        SL == []
    ->
        bcal_neg(Ts, F, SE0, SE)
    ;
        negate_disj(SL, SE1),
        bcal_neg(Ts, F, SE0 and SE1, SE)
    ).  

/** bigOrOfSeqs(+Seqs,-OrF) OrF = \/ Seqs */

bigOrOfSeqs([],lambda).
bigOrOfSeqs([S], S).
bigOrOfSeqs([S1,S2|Ss], S1 or S2Ss) :-
    bigOrOfSeqs([S2|Ss],S2Ss).


/* arg_type_seqs(+Type,+Func,Seqs) $Seqs=\acal_{Type}^{Func}$
   Func=Arity-Name. */

arg_type_seqs(Type,Func,Seqs)     :-
        findall(Seq, arg_type_seq(Type, Func, Seq), Seqs).

arg_type_seq(Type,Arity-_,Seq) :-
    Type == top, 
    !,
    ntops(Arity,Seq).

arg_type_seq(Type,Arity-Name,Seq) :- 
%    Type \== top,
    %functor(Body,Name,Arity),
    %Type ==> Body,
    %Body =.. [Name|Seq].
    typedef(Name/Arity,Seq,Type).

ntops(0,[]).
ntops(N,[top|Ones]) :-
    N>0,
    N1 is N-1,
    ntops(N1,Ones).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
/**

?- pos_main_functors([list(bot)],L).

L = [0-nil, 2-cons] ;


2 ?- etype(nat and not odd and not even).

Yes
3 ?- etype(list(bot)).

No

4 ?- etype(list(nat and not even and not odd)).

No

5 ?- etype(list(even and not nat)).

No
6 ?- etype(list(even) and not list(nat)).

Yes



*/

%%%%%%%%%%%%%%%%%%type%analysis%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

seqSet_leq(SeqSet1,[]) :- 
    !,
    bigOrOfSeqs(SeqSet1,Seq1),
    eseq(Seq1,[]). 

seqSet_leq(SeqSet1,SeqSet2) :- 
    bigOrOfSeqs(SeqSet1,Seq1),
    ( 
        SeqSet2=[] 
    ->
        eseq(Seq1,[])
    ;
        negate_disj(SeqSet2,Seq2), % Seq2 = push(tneg\tor SeqSet2)
        eseq(Seq1 and Seq2,[])
    ).
 


/* 
seqSet_leq([[nat,list(nat)]],[[top,top]]).

Yes
15 ?- seqSet_leq([[nat,list(nat)]],[[nat,top]]).

Yes
16 ?- seqSet_leq([[nat,list(nat)]],[[nat,list(top)]]).

Yes
17 ?- seqSet_leq([[nat,list(nat)]],[[nat,list(bot)]]).

No
18 ?- seqSet_leq([[bot,list(nat)]],[[nat,list(bot)]]).

Yes

22 ?- seqSet_leq([[bot,list(nat)]],[]).

24 ?- seqSet_leq([],[]).

Yes
*/

seqSet_reduce(SSi,SSo) :- 
    seqSet_reduce(SSi,[],SSo).


seqSet_reduce([],SSu,SSu).
seqSet_reduce([Si|SSi],SSu, SSo) :- 
    append(SSu,SSi,SSn),
    (
        seqSet_leq([Si],SSn)
    -> 
        seqSet_reduce(SSi,SSu,SSo)
    ;
        seqSet_reduce(SSi,[Si|SSu],SSo)
    ).

/*

34 ?- !seq.
seqSet_reduce([[nat,list(nat)],[nat,list(top)],[bot,list(top)]],V).

V = [[nat, list(top)]] ;

35 ?- seqSet_reduce([[nat],[even],[odd]],V).

V = [[odd], [even]] ;

36 ?- seqSet_reduce([[even],[nat],[odd]],V).

V = [[nat]] ;

*/






