:- module(ordsets, [ 
	ord_setproduct/3,
	ord_subset/2,
	ord_add_element/3,
	ord_intersection/3,
	ord_subtract/3,
	ord_union/3
   ]).

ord_union([],S,S) :-
    !.
ord_union(S,[],S) :-
    !.
ord_union([H1|S1],[H2|S2],[H1|S3]) :-
    H1 @< H2,
    !,
    ord_union(S1,[H2|S2],S3).
ord_union([H1|S1],[H2|S2],[H1|S3]) :-
    H1 == H2,
    !,
    ord_union(S1,S2,S3).
ord_union([H1|S1],[H2|S2],[H2|S3]) :-
    H1 @> H2,
    !,
    ord_union([H1|S1],S2,S3).

ord_add_element([],E,[E]).
ord_add_element([E1|S1],E,[E,E1|S1]) :-
    E @< E1,
    !.
ord_add_element([E1|S1],E,[E1|S1]) :-
    E == E1,
    !.
ord_add_element([E1|S1],E,[E1|S2]) :-
    ord_add_element(S1,E,S2).

ord_setproduct(S1,S2,So) :- 
    ord_setproduct_dl(S1,S2,So-[]). % using difference list

ord_setproduct_dl([],_,S-S).
ord_setproduct_dl([E|S1],S2,Front-Tail) :-
    scaling_dl(S2,E,Front-Middle),
    !,
    ord_setproduct_dl(S1,S2,Middle-Tail).

scaling_dl([],_,S-S).
scaling_dl([E1|S1],E,[(E,E1)|S2]-Tail) :-
    scaling_dl(S1,E,S2-Tail).

ord_subset([],_S) :-
    !.
ord_subset([H1|S1],[H2|S2]) :-
    H1 == H2,
    !,
    ord_subset(S1,S2).
ord_subset([H1|S1],[H2|S2]) :-
    H1 @> H2,
    !,
    ord_subset([H1|S1],S2).

ord_intersection([],_S,[]) :-
    !.
ord_intersection(_S,[],[]) :-
    !.
ord_intersection([H1|S1],[H2|S2],S3) :-
    H1 @< H2,
    !,
    ord_intersection(S1,[H2|S2],S3).
ord_intersection([H1|S1],[H2|S2],[H1|S3]) :-
    H1 == H2,
    !,
    ord_intersection(S1,S2,S3).
ord_intersection([H1|S1],[H2|S2],S3) :-
    H1 @> H2,
    !,
    ord_intersection([H1|S1],S2,S3).

ord_subtract([],_S,[]) :-
    !.
ord_subtract(S,[],S) :-
    !.
ord_subtract([H1|S1],[H2|S2],[H1|S3]) :-
    H1 @< H2,
    !,
    ord_subtract(S1,[H2|S2],S3).
ord_subtract([H1|S1],[H2|S2],S3) :-
    H1 == H2,
    !,
    ord_subtract(S1,S2,S3).
ord_subtract([H1|S1],[H2|S2],S3) :-
    H1 @> H2,
    !,
    ord_subtract([H1|S1],S2,S3).







