:- module(pos, [ 
		pos_eq_constraints/2,
                pos_init/0,
		pos_list_print/1,
		pos_unify_entry/5,
		pos_unify_exit/6,
		pos_bipai/3,
		pos_compile/2,
		pos_bot/2,
		pos_ident/2,
		pos_join/3,
		pos_leq/2,
		pos_internalize/2,
		pos_externalize/2,
		pos_bw_compile/2,
		pos_bip_has_cond/2,
		pos_top/2,
		pos_meet/3,
		pos_bw_unify_entry/5,
		pos_bw_unify_exit/6,
		pos_bw_bipai/3,
		pos_vid_grounded/2
	       ]).





:- use_module('basics', [
      mgu/3,
      vars/2,
      variable/1,
      variable_id/2,
      variable_ids/2    ]).


:- use_module('BDDS/bdds').

pos_vid_grounded(Id,F) :- 
    bdd_variable(Id,Fid),
    bdd_implies(F,Fid,1).

pos_unify_entry(Sub,Sigma,NVarR,NVar,Zeta) :- 
    bdd_rename_apart(Sigma,NVarR,NVar,SigmaR), 
    bdd_and(Sub,SigmaR,SigmaRSub),
    bdd_restrict(NVar,SigmaRSub,Zeta).

pos_unify_exit(Sub,Sigma,NVarR,Theta,NVar,Zeta) :-
    bdd_rename_apart(Sigma,NVarR,NVar,SigmaR), 
    bdd_and(SigmaR,Theta,SigmaRTheta),
    bdd_and(Sub,SigmaRTheta,SigmaRThetaSub),
    bdd_restrict(NVar,SigmaRThetaSub,Zeta).

pos_init :- 
    bdd_start.


pos_ident(_,1).

pos_top(_,1).

pos_bot(_NVar,0).

pos_join(Zeta,Eta,Beta) :- 
    bdd_or(Zeta,Eta,Beta).

pos_leq(F1,F2) :- 
    bdd_implies(F1,F2,1). 

pos_externalize(F,F).

pos_internalize(F,F).

pos_meet(Zeta,Eta,Beta) :- 
    bdd_and(Zeta,Eta,Beta).

pos_bw_unify_entry(CE,Sigma,NVarR,NVar,Eta) :-
    bdd_rename_apart(Sigma,NVarR,NVar,SigmaR),
    bdd_implies(CE,SigmaR,Zeta),
    bdd_proj_forall(Zeta,NVar,Eta).

pos_bw_unify_exit(CE,Answer,NVarR,Success,NVar,F3) :-
    bdd_rename_apart(Answer,NVarR,NVar,AnswerR),
    bdd_and(CE,AnswerR,F1), 
    bdd_restrict(NVar,F1,F2),
    bdd_implies(F2,Success,F3).



/* pos_bip_has_cond(+Call,-Form) +Form must be true for Call to be
   executed.  Fails if +Call does not have any condition. */

pos_bip_has_cond(abolish(P),Fo) :- vars(P,Vs),list_to_form(Vs,Fo).
pos_bip_has_cond(erase(P),Fo) :- vars(P,Vs),list_to_form(Vs,Fo).
pos_bip_has_cond(format(Fmt),Fo) :- vars(Fmt,Vs),list_to_form(Vs,Fo).
pos_bip_has_cond(format(Fmt,_),Fo) :- vars(Fmt,Vs),list_to_form(Vs,Fo).
pos_bip_has_cond(format(Strm,Fmt,_),Fo) :- vars((Strm,Fmt),Vs),list_to_form(Vs,Fo).

pos_bip_has_cond(=:=(N1,N2),Fo) :- vars((N1,N2),Vs),list_to_form(Vs,Fo).
pos_bip_has_cond(=\=(N1,N2),Fo) :- vars((N1,N2),Vs),list_to_form(Vs,Fo).
pos_bip_has_cond(<(N1,N2),Fo) :- vars((N1,N2),Vs),list_to_form(Vs,Fo).
pos_bip_has_cond(>(N1,N2),Fo) :- vars((N1,N2),Vs),list_to_form(Vs,Fo).
pos_bip_has_cond(=<(N1,N2),Fo) :- vars((N1,N2),Vs),list_to_form(Vs,Fo).
pos_bip_has_cond(>=(N1,N2),Fo) :- vars((N1,N2),Vs),list_to_form(Vs,Fo).
pos_bip_has_cond(tab(T),Fo) :- vars(T,Vs),list_to_form(Vs,Fo).
pos_bip_has_cond(put(T),Fo) :- vars(T,Vs),list_to_form(Vs,Fo).
pos_bip_has_cond(is(_,T),Fo) :- vars(T,Vs),list_to_form(Vs,Fo).
pos_bip_has_cond(arg(T,_,_),Fo) :- vars(T,Vs),list_to_form(Vs,Fo).
pos_bip_has_cond(statistics(T,_),Fo) :- vars(T,Vs),list_to_form(Vs,Fo).

pos_bip_has_cond(=..(Var,[F|_]), Fo) :- 
    variable(Var), 
    vars(F,Vs), 
    list_to_form(Vs,Fo).

pos_bip_has_cond(name(A,L),Fo) :- 
    vars(A,VsInA),
    list_to_form(VsInA,FA),
    vars(L,VsInL),
    list_to_form(VsInL,FL),
    bdd_or(FA,FL,Fo).


/* pos_bw_bipai(+Call,+Fr,-Fl) All states in Fl will be transformed to
   a state in Fr by Call should Call succeed. */

pos_bw_bipai(copy,F,F).

pos_bw_bipai(implies(Fc),Fi,Fo) :-
    bdd_implies(Fc,Fi,Fo).

pos_bw_bipai(exists(Vids),Fr,Fl) :- 
    exists_var_list(Vids,Fr,Fl). 

pos_bw_bipai(or(exists(VidsA),exists(VidsL)),Fr,Fl) :- 
    exists_var_list(VidsA,Fr,FlA),
    exists_var_list(VidsL,Fr,FlL),
    bdd_or(FlA,FlL,Fl).

pos_bw_bipai(if(Ft,Vids),Fr,Fl) :- 
    (
        bdd_implies(Fr,Ft,1)
     ->
        exists_var_list(Vids,Fr,Fl)
     ; 
        Fl=Fr
    ).


/* pos_bw_compile(+Call,+Fr,-Fl) All states in Fl will be transformed to
   a state in Fr by Call should Call succeeds. */


pos_bw_compile(abolish(_),copy).
pos_bw_compile(assert(_),copy).
pos_bw_compile(retract(_),copy). % might leave argument un-groundned 
pos_bw_compile(erase(_),copy).
pos_bw_compile(format(_),copy).
pos_bw_compile(format(_,_),copy).
pos_bw_compile(format(_,_,_),copy).

pos_bw_compile(var(_),copy).   /* var(X) doesn't change the state.*/
pos_bw_compile(nonvar(_),copy).
pos_bw_compile(==(_,_),copy).
pos_bw_compile(put(_),copy).
pos_bw_compile(\==(_,_),copy).
pos_bw_compile(tab(_t),copy).
pos_bw_compile(=:=(_,_),copy).
pos_bw_compile(=\=(_,_),copy).
pos_bw_compile(<(_,_),copy).
pos_bw_compile(>(_,_),copy).
pos_bw_compile(=<(_,_),copy).
pos_bw_compile(>=(_,_),copy).
pos_bw_compile(ground(_),copy).
pos_bw_compile(integer(_),copy).
pos_bw_compile(number(_),copy).
pos_bw_compile(repeat,copy). 
pos_bw_compile(listing,copy).
pos_bw_compile(listing(_),copy).
pos_bw_compile(float(_),copy).
pos_bw_compile(compound(_),copy).
pos_bw_compile(portray_clause(_),copy).
pos_bw_compile(\=(_,_),copy).
pos_bw_compile(!,copy).
pos_bw_compile(display(_),copy).
pos_bw_compile(nl,copy).
pos_bw_compile(print(_),copy).
pos_bw_compile(write(_),copy).
pos_bw_compile(writeq(_),copy).
pos_bw_compile(atom(_),copy).
pos_bw_compile(atomic(_),copy).
pos_bw_compile(@<(_,_),copy).
pos_bw_compile(@>(_,_),copy).
pos_bw_compile(@=<(_,_),copy).
pos_bw_compile(@>=(_,_),copy).
pos_bw_compile(true,copy).

pos_bw_compile(=(L,R),implies(Fe)) :- 
       mgu(L,R,E),
       pos_eq_constraints(E,Fe). % called only when mgu(L,R,E) succeeds
pos_bw_compile(sort(L,R),implies(Fe)) :- 
       mgu(L,R,E),
       pos_eq_constraints(E,Fe). % called only when mgu(L,R,E) succeeds
pos_bw_compile(keysort(L,R),implies(Fe)) :- 
       mgu(L,R,E),
       pos_eq_constraints(E,Fe). % called only when mgu(L,R,E) succeeds
pos_bw_compile(is(L,_R),exists(Vids)) :- 
    /* Fl <==> (\exits vars(L).Fr)
       For is(L,R) to succeeds, L must be a variable or a ground term 
       and R must be a ground term. The requirement that R be ground
       is already placed by pos_bip_has_cond/2. Thus, for every 
       program state in Fl, if is(L,R) succeeds then is(L,R) changes
       the program state in Fl to a program state in Fr. 
    */
    vars(L,Vs), variable_ids(Vs,Vids). 

pos_bw_compile(name(A,L),or(exists(VidsA),exists(VidsL))) :- 
    /* name(A,L) either instantiate A to ground or instantiate L to ground.
       Note that name(A,L) grounds both A and L since Prolog requires that
       either A or L be ground before name(A,L) is executed.      
    */
    vars(A,VsA),variable_ids(VsA,VidsA),
    vars(L,VsL),variable_ids(VsL,VidsL). 

pos_bw_compile(length(_,L),exists(Vids)) :- 
    /* So, L can be any term before execution. */
    vars(L,Vs),variable_ids(Vs,Vids).



pos_bw_compile(compare(O,_,_),exists(Vids)) :- 
    /*  So, O can be any term before execution.
    */
    vars(O,Vs), variable_ids(Vs,Vids).


pos_bw_compile(recorda(_,_,Ref),exists(Vids)) :- 
    /*  So, O can be any term before execution.
    */
    vars(Ref,Vs), variable_ids(Vs,Vids).

pos_bw_compile(recorded(_,_,Ref),exists(Vids)) :- 
    /*  So, O can be any term before execution.
    */
    vars(Ref,Vs), variable_ids(Vs,Vids).

pos_bw_compile(statistics(_,O),exists(Vids)) :- 
    /*  So, O can be any term before execution.  */
    vars(O,Vs),
    variable_ids(Vs,Vids).

pos_bw_compile(read(_),copy). 
    /* Though read(L) instantiates L, it doesn't necessarily 
       instantiate L to a ground term. So, it doesn't necessary 
       change the groundness of any variable in L.
    */

pos_bw_compile(arg(_,_,_),copy). 
    /* This can be made more precise if a more refined domain is used. 
    */

pos_bw_compile(=..(_,_),copy).
    /* This MIGHT be made more precise 
    */

pos_bw_compile(functor(T,F,A),if(Ft,Vids)) :- 
    vars(T,VsT),   list_to_form(VsT,Ft), 
    vars((F,A),Vs),variable_ids(Vs,Vids).



/* exists_var (V,Fi,Fo) is true iff Fo <==> exists V. Fi */

exists_var_list([],F,F). 
exists_var_list([V|Vs],Fi,Fo) :-
    bdd_restrict(V,Fi,Fm),
    exists_var_list(Vs,Fm,Fo). 

list_to_form([],True)  :-
    bdd_true(True). 
list_to_form([V|Vs],F) :-
    variable_id(V,Id),
    bdd_variable(Id,FId),
    list_to_form(Vs,FVs),
    bdd_and(FId,FVs,F). 


pos_eq_constraints([],True) :-
    bdd_true(True).
pos_eq_constraints([X/T|E],F) :-
    pos_eq_constraints(E,Fe),
    variable_id(X,Xid),
    vars(T,Vs),
    variable_ids(Vs,Tids),
    bdd_iff(Xid,Tids,Fxt),
    bdd_and(Fxt,Fe,F).

/* Fo <==> \forall \Phi(NVI). Fi where VI is the set of variables of
   interest, \Phi is the tagging substitution. */

bdd_proj_forall(Fi,Max,Fo) :- 
    bdd_implies(Fi,0,NFi), % NFi is the negation of Fi
    bdd_restrict(Max,NFi,NFo), /* NFo is \exists VI. NFi */
    bdd_implies(NFo,0,Fo). % Fo is the negation of NFo

pos_compile(abolish(_),copy).
pos_compile(assert(_),copy).
pos_compile(retract(_),copy).
pos_compile(erase(_),copy).
pos_compile(recorda(_,_,_),copy).
pos_compile(recorded(_,_,_),copy).
pos_compile(format(_),copy).
pos_compile(format(_,_),copy).
pos_compile(format(_,_,_),copy).
pos_compile(read(_),copy).
pos_compile(arg(_,_,_),copy).
pos_compile(\=(_,_),copy).
pos_compile(var(_),copy).   /* var(X) doesn't change the state.*/
pos_compile(nonvar(_),copy).
pos_compile(==(_,_),copy).
pos_compile(put(_),copy).
pos_compile(\==(_,_),copy).
pos_compile(repeat,copy). 
pos_compile(listing,copy).
pos_compile(listing(_),copy).
pos_compile(compound(_),copy).
pos_compile(portray_clause(_),copy).
pos_compile(!,copy).
pos_compile(display(_),copy).
pos_compile(nl,copy).
pos_compile(print(_),copy).
pos_compile(write(_),copy).
pos_compile(writeq(_),copy).
pos_compile(@<(_,_),copy).
pos_compile(@>(_,_),copy).
pos_compile(@=<(_,_),copy).
pos_compile(@>=(_,_),copy).
pos_compile(true,copy).

pos_compile(=\=(N1,N2),add(FVs)) :- vars((N1,N2),Vs),list_to_form(Vs,FVs).
pos_compile(tab(N),add(FVs)) :- vars(N,Vs),list_to_form(Vs,FVs).
pos_compile(=:=(N1,N2),add(FVs)) :- vars((N1,N2),Vs),list_to_form(Vs,FVs).
pos_compile(<(N1,N2),add(FVs)) :- vars((N1,N2),Vs),list_to_form(Vs,FVs).
pos_compile(>(N1,N2),add(FVs)) :- vars((N1,N2),Vs),list_to_form(Vs,FVs).
pos_compile(=<(N1,N2),add(FVs)) :- vars((N1,N2),Vs),list_to_form(Vs,FVs).
pos_compile(>=(N1,N2),add(FVs)) :- vars((N1,N2),Vs),list_to_form(Vs,FVs).
pos_compile(name(A,L),add(FVs)) :- vars((A,L),Vs),list_to_form(Vs,FVs).
pos_compile(is(L,_),add(FVs)) :- vars(L,Vs),list_to_form(Vs,FVs).
pos_compile(length(_,L),add(FVs)) :- vars(L,Vs),list_to_form(Vs,FVs).
pos_compile(compare(O,_,_),add(FVs)) :- vars(O,Vs),list_to_form(Vs,FVs).
pos_compile(functor(_,F,A),add(FVs)) :- vars((F,A),Vs),list_to_form(Vs,FVs).
pos_compile(statistics(_,O),add(FVs)) :- vars(O,Vs),list_to_form(Vs,FVs).
pos_compile(ground(N),add(FVs)) :- vars(N,Vs),list_to_form(Vs,FVs).
pos_compile(integer(N),add(FVs)) :- vars(N,Vs),list_to_form(Vs,FVs).
pos_compile(number(N),add(FVs)) :- vars(N,Vs),list_to_form(Vs,FVs).
pos_compile(float(N),add(FVs)) :- vars(N,Vs),list_to_form(Vs,FVs).
pos_compile(atom(N),add(FVs)) :- vars(N,Vs),list_to_form(Vs,FVs).
pos_compile(atomic(N),add(FVs)) :- vars(N,Vs),list_to_form(Vs,FVs).

pos_compile(=(L,R),add(Fe)) :- 
    /* If L and R unify then L=R is a unification. 
       Otherwise, the right point will not be reached. */       
       mgu(L,R,E),
       pos_eq_constraints(E,Fe).

pos_compile(sort(L,R),add(Fe)) :- 
    /* If L and R unify then L=R is a unification. 
       Otherwise, the right point will not be reached. */       
       mgu(L,R,E),
       pos_eq_constraints(E,Fe).

pos_compile(keysort(L,R),add(Fe)) :- 
    /* If L and R unify then L=R is a unification. 
       Otherwise, the right point will not be reached. */       
       mgu(L,R,E),
       pos_eq_constraints(E,Fe).

pos_compile(=..(L,R),add(Fe12)) :-
   vars(L,VsL),
   vars(R,VsR),
   list_to_form(VsL,FVsL),
   list_to_form(VsR,FVsR),
   bdd_implies(FVsL,FVsR,Fe1), 
   bdd_implies(FVsR,FVsL,Fe2), 
   bdd_and(Fe1,Fe2,Fe12).

/* bipai(+Eqs,+Fl,-Fr) All states in Fl will be transformed to
   a state in Fr by Call if Call succeeds. */

pos_bipai(copy,F,F).
pos_bipai(add(FVs),Fl,Fr) :-
    bdd_and(FVs,Fl,Fr).


pos_list_print([]).
pos_list_print([A|As]) :- 
    write('>>>>>>'),nl,
    write(A), write(' : '), nl,
    bdd_print(A), nl, bdd_true(_),
    write('<<<<<<'),nl,nl, 
    !,
    pos_list_print(As).

