/* coding(+Ps,+Ns,Num) returns a unique code for a conjunction of
   primitive type literals */

coding(Ps,Ns,Num) :- 
   pcoding(Ps,PNum),
   ncoding(Ns,NNum),
   Num is PNum+NNum. 

pcoding(Ps,Num) :- 
   pcoding(Ps,0,Num).

   pcoding([],Num,Num).
   pcoding([P|Ps],Num0,Num) :- 
      pcode(P,Bit),
      Num1 is Num0+Bit,
      pcoding(Ps,Num1,Num).

ncoding(Ps,Num) :- 
   ncoding(Ps,0,Num).

   ncoding([],Num,Num).
   ncoding([P|Ps],Num0,Num) :- 
      ncode(P,Bit),
      Num1 is Num0+Bit,
      ncoding(Ps,Num1,Num).

membership_conj_prim(Ps,Ns,X) :- 
    membership_pos_prim(Ps,X),
    membership_neg_prim(Ns,X).

    membership_pos_prim([],_). 
    membership_pos_prim([P|Ps],X) :- 
        membership_prim(P,X),
        membership_pos_prim(Ps,X).

    membership_neg_prim([],_). 
    membership_neg_prim([P|Ps],X) :- 
        \+ membership_prim(P,X),
        membership_neg_prim(Ps,X).

conj_splitting([],[],[]).
conj_splitting([A|As],Us,Ps) :- 
    (
         prim_type(A)
    -> 
         Ps = [A|Ps1],
         conj_splitting(As,Us,Ps1)
     ;
         Us= [A|Us1],
         conj_splitting(As,Us1,Ps)
    ). 



etype_conj_prim(Ps,Ns) :-  % assumming Ps and Ns do not intersect
    % provided that all generated facts for '$etype_conj_prim_coded'/1 are included
    coding(Ps,Ns,Num), 
    '$etype_conj_prim_coded'(Num).

% the following is specific to a type system and need be carried from 
% generating program tool_rtype.pl 

primitive_types([atom,atomic,float,integer,number]). % primitive types

pcode(atom,   1).    % code for an atom in positive part
pcode(atomic, 2).
pcode(float,  4).
pcode(integer,8).
pcode(number,16).


ncode(atom,     512). % code for an atom in negative part
ncode(atomic,  1024).
ncode(float,   2048).
ncode(integer, 4096).
ncode(number,  8192).




% add following for each primitive type 

prim_type(atom).
prim_type(atomic).
prim_type(float).
prim_type(integer).
prim_type(number).



membership_prim(atom,  X) :- 
    (   
        name(X,S), 
        \+ S=[36,97,114,98,105,116,114,97,114,121,95|_], 
        % X is not of the form $arbitrary_
        atom(X) 
     -> 
        true 
     ; 
        X = '$arbitrary_atom_constant' 
    ).
membership_prim(float, X) :- 
    ( float(X) -> true ; X = '$arbitrary_float_constant' ).
membership_prim(integer,X):- 
    ( integer(X) -> true ; X = '$arbitrary_integer_constant' ).


% following clauses are for non-leave primitive types 
 
membership_prim(number,X) :- 
    ( membership_prim(integer,X) -> true ; membership_prim(float,X) ). 

membership_prim(atomic,X) :- 
    ( 
        membership_prim(number,X) 
    -> 
        true 
    ; 
        membership_prim(atom,X) 
    ).


%%%%output_of_generate/0_in_"tool_rtype.pl"%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

'$etype_conj_prim_coded'(31).%empty([atom,atomic,float,integer,number],[])
'$etype_conj_prim_coded'(8207).%empty([atom,atomic,float,integer],[number])
'$etype_conj_prim_coded'(4119).%empty([atom,atomic,float,number],[integer])
'$etype_conj_prim_coded'(12295).%empty([atom,atomic,float],[integer,number])
'$etype_conj_prim_coded'(2075).%empty([atom,atomic,integer,number],[float])
'$etype_conj_prim_coded'(10251).%empty([atom,atomic,integer],[float,number])
'$etype_conj_prim_coded'(6163).%empty([atom,atomic,number],[float,integer])
'$etype_conj_prim_coded'(1053).%empty([atom,float,integer,number],[atomic])
'$etype_conj_prim_coded'(9229).%empty([atom,float,integer],[atomic,number])
'$etype_conj_prim_coded'(5141).%empty([atom,float,number],[atomic,integer])
'$etype_conj_prim_coded'(13317).%empty([atom,float],[atomic,integer,number])
'$etype_conj_prim_coded'(3097).%empty([atom,integer,number],[atomic,float])
'$etype_conj_prim_coded'(11273).%empty([atom,integer],[atomic,float,number])
'$etype_conj_prim_coded'(7185).%empty([atom,number],[atomic,float,integer])
'$etype_conj_prim_coded'(15361).%empty([atom],[atomic,float,integer,number])
'$etype_conj_prim_coded'(542).%empty([atomic,float,integer,number],[atom])
'$etype_conj_prim_coded'(8718).%empty([atomic,float,integer],[atom,number])
'$etype_conj_prim_coded'(12806).%empty([atomic,float],[atom,integer,number])
'$etype_conj_prim_coded'(10762).%empty([atomic,integer],[atom,float,number])
'$etype_conj_prim_coded'(1564).%empty([float,integer,number],[atom,atomic])
'$etype_conj_prim_coded'(9740).%empty([float,integer],[atom,atomic,number])
'$etype_conj_prim_coded'(5652).%empty([float,number],[atom,atomic,integer])
'$etype_conj_prim_coded'(3608).%empty([integer,number],[atom,atomic,float])
'$etype_conj_prim_coded'(7696).%empty([number],[atom,atomic,float,integer])
'$etype_conj_prim_coded'(15).%empty([atom,atomic,float,integer],[])
'$etype_conj_prim_coded'(4103).%empty([atom,atomic,float],[integer])
'$etype_conj_prim_coded'(2059).%empty([atom,atomic,integer],[float])
'$etype_conj_prim_coded'(1037).%empty([atom,float,integer],[atomic])
'$etype_conj_prim_coded'(5125).%empty([atom,float],[atomic,integer])
'$etype_conj_prim_coded'(3081).%empty([atom,integer],[atomic,float])
'$etype_conj_prim_coded'(7169).%empty([atom],[atomic,float,integer])
'$etype_conj_prim_coded'(526).%empty([atomic,float,integer],[atom])
'$etype_conj_prim_coded'(1548).%empty([float,integer],[atom,atomic])
'$etype_conj_prim_coded'(5636).%empty([float],[atom,atomic,integer])
'$etype_conj_prim_coded'(3592).%empty([integer],[atom,atomic,float])
'$etype_conj_prim_coded'(23).%empty([atom,atomic,float,number],[])
'$etype_conj_prim_coded'(8199).%empty([atom,atomic,float],[number])
'$etype_conj_prim_coded'(2067).%empty([atom,atomic,number],[float])
'$etype_conj_prim_coded'(1045).%empty([atom,float,number],[atomic])
'$etype_conj_prim_coded'(9221).%empty([atom,float],[atomic,number])
'$etype_conj_prim_coded'(3089).%empty([atom,number],[atomic,float])
'$etype_conj_prim_coded'(11265).%empty([atom],[atomic,float,number])
'$etype_conj_prim_coded'(8710).%empty([atomic,float],[atom,number])
'$etype_conj_prim_coded'(1556).%empty([float,number],[atom,atomic])
'$etype_conj_prim_coded'(3600).%empty([number],[atom,atomic,float])
'$etype_conj_prim_coded'(7).%empty([atom,atomic,float],[])
'$etype_conj_prim_coded'(1029).%empty([atom,float],[atomic])
'$etype_conj_prim_coded'(3073).%empty([atom],[atomic,float])
'$etype_conj_prim_coded'(1540).%empty([float],[atom,atomic])
'$etype_conj_prim_coded'(27).%empty([atom,atomic,integer,number],[])
'$etype_conj_prim_coded'(8203).%empty([atom,atomic,integer],[number])
'$etype_conj_prim_coded'(4115).%empty([atom,atomic,number],[integer])
'$etype_conj_prim_coded'(1049).%empty([atom,integer,number],[atomic])
'$etype_conj_prim_coded'(9225).%empty([atom,integer],[atomic,number])
'$etype_conj_prim_coded'(5137).%empty([atom,number],[atomic,integer])
'$etype_conj_prim_coded'(13313).%empty([atom],[atomic,integer,number])
'$etype_conj_prim_coded'(8714).%empty([atomic,integer],[atom,number])
'$etype_conj_prim_coded'(1560).%empty([integer,number],[atom,atomic])
'$etype_conj_prim_coded'(5648).%empty([number],[atom,atomic,integer])
'$etype_conj_prim_coded'(11).%empty([atom,atomic,integer],[])
'$etype_conj_prim_coded'(1033).%empty([atom,integer],[atomic])
'$etype_conj_prim_coded'(5121).%empty([atom],[atomic,integer])
'$etype_conj_prim_coded'(1544).%empty([integer],[atom,atomic])
'$etype_conj_prim_coded'(19).%empty([atom,atomic,number],[])
'$etype_conj_prim_coded'(1041).%empty([atom,number],[atomic])
'$etype_conj_prim_coded'(9217).%empty([atom],[atomic,number])
'$etype_conj_prim_coded'(1552).%empty([number],[atom,atomic])
'$etype_conj_prim_coded'(1025).%empty([atom],[atomic])
'$etype_conj_prim_coded'(29).%empty([atom,float,integer,number],[])
'$etype_conj_prim_coded'(8205).%empty([atom,float,integer],[number])
'$etype_conj_prim_coded'(4117).%empty([atom,float,number],[integer])
'$etype_conj_prim_coded'(12293).%empty([atom,float],[integer,number])
'$etype_conj_prim_coded'(2073).%empty([atom,integer,number],[float])
'$etype_conj_prim_coded'(10249).%empty([atom,integer],[float,number])
'$etype_conj_prim_coded'(6161).%empty([atom,number],[float,integer])
'$etype_conj_prim_coded'(540).%empty([float,integer,number],[atom])
'$etype_conj_prim_coded'(8716).%empty([float,integer],[atom,number])
'$etype_conj_prim_coded'(12804).%empty([float],[atom,integer,number])
'$etype_conj_prim_coded'(10760).%empty([integer],[atom,float,number])
'$etype_conj_prim_coded'(13).%empty([atom,float,integer],[])
'$etype_conj_prim_coded'(4101).%empty([atom,float],[integer])
'$etype_conj_prim_coded'(2057).%empty([atom,integer],[float])
'$etype_conj_prim_coded'(524).%empty([float,integer],[atom])
'$etype_conj_prim_coded'(21).%empty([atom,float,number],[])
'$etype_conj_prim_coded'(8197).%empty([atom,float],[number])
'$etype_conj_prim_coded'(2065).%empty([atom,number],[float])
'$etype_conj_prim_coded'(8708).%empty([float],[atom,number])
'$etype_conj_prim_coded'(5).%empty([atom,float],[])
'$etype_conj_prim_coded'(25).%empty([atom,integer,number],[])
'$etype_conj_prim_coded'(8201).%empty([atom,integer],[number])
'$etype_conj_prim_coded'(4113).%empty([atom,number],[integer])
'$etype_conj_prim_coded'(8712).%empty([integer],[atom,number])
'$etype_conj_prim_coded'(9).%empty([atom,integer],[])
'$etype_conj_prim_coded'(17).%empty([atom,number],[])
'$etype_conj_prim_coded'(30).%empty([atomic,float,integer,number],[])
'$etype_conj_prim_coded'(8206).%empty([atomic,float,integer],[number])
'$etype_conj_prim_coded'(12294).%empty([atomic,float],[integer,number])
'$etype_conj_prim_coded'(10250).%empty([atomic,integer],[float,number])
'$etype_conj_prim_coded'(1052).%empty([float,integer,number],[atomic])
'$etype_conj_prim_coded'(9228).%empty([float,integer],[atomic,number])
'$etype_conj_prim_coded'(5140).%empty([float,number],[atomic,integer])
'$etype_conj_prim_coded'(3096).%empty([integer,number],[atomic,float])
'$etype_conj_prim_coded'(7184).%empty([number],[atomic,float,integer])
'$etype_conj_prim_coded'(14).%empty([atomic,float,integer],[])
'$etype_conj_prim_coded'(1036).%empty([float,integer],[atomic])
'$etype_conj_prim_coded'(5124).%empty([float],[atomic,integer])
'$etype_conj_prim_coded'(3080).%empty([integer],[atomic,float])
'$etype_conj_prim_coded'(8198).%empty([atomic,float],[number])
'$etype_conj_prim_coded'(1044).%empty([float,number],[atomic])
'$etype_conj_prim_coded'(3088).%empty([number],[atomic,float])
'$etype_conj_prim_coded'(1028).%empty([float],[atomic])
'$etype_conj_prim_coded'(8202).%empty([atomic,integer],[number])
'$etype_conj_prim_coded'(1048).%empty([integer,number],[atomic])
'$etype_conj_prim_coded'(5136).%empty([number],[atomic,integer])
'$etype_conj_prim_coded'(1032).%empty([integer],[atomic])
'$etype_conj_prim_coded'(1040).%empty([number],[atomic])
'$etype_conj_prim_coded'(28).%empty([float,integer,number],[])
'$etype_conj_prim_coded'(8204).%empty([float,integer],[number])
'$etype_conj_prim_coded'(12292).%empty([float],[integer,number])
'$etype_conj_prim_coded'(10248).%empty([integer],[float,number])
'$etype_conj_prim_coded'(12).%empty([float,integer],[])
'$etype_conj_prim_coded'(8196).%empty([float],[number])
'$etype_conj_prim_coded'(8200).%empty([integer],[number])

