/* 

An abstract substitution on a set of variables Vars is represented as
a set of variable-typings whose domain is a subset of Vars. A
variable-typing maps a variable not in its domain to the type
"top". The choice of a partial function instead of a total function is
non-essential.

*/

:- module(rtype, [ 
		rtype_eq_constraints/2,
                rtype_init/0,
		rtype_list_print/1,
		rtype_unify_entry/5,
		rtype_unify_exit/6,
		rtype_bipai/3,
		rtype_compile/2,
		rtype_bot/2,
		rtype_ident/2,
		rtype_join/3,
		rtype_leq/2,
		rtype_internalize/2,
		rtype_externalize/2,
		rtype_bw_compile/2,
		rtype_bip_has_cond/2,
		rtype_top/2,
		rtype_meet/3,
		rtype_bw_unify_entry/5,
		rtype_bw_unify_exit/6,
		rtype_bw_bipai/3
	       ]).


rtype_list_print(_).


rtype_unify_entry(Sub,Sigma,_NVarR,NVar,Zeta) :- 
           vts_tag(Sigma,NVar,SigmaR), 
	   rtype_solve(Sub,SigmaR,SigmaRSub),
           vts_restrict(SigmaRSub,NVar,Zeta).


    vts_tag([],_Shift,[]).
    vts_tag([VTi|VTSi],Shift,[VTo|VTSo]) :-
            vt_tag(VTi,Shift,VTo),
            vts_tag(VTSi,Shift,VTSo).

    vt_tag([],_Shift,[]).
    vt_tag(['$VAR'(Ni)/Ti|VTi],Shift,['$VAR'(No)/Ti|VTo]) :-
	    No is Ni + Shift,
	    vt_tag(VTi,Shift,VTo).


    vts_restrict([],_NVar,[]).
    vts_restrict([VTi|VTSi],NVar,[VTo|VTSo]) :-
            vt_restrict(VTi,NVar,VTo),
            vts_restrict(VTSi,NVar,VTSo).

    vt_restrict([],_NVar,[]).
    vt_restrict(['$VAR'(Ni)/Ti|VTi],NVar,VTo) :-
	    ( Ni =< NVar -> VTo = ['$VAR'(Ni)/Ti|VT] ; VTo=VT ), 
	    vt_restrict(VTi,NVar,VT).


rtype_unify_exit(Sub,Sigma,_NVarR,Theta,NVar,Zeta) :-
           vts_tag(Sigma,NVar,SigmaR), 
           rtype_comb(SigmaR,Theta,SigmaRTheta),
	   rtype_solve(Sub,SigmaRTheta,SigmaRThetaSub),
           vts_restrict(SigmaRThetaSub,NVar,Zeta).


rtype_meet(Zeta,Eta,Beta) :-  
    vtsglb(Zeta,Eta,Beta1), 
    reduce(Beta1,Beta).

rtype_bw_unify_exit(CE,Answer,_NVarR,DemandRt,NVar,DemandLt) :-
    vts_tag(Answer,NVar,AnswerR),
    rtype_solve(CE,AnswerR,CEAnswerR), 
    vts_restrict(CEAnswerR,NVar,CEAnswerR1),
    pseduo_complement(CEAnswerR1,DemandRt,DemandLt).


/*
rtype_bw_unify_exit(CE,Answer,_NVarR,DemandRt,NVar,DemandLt) :-
	  vts_tag(Answer,NVar,AnswerR),
          vts_tag(DemandRt,NVar,DemandRtR),
          pseduo_complement(AnswerR,DemandRtR,DemandAfterCE),
          rtype_solve_bw(CE,DemandAfterCE,DemandLtUnRestricted), 
          vts_proj_forall(DemandLtUnRestricted,NVar,DemandLt).
*/


rtype_bw_unify_entry(CE,Sigma,_NVarR,NVar,Eta) :-
    vts_tag(Sigma,NVar,SigmaR),
    rtype_solve_bw(CE,SigmaR,Zeta),
    vts_proj_forall(Zeta,NVar,Eta).



/*

:- use_module(ordsets,	[
	ord_add_element/3,
	ord_intersection/3,
        ord_subtract/3,
	ord_subset/2,
	ord_setproduct/3,
	ord_union/3
   ]).

*/

:- use_module(basics, [
	mgu/3,
        variable/1,
	vars/2
   ]).


:- use_module(typedef, [
	typedef_constructors/1,
	typedef_functors/1,
	typedef/3
   ]).
     



/* rtype_eq_constraints(+Sub,-Eqs) - do nothing */

rtype_eq_constraints(Sub,Sub).
rtype_internalize(ASub,ASub).
rtype_externalize(ASub,ASub).





/* 

The combination of two abstract substitution is formed by conjucting
two variable-typings each from one abstract substitution in every
possibility.

*/

rtype_comb(VTS1,VTS2,VTSo) :-
	rtype_comb(VTS1,VTS2,[],VTSo).

rtype_comb([],_,VTSo,VTSo).

rtype_comb([VT1|VTS1],VTS2,VTSo1,VTSo)	:-
	rtype_comb_1(VTS2,VT1,VTSo1,VTSo2),
	rtype_comb(VTS1,VTS2,VTSo2,VTSo).

rtype_comb_1([],_,VTS,VTS).
rtype_comb_1([VT2|VTS2],VT1,VTS0,VTS)	:-
	merge(VT1,VT2,VT12),
	rtype_comb_1(VTS2,VT1, [VT12|VTS0],VTS).


/*

The identity abstract substitution on Vars constains a single variable
typing that maps every variable into "top".


rtype_ident(Vs,ASubId) :- 
    rtype_top(Vs,ASubId).

*/

rtype_ident(_Vars,[[]]). 

/* 

The top abstract substitution on Vars constains a single variable
typing that maps every variable into "top".


rtype_top(Vs,[VtTop]) :-
    vt_top(Vs,VtTop).

vt_top([],[]).
vt_top([V|Vs],[V/top|VtTop]) :- 
    vt_top(Vs,VtTop). 

*/

rtype_top(_Vars,[[]]).



/* 

The bottom abstract substitution on Vars is the empty set of variable
typings. The following implementation also tests if a given set of
variable typings is equivalent to the bottom abstract
substitution. This is necesary due to redundency in an abstract
substitution.

*/

rtype_bot(_Vars, []) :- !.
rtype_bot(_Vars, ASub) :- reduce(ASub,[]).


/* 

rtype_ext(+ASubIn,+Vars,-ASubOut) ASubOut=ASubIn since default type
assigned to a variable by a variable-typing is "top"


rtype_ext([],_Vs,[]).
rtype_ext([VTi|VTSi],Vs,[VTo|VTSo]) :-
        vt_ext(Vs,VTi,VTo),
        rtype_ext(VTSi,Vs,VTSo).

vt_ext([],VT,VT) :-!.
vt_ext(Vs,[],VT) :-
        !,
        vt_top(Vs,VT).
vt_ext([V|Vs],[Vi/Ti|VTi],[Vo/To|VTo]) :-
        (
            V@<Vi
        ->
            Vo=V, To=top,
            vt_ext(Vs,[Vi/Ti|VTi],VTo)
	 ;
            V@>Vi
        ->  
            Vo=Vi,To=Ti,
            vt_ext([V|Vs],VTi,VTo)
         ; 
            Vo=Vi,To=Ti,
            vt_ext(Vs,VTi,VTo)         
        ).

*/


rtype_ext(VTS,_,VTS). 


/* 

rtype_solve(+Sub, +ASubi, -ASubo) -ASubo is an abstract substitution
on the same domain as +ASubi.-ASubo approximates the set of
substitutions that satisfy both constraints +Sub and +ASubi.

*/

rtype_solve(EQS,ASubi,ASubo) :-
	rtype_down(EQS,ASubi,ASubd),
	rtype_up(EQS,ASubd,ASubu),
        reduce(ASubu,ASubo).

rtype_down([],ASub,ASub).
rtype_down([LHS/RHS|EQS],ASubi,ASubo) :- 
        rtype_down_1(ASubi,LHS,RHS,ASub1),
        rtype_down(EQS,ASub1,ASubo). 
        

rtype_down_1([],_,_,[]).
rtype_down_1([VTi|VTSi],LHS,RHS,VTSo) :-
        ( 
            mapsto(LHS,VTi,Tp), 
            \+ Tp = top
        -> 
            vts_UpApprox(Tp,RHS,VTS11),
            vtsglb([VTi],VTS11,VTS1)
	 ;
            VTS1=[VTi]
        ),
        rtype_down_1(VTSi,LHS,RHS,VTS2),
        append(VTS1,VTS2,VTSo).


rtype_up([],ASub,ASub).
rtype_up([LHS/RHS|EQS],ASubi,ASubo) :- 
        rtype_up_1(ASubi,LHS,RHS,ASub1),
        rtype_up(EQS,ASub1,ASubo). 

rtype_up_1([],_,_,[]).
rtype_up_1([VTi|VTSi],LHS,RHS,[VTo|VTSo]) :-
        type_UpApprox(RHS,VTi,Type),
        ( Type=top -> VTo=VTi ; vt_strengthen(VTi,LHS,Type,VTo) ), % CORRECTED
        rtype_up_1(VTSi,LHS,RHS,VTSo).



/* rtype_bipai(+Call,+ASubi,-ASubo) ASubo describes the set of
   substitutions that satisfy ASubi and that are answers to Call.
*/

rtype_bipai(copy,ASubi,ASubi). 
rtype_bipai(solve(E0),ASubi,ASubo) :- 	rtype_solve(E0,ASubi,ASubo).
rtype_bipai(glb(VTS),ASubi,ASubo)  :-   vtsglb(VTS,ASubi,ASubo). 
rtype_bipai(glb_solve(VTS,E0),ASubi,ASubo)  :-   
           vtsglb(VTS,ASubi,ASub1),rtype_solve(E0,ASub1,ASubo). 
rtype_bipai(var(V),ASubi,ASubo)    :- 
           rtype_bipai_var(ASubi,V,ASubo).
           % keep those variable typings in which V is of type "top" 

    rtype_bipai_var([],_,[]).
    rtype_bipai_var([VTi|VTSi],V,VTSo) :- 
        type_UpApprox(V,VTi,Tp),
        (
            etype_tabled(not(Tp)) % Tp is equivalent to top ?
        ->
            rtype_bipai_var(VTSi,V,VTSo)
	 ; 
            VTSo=[VTi|VTSo1],
            rtype_bipai_var(VTSi,V,VTSo1)
        ). 


/* rtype_compile(+Call,-Compiled) compiles a bip call to its compiled
   form to factor out computation that is independent of abstract
   input state. */

rtype_compile(abolish(_),       copy).
rtype_compile(assert(_),        copy).
rtype_compile(retract(_),       copy).
rtype_compile(erase(I), glb(VTS)) :- 
    vts_UpApprox(integer,I,VTS).
rtype_compile(recorda(_Key,_Term,Ref), glb(VTS))    :-
    vts_UpApprox(integer,Ref,VTS).
rtype_compile(recorded(_Key,_Term,Ref), glb(VTS))    :-
    vts_UpApprox(integer,Ref,VTS).
rtype_compile(format(Fmt), glb(VTS))      :-
    vts_UpApprox(or(atom, or(string,list(integer))),Fmt,VTS).
rtype_compile(format(Fmt,_), glb(VTS))    :-
    vts_UpApprox(or(atom, or(string,list(integer))),Fmt,VTS).
rtype_compile(format(_,Fmt,_), glb(VTS))  :-
    vts_UpApprox(or(atom, or(string,list(integer))),Fmt,VTS).


rtype_compile(repeat,           copy).
rtype_compile(listing,          copy).
rtype_compile(listing(_),       copy).
rtype_compile(compound(_),      copy).
rtype_compile(portray_clause(_),copy).
rtype_compile(ground(_),        copy).
rtype_compile(!,	        copy).
rtype_compile(display(_),	copy).
rtype_compile(put(_),           copy).
rtype_compile(nl,               copy).
rtype_compile(print(_),         copy).
rtype_compile(write(_),         copy).
rtype_compile(writeq(_),	copy).
rtype_compile(keysort(_,_),     copy). % sorting on term order
rtype_compile(sort(_,_),	copy).
rtype_compile(@<(_,_),	        copy). % comparing on term order
rtype_compile(@>(_,_),	        copy).
rtype_compile(@=<(_,_),	        copy).
rtype_compile(@>=(_,_),         copy).
rtype_compile(true,             copy).
rtype_compile(nonvar(_),	copy).
rtype_compile(read(_),	        copy).
rtype_compile(\==(_,_),         copy).
rtype_compile(\=(_,_),          copy).
rtype_compile(=\=(_,_),	        copy).
rtype_compile(statistics(_N1,_N2), copy).
rtype_compile(compare(_,_,_), copy).


% these built-ins change type information as fail

%rtype_compile(abort,            _) :- fail.
%rtype_compile(fail,             _) :- fail.
%rtype_compile(false,	         _) :- fail.

% these built-ins change type information as unification

rtype_compile(=(Left,Right),    solve(E0))     :-
	mgu(Left,Right,E0).
rtype_compile(=@=(Left,Right),  solve(E0))     :-
	mgu(Left,Right,E0).
        % structurally identical terms must unify
rtype_compile(==(Left,Right),   solve(E0))     :-
	mgu(Left,Right,E0).
	% identical terms must unify
rtype_compile(is(Left,Right),   solve(E0))     :-
	mgu(Left,Right,E0).
	% identical terms must unify

% these built-ins type check their arguments

rtype_compile(float(Tm), glb(VTS))     :-
        vts_UpApprox(float,Tm,VTS).
rtype_compile(string(Tm), glb(VTS))    :-
        vts_UpApprox(string,Tm,VTS).
rtype_compile(atom(Tm),  glb(VTS))     :-
        vts_UpApprox(atom,Tm,VTS).
rtype_compile(atomic(Tm), glb(VTS))    :-
        vts_UpApprox(atomic,Tm,VTS).
rtype_compile(integer(Tm), glb(VTS))   :-
        vts_UpApprox(integer,Tm,VTS).
rtype_compile(number(Tm), glb(VTS))    :-
        vts_UpApprox(number,Tm,VTS).
rtype_compile(tab(Tm), glb(VTS))       :-
        vts_UpApprox(integer,Tm,VTS).
rtype_compile(<(N1,N2), glb(VTS))      :-
        vts_UpApprox(number,N1,VTS1),
        vts_UpApprox(number,N2,VTS2),
        vtsglb(VTS1,VTS2,VTS). 
rtype_compile(>(N1,N2), glb(VTS))     :-
        vts_UpApprox(number,N1,VTS1),
        vts_UpApprox(number,N2,VTS2),
        vtsglb(VTS1,VTS2,VTS). 
rtype_compile(=<(N1,N2), glb(VTS))     :-
        vts_UpApprox(number,N1,VTS1),
        vts_UpApprox(number,N2,VTS2),
        vtsglb(VTS1,VTS2,VTS). 
rtype_compile(>=(N1,N2), glb(VTS))     :-
        vts_UpApprox(number,N1,VTS1),
        vts_UpApprox(number,N2,VTS2),
        vtsglb(VTS1,VTS2,VTS). 
rtype_compile(length(L,N),  glb(VTS))     :-
        vts_UpApprox(integer,N,VTS1),
        vts_UpApprox(list(top),L,VTS2),
        vtsglb(VTS1,VTS2,VTS). 
rtype_compile(=..(_,L), glb(VTS)) :-
        vts_UpApprox(list(top),L,VTS).
rtype_compile(functor(_,Nm,Ar),  glb(VTS))     :-
        vts_UpApprox(integer,Ar,VTS1),
        vts_UpApprox(atom,Nm,VTS2),
        vtsglb(VTS1,VTS2,VTS). 

rtype_compile(arg(N,_,_), glb(VTS))     :- % may be made more precise??
        vts_UpApprox(integer,N,VTS).
rtype_compile(name(AI,S), glb(VTS))     :-
        vts_UpApprox(or(atom,integer),AI,VTS1),
        vts_UpApprox(string,S,VTS2),
        vtsglb(VTS1,VTS2,VTS). 

rtype_compile(=:=(N1,N2), glb_solve(VTS,E0))     :- % ?? two sides must unify
        vts_UpApprox(number,N1,VTS1),
        vts_UpApprox(number,N2,VTS2),
        vtsglb(VTS1,VTS2,VTS),
        mgu(N1,N2,E0).

rtype_compile(var(V), var(V)) :- !, variable(V).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

/* rtype_bip_has_cond(+Call,-Cond) is true iff +Call has type
   condition -Cond.  */

%rtype_bip_has_cond(assert(_),        _) :- fail.
%rtype_bip_has_cond(retract(_),       _) :- fail.
%rtype_bip_has_cond(repeat,           _) :- fail.
%rtype_bip_has_cond(listing,          _) :- fail.
%rtype_bip_has_cond(listing(_),       _) :- fail.
%rtype_bip_has_cond(compound(_),      _) :- fail.
%rtype_bip_has_cond(ground(_),        _) :- fail.
%rtype_bip_has_cond(!,	        _) :- fail.
%rtype_bip_has_cond(display(_),	_) :- fail.
%rtype_bip_has_cond(nl,               _) :- fail.
%rtype_bip_has_cond(print(_),         _) :- fail.
%rtype_bip_has_cond(write(_),         _) :- fail.
%rtype_bip_has_cond(writeq(_),	_) :- fail.
%rtype_bip_has_cond(abort,            _) :- fail.
%rtype_bip_has_cond(fail,             _) :- fail.
%rtype_bip_has_cond(false,            _) :- fail.
%rtype_bip_has_cond(@<(_,_),	      _) :- fail. 
%rtype_bip_has_cond(@>(_,_),	        _) :- fail.
%rtype_bip_has_cond(@=<(_,_),	        _) :- fail.
%rtype_bip_has_cond(@>=(_,_),         _) :- fail.
%rtype_bip_has_cond(true,             _) :- fail.
%rtype_bip_has_cond(nonvar(_),	_) :- fail.
%rtype_bip_has_cond(read(_),	_) :- fail.
%rtype_bip_has_cond(\==(_,_),   _) :- fail.
%rtype_bip_has_cond(\=(_,_),    _) :- fail.
%rtype_bip_has_cond(=\=(_,_),	_) :- fail.
%rtype_bip_has_cond(var(_), _)     :- fail.
%rtype_bip_has_cond(=(_,_), _)     :- fail.
%rtype_bip_has_cond(=@=(_,_), _)   :- fail.
%rtype_bip_has_cond(==(_,_),_)     :- fail.
%rtype_bip_has_cond(is(_,_), _)    :- fail.
%rtype_bip_has_cond(float(_), _)   :- fail.
%rtype_bip_has_cond(string(_),_)   :- fail.
%rtype_bip_has_cond(atom(_),_)     :- fail.
%rtype_bip_has_cond(atomic(_),_)   :- fail.
%rtype_bip_has_cond(integer(_),_)  :- fail.
%rtype_bip_has_cond(number(_),_)   :- fail.
%rtype_bip_has_cond(length(_,_),_)  :- fail.
%rtype_bip_has_cond(=..(_,_),_)     :- fail.
%rtype_bip_has_cond(portray_clause(_),_) :- fail.
%rtype_bip_has_cond(compare(_,_,_), _)   :- fail.


rtype_bip_has_cond(erase(I),VTS)  :- 
    vts_LowApprox(integer,I,VTS).
rtype_bip_has_cond(format(Fmt), VTS)      :-
    vts_LowApprox(or(atom, or(string,list(integer))),Fmt,VTS).
rtype_bip_has_cond(format(Fmt,_), VTS)    :-
    vts_LowApprox(or(atom, or(string,list(integer))),Fmt,VTS).
rtype_bip_has_cond(format(_,Fmt,_), VTS)  :-
    vts_LowApprox(or(atom, or(string,list(integer))),Fmt,VTS).
rtype_bip_has_cond(keysort(Li,_), VTS) :- 
    vts_LowApprox(list(pair(top,top)),Li,VTS).
rtype_bip_has_cond(sort(Li,_), VTS)    :- 
    vts_LowApprox(list(top),Li,VTS).

% following conditions can be relaxed in a more powerful type system.

rtype_bip_has_cond(put(N),VTS) :- vts_LowApprox(integer,N,VTS).
rtype_bip_has_cond(statistics(Key,_), VTS) :- 
    vts_LowApprox(atom,Key,VTS). % needs be refined

rtype_bip_has_cond(functor(_,_,_),_) :-
    write("The program contains functor/3!"),nl,halt.
rtype_bip_has_cond(arg(_,_,_),_) :- 
    write("The program contains arg/3!"),nl,halt.
rtype_bip_has_cond(abolish(_),_) :- 
     write("The program contains abolish/1!"),nl,halt.

rtype_bip_has_cond(recorda(_,_,_),_)    :-
     write("The program contains recorda/3!"),nl,halt.
rtype_bip_has_cond(recorded(_,_,_),_)   :-
    write("The program contains recorded/3!"),nl,halt.



% these built-ins type check their arguments

rtype_bip_has_cond(tab(Tm), VTS)       :-
        vts_LowApprox(integer,Tm,VTS).
rtype_bip_has_cond(<(N1,N2), VTS)      :-
        vts_LowApprox(number,N1,VTS1),
        vts_LowApprox(number,N2,VTS2),
        vtsglb(VTS1,VTS2,VTS). 
rtype_bip_has_cond(>(N1,N2), VTS)      :-
        vts_LowApprox(number,N1,VTS1),
        vts_LowApprox(number,N2,VTS2),
        vtsglb(VTS1,VTS2,VTS). 
rtype_bip_has_cond(=<(N1,N2), VTS)     :-
        vts_LowApprox(number,N1,VTS1),
        vts_LowApprox(number,N2,VTS2),
        vtsglb(VTS1,VTS2,VTS). 
rtype_bip_has_cond(>=(N1,N2), VTS)     :-
        vts_LowApprox(number,N1,VTS1),
        vts_LowApprox(number,N2,VTS2),
        vtsglb(VTS1,VTS2,VTS). 
rtype_bip_has_cond(=:=(N1,N2),VTS)     :- 
        vts_LowApprox(number,N1,VTS1),
        vts_LowApprox(number,N2,VTS2),
        vtsglb(VTS1,VTS2,VTS).
rtype_bip_has_cond(name(AI,S), VTS)    :-
        vts_LowApprox(or(atom,integer),AI,VTS1),
        vts_LowApprox(string,S,VTS2),
        append(VTS1,VTS2,VTS). 










%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
/*

% rtype_leq(+VTS1,+VTS2) VTS1 is subsumed by VTS2

rtype_leq(VTS1,VTS2) :- 
    vtSet_to_seqSet(VTS1,SeqS1),
    vtSet_to_seqSet(VTS2,SeqS2),
    !,
    seqSet_leq(SeqS1,SeqS2).

    vtSet_to_seqSet([],[]).
    vtSet_to_seqSet([VT|VTS],[Seq|Seqs]) :- 
        vt_to_seq(VT,Seq),
        vtSet_to_seqSet(VTS,Seqs).

    vt_to_seq([],[]).
    vt_to_seq([_V/T|VT],[T|Seq]) :- 
        vt_to_seq(VT,Seq).
*/


% rtype_leq(+VTS1,+VTS2) VTS1 is subsumed by VTS2

rtype_leq([],_).
rtype_leq([VT1|VTS1],VTS2) :- 
    ( vt_leq_one(VT1,VTS2); vt_bot(VT1) ),
    !,
    rtype_leq(VTS1,VTS2).

vt_bot([_/T|VT]) :- 
    ( etype_tabled(T) -> true; vt_bot(VT) ).


% rtype_join(+VTS1,+VTS2,-VTS) VTS=VTS1+VTS2 redundency removed.

rtype_join(VTS1, VTS2, VTSo) :-
        append(VTS1, VTS2, VTS12),
        !,
        reduce(VTS12,VTSo).

reduce(VTSi,VTSo) :- 
    sort(VTSi,VTSiSorted), % removing duplicates
    reduce(VTSiSorted,[],VTSo). 

reduce([],VTS,VTS). 
reduce([VTi|VTSi],VTSacc,VTSo) :- 
    (
        ( vt_bot(VTi);     
	  vt_leq_one(VTi,VTSi);
	  vt_leq_one(VTi,VTSacc) 
	)
    ->
        reduce(VTSi,VTSacc,VTSo)
     ; 
        reduce(VTSi,[VTi|VTSacc],VTSo)
    ).

vt_leq_one(VT1,[VT2|VTS2]) :- 
    ( vt_leq(VT1,VT2) -> true ; vt_leq_one(VT1,VTS2) ). 

vt_leq([],[])    :- !.
vt_leq([],[H|L]) :- vt_id([H|L]).  %%%
vt_leq([_|_],[]) :- !.
vt_leq([P1/T1|VT1],[P2/T2|VT2]) :-
	(
		P1 @< P2 
	->      % skip P1 as VT2(P1) = top
		vt_leq(VT1,[P2/T2|VT2])
	;
		P1 @> P2
	->      % 
                etype_tabled(not(T2)), % T2 must be equivalent to "top"
		vt_leq([P1/T1|VT1],VT2)
	;
                % test T1<=T2
                etype_tabled(and(T1,not(T2))),
		vt_leq(VT1,VT2)
	).

vt_id([]).
vt_id([_/T|VT]) :- 
    etype_tabled(not(T)),
    vt_id(VT). 

/*
%  specifiaction 

reduce([],[]). 
reduce([VTi|VTSi],VTSo) :- 
    domain(VTi,Vs),
    vtSet_to_seqSet([VTi|VTSi],SeqSi),
    seqSet_reduce(SeqSi,SeqSo),
    seqSet_varSet_vtSet(SeqSo,Vs,VTSo).

    domain([],[]).
    domain([V/_|VT],[V|Vs]) :- 
        domain(VT,Vs). 

    seqSet_varSet_vtSet([],_,[]).
    seqSet_varSet_vtSet([Seq|SeqS],Vs,[VT|VTS]) :- 
        varSet_Seq_to_vt(Vs,Seq,VT),
        seqSet_varSet_vtSet(SeqS,Vs,VTS).

    varSet_Seq_to_vt([],[],[]).
    varSet_Seq_to_vt([V|Vs],[T|Seq],[V/T|VT]) :- 
        varSet_Seq_to_vt(Vs,Seq,VT).
*/
    

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


/* vt_strengthen(+VTi,+V,+T,-VTo) VTo = VT1[Var/(VTi(V) and T)] */

vt_strengthen([],V,T,[V/T]).
vt_strengthen([Key1/Type1|VT1],Key2,Type2,[Key3/Type3|VT3]) :-
	(
		Key1 @< Key2
	->
		Key3 = Key1,Type3=Type1,
		vt_strengthen(VT1,Key2,Type2,VT3)
	;
		Key1 @> Key2
	->
		Key3 = Key2,Type3=Type2,VT3=[Key1/Type1|VT1]		
	;
		Key3 = Key2, tand(Type1,Type2,Type3),
                VT3=VT1
	).


make_map([],[],[]).
make_map([Key|Keys],[Val|Vals],[Key/Val|Map]) :- 
    make_map(Keys,Vals,Map).


:- [empty].


type_UpApprox(V,VT,Tp) :- 
    variable(V),
    !,
    ( mapsto(V,VT,Tp1) -> type_dnf(Tp1,Tp) ; Tp=top ).
type_UpApprox(Tm,VT,Tp) :- 
    functor(Tm,Fn,Ar),
    (
        Ar =:= 0
    ->
        type_UpApprox_one_level(Fn/Ar,[],Tp)
     ;  
        Tm =.. [_|Args],
        type_UpApprox_map(Args,VT,TArgs),
%        findall(CT,(one_mix(TArgs,CTArgs),
%		    type_UpApprox_one_level(Fn/Ar,CTArgs,CT)),CTs),
%        bigtlub(CTs,Tp)
	 type_UpApprox_one_level(Fn/Ar,TArgs,Tp)
     ).

one_mix([],[]).
one_mix([or(C,_)|ORs],[C|Cs]) :- 
    one_mix(ORs,Cs).
one_mix([or(_,R)|ORs],[C|Cs]) :- 
    one_mix([R|ORs],[C|Cs]).
one_mix([C|ORs],[C|Cs]) :- 
    \+ C=or(_,_),
    one_mix(ORs,Cs).
    

type_UpApprox_map([],_,[]).
type_UpApprox_map([Tm|Tms],VT,[Tp|Tps]) :- 
    type_UpApprox(Tm,VT,Tp),
    type_UpApprox_map(Tms,VT,Tps).

type_UpApprox_one_level(Fn/Ar,CTs,CT) :- 
    findall(AT,type_UpApprox_one_level_one_rule(Fn/Ar,CTs,AT),ATs),
    bigtglb(ATs,CT).

type_UpApprox_one_level_one_rule(Fn/0,[],AT)   :- % non-deterministic 
    prim_type(AT),  % extra clause for constant and types 
    membership_prim(AT,Fn). 

type_UpApprox_one_level_one_rule(Fn/Ar,CTs,AT) :- % non-deterministic 
    typedef(Fn/Ar,Tis,To),
    tvs_UpApprox_vec(CTs,Tis,[[]],TVS), % [] maps every parameter to bot
    tvs_apply(TVS,To,AT).


bigtglb(Tps,Tp) :- 
   bigtglb(Tps,top,Tp).

bigtglb([],Tp,Tp).
bigtglb([A|As],Tp0,Tp) :-
    tand(A,Tp0,Tp1),
    bigtglb(As,Tp1,Tp).

bigtlub(Tps,Tp) :- 
   bigtlub(Tps,bot,Tp).

bigtlub([],Tp,Tp).
bigtlub([A|As],Tp0,Tp) :-
    tor(A,Tp0,Tp1),
    bigtlub(As,Tp1,Tp).



/* tvs_UpApprox_vec(+RTs,+Tis,+TVSi,-TVSo) TVSo = TVi \/ K such that
   RTs <= K(Tis) */

tvs_UpApprox_vec([],[],TVSi,TVSi). 
tvs_UpApprox_vec([R|Rs],[P|Ps],TVSi,TVSo) :-
    type_dnf(R,DC), % make sure tvs_UpApprox is called with first arg in dnf
    tvs_UpApprox(DC,P,TVS1),
    tvs_lub(TVS1,TVSi,TVS2),
    tvs_UpApprox_vec(Rs,Ps,TVS2,TVSo).


/* tvs_UpApprox(+RT,+Poly,-TVS) TVS(Poly)>=RT specified in the paper*/


tvs_UpApprox(or(A,C),P,TVSo) :- 
    !,
    tvs_UpApprox(A,P,TVSa),
    tvs_UpApprox(C,P,TVSc),
    append(TVSa,TVSc,TVSo). 
tvs_UpApprox(and(A,C),P,TVSo) :- 
    !,
    tvs_UpApprox(A,P,TVSa),
    tvs_UpApprox(C,P,TVSc),
    tvs_glb(TVSa,TVSc,TVSo). 
tvs_UpApprox(A,P,[[P/A]]) :-
    var(P), %P is a parameter in a type rule
    !.
tvs_UpApprox(not(A),P,TVSo) :- 
    !,
    negate_atom_UpApprox(A,DC),
    tvs_UpApprox(DC,P,TVSo).
tvs_UpApprox(top,_,[tv_top]) :-!.
tvs_UpApprox(bot,_,[tv_bot]) :-!.

tvs_UpApprox(A,P,TVS) :- 
    (
        A =..[Cons|Rs],P =..[Cons|Ps]
    ->  
        make_map(Ps,Rs,TVal), TVS=[TVal]
     ;
        TVS = [tv_top]
    ).


 


/* tv_lub(+TVal1,+TVal2,-TValo). Note that the default value for any
   paramter outside the domain of a type valustion is bot. */

tv_lub(tv_top,	_,	tv_top)	:-
	!.
tv_lub(_,	tv_top,	tv_top)	:-
	!.
tv_lub(tv_bot,	TVal,	TVal)	:-
	!.
tv_lub(TVal,	tv_bot,	TVal)	:-
	!.
tv_lub(TVal1,	TVal2,	TValo)	:-
	tv_lub_1(TVal1,TVal2,TValo).

tv_lub_1([],[],[]).
tv_lub_1([],[P/T|TVal],[P/T|TVal]).
tv_lub_1([P/T|TVal],[],[P/T|TVal]).
tv_lub_1([P1/T1|TVal1],[P2/T2|TVal2],[P/T|TVal]) :-
	(
		P1 @< P2
	->
		P=P1,
		T=T1,
		tv_lub_1(TVal1,[P2/T2|TVal2],TVal)
	;
		P1 @> P2
	->
		P=P2,
		T=T2,
		tv_lub_1([P1/T1|TVal1],TVal2,TVal)
	;
		P=P1, 
		tor(T1,T2,T),
		tv_lub_1(TVal1,TVal2,TVal)
	).

/* 

tv_glb(+TVal1,+TVal2,-TValo) Note that the default value for a type
parameter outside the domain of a type valuation/substitution is bot

*/

tv_glb(tv_top, K, K)	:-
	!.
tv_glb(K, tv_top, K)	:-
	!.
tv_glb(tv_bot, _, tv_bot)	:-
	!.
tv_glb(_, tv_bot, tv_bot)	:-
	!.
tv_glb(TVal1,TVal2,TValo)	:-
	tv_glb_1(TVal1,TVal2,TValo).

tv_glb_1([],_,[]) :- !.
tv_glb_1(_,[],[]) :- !.
tv_glb_1([P1/T1|TVal1],[P2/T2|TVal2],TVal) :-
	(
		P1 @< P2 
	->      % skip P1
		tv_glb_1(TVal1,[P2/T2|TVal2],TVal)
	;
		P1 @> P2
	->      % skip P2
		tv_glb_1([P1/T1|TVal1],TVal2,TVal)
	;
                % narrow the type assigned to P1==P2
                tand(T1,T2,T3),
                TVal=[P1/T3|TVal12],
		tv_glb_1(TVal1,TVal2,TVal12)
	).

/* tv_apply(+TSub,+TpIn,-Tp) TpO=TSub(TpIn). It unravals the input
   type TpIn in the top-down manner and constructs the output type TpO
   in the bottom-up manner. */

tv_apply(tv_top,_,top) :-  % special rule for tv_top
    !.
tv_apply(tv_bot,_,bot) :-  % special rule for tv_bot
    !.

tv_apply(TSub,TpIn,TpO) :-
	(    
		var(TpIn) %parameter(TpIn)
	->
		( 
			mapsto(TpIn,TSub,TpO) 
		-> 
			true
		; 
			TpO=bot % note default value for any
				    % Tp parameter is bot
		)
	; 
		TpIn =.. [F|ArgsIn],
		tv_apply_all(ArgsIn,TSub,ArgsO),
		TpO =.. [F|ArgsO]         
	).

tv_apply_all([],_,[]).
tv_apply_all([ArgIn|ArgsIn],TSub,[ArgO|ArgsO]) :-
	tv_apply(TSub,ArgIn,ArgO),
	tv_apply_all(ArgsIn,TSub,ArgsO).


/* mapsto(+Key,+Map,-Value) If +Key is in the domain of +Map then
   mapsto(Key,Map,Value) succeeds and Value is the value assigned to
   Key by Map. Otherwise mapsto(Key,Map,Value) fails. */

mapsto(Key,[Key1/Value1|Map],Value) :-
	( Key==Key1 -> Value=Value1 ; mapsto(Key,Map,Value) ).

/* vtglb(+VT1,+VT2,-VT3) +VT1 and +VT2 are variable-typings. vtglb/3
   succeeds with dom(VT3)=dom(VT1)+dom(VT2) and for each varaible X in
   dom(VT3) such that X/T is in VT1 and X/T2 in VT2, X/T3 is in VT3
   and T3 is the greatest lower bound of T1 and T2.

   This implementation assumes that variable typings are sorted lists
   that are sorted according to the ordering on the names of
   variables.*/


vtglb([],VT,VT) 	:-
        !.
vtglb(VT,[],VT) 	:-
        !.
vtglb([X1/T1|VT1],[X2/T2|VT2],[X1/T3|VT3]) 	:-
    	X1 == X2,
    	!,
    	tand(T1,T2,T3),
    	vtglb(VT1,VT2,VT3).
vtglb([X1/T1|VT1],[X2/T2|VT2],[X1/T1|VT3]) 	:-
    	X1 @< X2,
    	!,
    	vtglb(VT1,[X2/T2|VT2],VT3).
vtglb([X1/T1|VT1],[X2/T2|VT2],[X2/T2|VT3]) 	:-
    	X1 @> X2,
    	!,
    	vtglb([X1/T1|VT1],VT2,VT3).



/* The lower bound of two sets of variable typings. This is different
   from the combination of two abstract substitution because the
   former works on two sets of variable typings on the overlapping
   domains while the later on disjoint domains. */

vtsglb(VTS1,VTS2,VTSo) :-
	vtsglb(VTS1,VTS2,[],VTSo).

vtsglb([],_,VTSo,VTSo).

vtsglb([VT1|VTS1],VTS2,VTSo1,VTSo)	:-
	vtsglb_1(VTS2,VT1,VTSo1,VTSo2),
        reduce(VTSo2,VTSo3),
	vtsglb(VTS1,VTS2,VTSo3,VTSo).

vtsglb_1([],_,VTS,VTS).
vtsglb_1([VT2|VTS2],VT1,VTS0,VTS)	:-
	vtglb(VT1,VT2,VT12),
	vtsglb_1(VTS2,VT1, [VT12|VTS0],VTS).


/* vts_UpApprox(+R,+Tm,-VTS) VTS describes the set of all substitutions
   \theta such that \theta(Tm) is of type R. 


vts_UpApprox(and(R1,R2),Tm,VTS) :- 
    !,
    vts_UpApprox(R1,Tm,VTS1),
    vts_UpApprox(R2,Tm,VTS2),
    vtsglb(VTS1,VTS2,VTS).

vts_UpApprox(or(R1,R2),Tm,VTS) :- 
    !,
    vts_UpApprox(R1,Tm,VTS1),
    vts_UpApprox(R2,Tm,VTS2),
    append(VTS1,VTS2,VTS).

vts_UpApprox(R,V,[[V/R]]) :-
    variable(V),
    !.

vts_UpApprox(top,_,[[]]) :- 
    !.

vts_UpApprox(ATp,Tm,VTS) :- 
    functor(Tm,Name,Arity),
    Tm =.. [_|TArgs],
    findall(VTSr, vts_UpApprox_one_rule(Name,Arity,TArgs,ATp,VTSr), VTSrSet),
    append_all(VTSrSet,VTS).

    append_all([],[]).
    append_all([VTSi|VTSSi],VTSo) :-
       append_all(VTSSi,VTS1),
       append(VTSi,VTS1,VTSo). 

    vts_UpApprox_one_rule(Name,0,[],ATp,[[]]) :- % non-deterministic 
        prim_type(ATp),     % extra clause for contants and primitive types
        membership_prim(ATp,Name).            % do not add cut !.

    vts_UpApprox_one_rule(Name,Arity,Targs,ATp,VTSr) :- % non-deterministic
        typedef(Name/Arity, Tis, ATp),               % multiple type rules
        once(vts_UpApprox_one_rule_args(Tis,Targs,[[]], VTSr)). 
                                         %[[]] empty set of variable  typing

        vts_UpApprox_one_rule_args([],[],VTSr, VTSr). 
        vts_UpApprox_one_rule_args([R|Rs],[Tm|Tms],VTSr0, VTSr) :- 
            vts_UpApprox(R,Tm,VTSr1),
            vtsglb(VTSr0,VTSr1,VTSr2),
            vts_UpApprox_one_rule_args(Rs,Tms,VTSr2,VTSr).
 
*/
    
/*

3 ?- vts_UpApprox(list(nat), cons(nil,'$VAR'(1)), L).

L = [] ;


[debug] 5 ?- vts_UpApprox(list(nat),cons('$VAR'(1),'$VAR'(2)),L).

L = [[B/nat, C/list(nat)]] ;

No
[debug] 6 ?- vts_UpApprox(list(int),cons('$VAR'(1),'$VAR'(2)),L).

L = [[B/int, C/list(int)]] ;

No
[debug] 7 ?- vts_UpApprox(or(list(nat),list(list(top))),cons('$VAR'(1),'$VAR'(2)),L).

L = [[B/nat, C/list(nat)], [B/list(top), C/list(list(top))]] ;

No
[debug] 8 ?- vts_UpApprox(and(list(nat),list(list(top))),cons('$VAR'(1),'$VAR'(2)),L).

L = [[B/ (nat and list(top)), C/ (list(nat)and list(list(top)))]] ;

No
[debug] 9 ?- vts_UpApprox(list(top),cons('$VAR'(1),'$VAR'(2)),L).

L = [[B/top, C/list(top)]] ;

No
[debug] 10 ?- vts_UpApprox(top,cons('$VAR'(1),'$VAR'(2)),L).

L = [[]] ;

?- rtype_down([['$VAR'(1)/list(even),'$VAR'(2)/list(nat)]], '$VAR'(1), 
|    cons('$VAR'(2),nil), L).

L = [[B/list(even), C/ (list(nat)and even)]] ;


*/

tand(T1,T2,T3)  :- 
    ( 
        etype_tabled(and(T1,not(T2))) % T1<=T2
    ->  
        T3 = T1
    ; 
        etype_tabled(and(T2,not(T1))) % T2<=T1
    -> 
        T3 = T2
    ;
        T3 = and(T1,T2) 
    ).

/*

tand(top,Tp,Tp) :- !.
tand(Tp,top,Tp) :- !.
tand(T1,T2,T3)  :- ( T1==T2 -> T3 = T1; T3 = and(T1,T2) ).

*/


tor(T1,T2,T3)  :- 
    ( 
        etype_tabled(and(T1,not(T2))) % T1<=T2
    ->  
        T3 = T2
    ; 
        etype_tabled(and(T2,not(T1))) % T2<=T1
    -> 
        T3 = T1
    ;
        T3 = or(T1,T2) 
    ).

tleq(T1,T2) :- etype_tabled(and(T1,not(T2))).

/*

tor(bot,Tp,Tp) :- !.
tor(Tp,bot,Tp) :- !.
tor(T1,T2,T3)  :-  ( T1==T2 -> T3 = T1; T3 = or(T1,T2) ).

*/


:- use_module(counter).

:- use_module(sicstus). 

% tabling etype, using XSB, we can expect much better performance


:- dynamic '$etype_tabled'/2. 

rtype_init :- retractall('$etype_tabled'(_,_)).

/*

% this code implements tabling

etype_tabled(T) :- 
    inc_counter('total_emptiness_checks'),
    (
        '$etype_tabled'(T,Ans) 
    -> 
        Ans = yes
    ;
        inc_counter('distinct_emptiness_checks'),
        current_time(TStart),
        ( 
            etype(T) 
        -> 
            current_time(TEnd),
            Lapse is TEnd-TStart,
            add_counter(checking_time,Lapse),
            assert('$etype_tabled'(T,yes)) 
	 ;
            current_time(TEnd),
            Lapse is TEnd-TStart,
            add_counter(checking_time,Lapse),
            assert('$etype_tabled'(T,no)),
            !,
            fail 
        )
    ).

*/

% the following implementation doesn't memoize but measure the time 
 
etype_tabled(T) :- 
    inc_counter('total_emptiness_checks'),
    (
        current_time(TStart),
        ( 
            etype(T) 
        -> 
            current_time(TEnd),
            Lapse is TEnd-TStart,
            add_counter(checking_time,Lapse)
	 ;
            current_time(TEnd),
            Lapse is TEnd-TStart,
            add_counter(checking_time,Lapse),
            !,
            fail 
        )
    ).




/* type_LowApprox(+Tm,+VT,-Tp) computes -Tp such that
   \theta\in\gamma_vt(VT) if \theta(Tm)\in\gamma_t(Tp) */

type_LowApprox(V,VT,Tp) :- 
    variable(V),
    !,
    ( mapsto(V,VT,Tp1) -> type_dnf(Tp1,Tp) ; Tp=top ).

type_LowApprox(Tm,VT,Tp) :- 
    functor(Tm,Fn,Ar),
    (
        Ar =:= 0
    ->
        Tp = top % No type contstraint on Tm is induced by VT
     ;  
        Tm =.. [_|Args],
        type_LowApprox_map(Args,VT,TArgs),
%        findall(CT,(one_mix(TArgs,CTArgs),
%		    type_LowApprox_one_level(Fn/Ar,CTArgs,CT)),CTs),
%        bigtlub(CTs,Tp) % if two types satisfy the spec then their union does 
         type_LowApprox_one_level(Fn/Ar,TArgs,Tp)

     ).

type_LowApprox_map([],_,[]).
type_LowApprox_map([Tm|Tms],VT,[Tp|Tps]) :- 
    type_LowApprox(Tm,VT,Tp),
    type_LowApprox_map(Tms,VT,Tps).

/* type_LowApprox_one_level(+Fn/Ar,+CTs,-CT) computes CT such that
   ti\in\gamma_t(CTs(i)) if Fn(t1,...tAr)\in\gamma_t(CT). */


type_LowApprox_one_level(Fn/Ar,RTs,RT) :- % Ar > 0
    findall(AT,type_LowApprox_one_level_one_rule(Fn/Ar,RTs,AT),ATs),
    bigtlub(ATs,RT). % if two types satisfy the spec then their union does 

type_LowApprox_one_level_one_rule(Fn/Ar,RTs,AT) :- % non-deterministic
    % Ar > 0,
    typedef(Fn/Ar,Tis,To),
    tvs_LowApprox_vec(RTs,Tis,[tv_top],TVS), 
    tvs_apply(TVS,To,AT).

/* tvs_LowApprox_vec(+RTs,+Tis,+TVi,-TVo) TVo=TVi\/K such that
   RTs>=K(Tis) */

tvs_LowApprox_vec([],[],TVSi,TVSi). 
tvs_LowApprox_vec([R|Rs],[P|Ps],TVSi,TVSo) :-
    type_dnf(R,DC),
    tvs_LowApprox(DC,P,TVS1),
    tvs_glb(TVS1,TVSi,TVS2),
    tvs_LowApprox_vec(Rs,Ps,TVS2,TVSo).


/* tvs_LowApprox(+RT,+PT,-TVS) TVS(PT) <= RT specified in the paper */

tvs_LowApprox(or(C1,C2),P,TVSo) :- 
    !,
    tvs_LowApprox(C1,P,TVS1),
    tvs_LowApprox(C2,P,TVS2),
    append(TVS1,TVS2,TVSo). 
tvs_LowApprox(and(A,C),P,TVSo) :- 
    !,
    tvs_LowApprox(A,P,TVSa),
    tvs_LowApprox(C,P,TVSc),
    tvs_glb(TVSa,TVSc,TVSo). 
tvs_LowApprox(A,P,[[P/A]]) :-
    var(P), %P is a parameter in a type rule
    !.
tvs_LowApprox(not(A),P,TVSo) :- 
    !,
    negate_atom_LowApprox(A,DC),
    tvs_LowApprox(DC,P,TVSo).

tvs_LowApprox(top,P,[TV]) :- 
    !,
    P =..[Cons|Ps],
    functor(P,Cons,Arity),
    tops(Arity,Ones),
    make_map(Ps,Ones,TV).

tvs_LowApprox(A,P,TVS) :- 
    (
        A =..[Cons|Rs],P =..[Cons|Ps]
    ->  
        make_map(Ps,Rs,TV), TVS=[TV]
     ;
        TVS = [tv_bot]
    ).


%% Following code is for dealing with negative types 

/* negate_atom_UpApprox(A,R) R is an upper approximation to not(A). R is a
   disjunction of atomic types. */

negate_atom_UpApprox(Atom,R) :- 
    typedef_constructors(Cnsts),
    functor(Atom,Cnst,Arity),
    delete(Cnsts,Cnst/Arity,Cnsts0), delete(Cnsts0,top/0,Cnsts1),
    most_gen_types(Cnsts1,bot,R1),
    Atom =.. [Cnst|Seq],
    ( 
      Seq = [] 
    -> 
      R = R1 
     ;
      negate_seq(Seq,DSeq),
      apply_cnst_to_dseq(DSeq,Cnst,R2),
      tor(R1,R2,R)
    ).

    most_gen_types([],R,R).
    most_gen_types([Cnst/Arity|Cnsts],Racc,R) :-
        tops(Arity,Ones),
        A =..[Cnst|Ones],
        tor(A,Racc,Racc1),
        most_gen_types(Cnsts,Racc1,R).


        tops(0,[]).
        tops(N,[top|Tops]) :-
            N > 0,
            N1 is N-1,
            tops(N1,Tops).

    apply_cnst_to_dseq(or(DSeq1,DSeq2),Cnst,R) :- 
        !,
        apply_cnst_to_dseq(DSeq1,Cnst,R1),
        apply_cnst_to_dseq(DSeq2,Cnst,R2),
        tor(R1,R2,R).
    apply_cnst_to_dseq(Seq,Cnst,R) :- 
        R =.. [Cnst|Seq]. 

/* negate_atom_LowApprox(A,R) R is an Lower approximation to not(A). R
   is a disjunction of atomic types. */

negate_atom_LowApprox(Atom,R) :- 
    typedef_constructors(Cnsts),
    functor(Atom,Cnst,Arity),
    findall(NA,overlap_cnst(Cnst/Arity,NA),Overlaps),
    delete_all([Cnst/Arity,top/0|Overlaps],Cnsts,Cnsts1),
    most_gen_types(Cnsts1,bot,R1),
    Atom =.. [Cnst|Seq],
    ( 
      Seq = [] 
    -> 
      R = R1 
     ;
      negate_seq(Seq,DSeq),
      apply_cnst_to_dseq(DSeq,Cnst,R2),
      tor(R1,R2,R)
    ).

    overlap_cnst(Cnst1/Arity1,Cnst2/Arity2) :- 
        functor(AT1,Cnst1,Arity1),
        typedef(Fn,_,AT1),
        typedef(Fn,_,AT2),
        functor(AT2,Cnst2,Arity2).

    delete_all([],L,L).
    delete_all([D|Ds],Li,Lo) :- 
        delete(Li,D,Li1),
        delete_all(Ds,Li1,Lo).

tvs_glb(TVSi1,TVSi2,TVSo) :- % pair-wise application of tv_glb
    tvs_glb(TVSi1,TVSi2,[],TVSo).

    tvs_glb([],_TVS,TVSo,TVSo).

    tvs_glb([TV|TVSi1],TVSi2,TVSoAcc,TVSo) :- 
        tvs_glb_1(TVSi2,TV,TVSoAcc,TVSoAcc1),
        tvs_glb(TVSi1,TVSi2,TVSoAcc1,TVSo).

    tvs_glb_1([],_TV,TVSo,TVSo).
    tvs_glb_1([TV2|TVS2],TV1,TVSoAcc,TVSo) :-
        tv_glb(TV1,TV2,TV12),
        tvs_glb_1(TVS2,TV1,[TV12|TVSoAcc],TVSo).

tvs_lub(TVSi1,TVSi2,TVSo) :- % pair-wise application of tv_lub
    tvs_lub(TVSi1,TVSi2,[],TVSo).

    tvs_lub([],_TVS,TVSo,TVSo).

    tvs_lub([TV|TVSi1],TVSi2,TVSoAcc,TVSo) :- 
        tvs_lub_1(TVSi2,TV,TVSoAcc,TVSoAcc1),
        tvs_lub(TVSi1,TVSi2,TVSoAcc1,TVSo).

    tvs_lub_1([],_TV,TVSo,TVSo).
    tvs_lub_1([TV2|TVS2],TV1,TVSoAcc,TVSo) :-
        tv_lub(TV1,TV2,TV12),
        tvs_lub_1(TVS2,TV1,[TV12|TVSoAcc],TVSo).
    
vts_UpApprox(and(R1,R2),Tm,VTS) :- 
%%always called with first argument begin in disjuctive normal form
    !,
    vts_UpApprox(R1,Tm,VTS1),
    vts_UpApprox(R2,Tm,VTS2),
    vtsglb(VTS1,VTS2,VTS).

vts_UpApprox(or(R1,R2),Tm,VTS) :- 
    !,
    vts_UpApprox(R1,Tm,VTS1),
    vts_UpApprox(R2,Tm,VTS2),
    append(VTS1,VTS2,VTS).

vts_UpApprox(top,_,[[]]) :- 
    !.

vts_UpApprox(R,V,[[V/R]]) :-
    variable(V),
    !.

vts_UpApprox(not(A),Tm,VTS) :- 
    !,
    negate_atom_UpApprox(A,R),
    vts_UpApprox(R,Tm,VTS).

vts_UpApprox(ATp,Tm,VTS) :- 
    functor(Tm,Name,Arity),
    Tm =.. [_|TArgs],
    findall(VTSr, vts_UpApprox_one_rule(Name,Arity,TArgs,ATp,VTSr), VTSrSet),
    append_all(VTSrSet,VTS).

    append_all([],[]).
    append_all([VTSi|VTSSi],VTSo) :-
       append_all(VTSSi,VTS1),
       append(VTSi,VTS1,VTSo). 

    vts_UpApprox_one_rule(Name,0,[],ATp,[[]]) :- % non-deterministic 
        prim_type(ATp),     % extra clause for contants and primitive types
        membership_prim(ATp,Name).            % do not add cut !.

    vts_UpApprox_one_rule(Name,Arity,Targs,ATp,VTSr) :- % non-deterministic
        typedef(Name/Arity, Tis, ATp),               % multiple type rules
        vts_UpApprox_one_rule_args(Tis,Targs,[[]], VTSr). 
                                         %[[]] empty set of variable  typing

        vts_UpApprox_one_rule_args([],[],VTSr, VTSr). 
        vts_UpApprox_one_rule_args([R|Rs],[Tm|Tms],VTSr0, VTSr) :-
            type_dnf(R,DC), 
            % make sure vts_UpApprox is called with first argument in DNF
            vts_UpApprox(DC,Tm,VTSr1),
            vtsglb(VTSr0,VTSr1,VTSr2),
            vts_UpApprox_one_rule_args(Rs,Tms,VTSr2,VTSr),
            !.


vts_LowApprox(and(R1,R2),Tm,VTS) :- 
%%always called with first argument begin in disjuctive normal form
%%only the clause for negative types differs from that for upper approx 
    !,
    vts_LowApprox(R1,Tm,VTS1),
    vts_LowApprox(R2,Tm,VTS2),
    vtsglb(VTS1,VTS2,VTS).

vts_LowApprox(or(R1,R2),Tm,VTS) :- 
    !,
    vts_LowApprox(R1,Tm,VTS1),
    vts_LowApprox(R2,Tm,VTS2),
    append(VTS1,VTS2,VTS).

vts_LowApprox(top,_,[[]]) :- 
    !.

vts_LowApprox(R,V,[[V/R]]) :-
    variable(V),
    !.

vts_LowApprox(not(A),Tm,VTS) :- 
    !,
    negate_atom_LowApprox(A,R),
    vts_LowApprox(R,Tm,VTS).


vts_LowApprox(ATp,Tm,VTS) :- 
    functor(Tm,Name,Arity),
    Tm =.. [_|TArgs],
    findall(VTSr, vts_LowApprox_one_rule(Name,Arity,TArgs,ATp,VTSr), VTSrSet),
    append_all(VTSrSet,VTS).

    vts_LowApprox_one_rule(Name,0,[],ATp,[[]]) :- % non-deterministic 
        prim_type(ATp),     % extra clause for contants and primitive types
        membership_prim(ATp,Name).            % do not add cut !.

    vts_LowApprox_one_rule(Name,Arity,Targs,ATp,VTSr) :- % non-deterministic
        typedef(Name/Arity, Tis, ATp),               % multiple type rules
        vts_LowApprox_one_rule_args(Tis,Targs,[[]], VTSr). 
                                         %[[]] empty set of variable  typing

        vts_LowApprox_one_rule_args([],[],VTSr, VTSr). 
        vts_LowApprox_one_rule_args([R|Rs],[Tm|Tms],VTSr0, VTSr) :-
            type_dnf(R,DC), 
            % make sure vts_LowApprox is called with first argument in DNF
            vts_LowApprox(DC,Tm,VTSr1),
            vtsglb(VTSr0,VTSr1,VTSr2),
            vts_LowApprox_one_rule_args(Rs,Tms,VTSr2,VTSr),
            !.





tvs_apply(TVS,PT,RT) :- tvs_apply(TVS,PT,bot,RT).

tvs_apply([],_,RT,RT).
tvs_apply([TV|TVS],PT,Racc,RT) :- 
    tv_apply(TV,PT,RT1),
    tor(Racc,RT1,RT2),
    tvs_apply(TVS,PT,RT2,RT).


/* vts_negate(VTSi,VTSo) VTSo = ~VTSi */

vts_negate(VTSi,VTSo) :- 
    vts_negate(VTSi,[[]],VTSo).

    vts_negate([],VTSo,VTSo).
    vts_negate([VTi|VTSi],VTSacc,VTSo) :- 
        vt_negate(VTi,[],VTS1),
        vtsglb(VTS1,VTSacc,VTSacc1),
        vts_negate(VTSi,VTSacc1,VTSo).

    vt_negate([],VTSo,VTSo).
    vt_negate([X/T|VT],VTSacc,VTSo) :- 
        type_dnf(not(T),NotT),
        !,
        vt_negate(VT,[[X/NotT]|VTSacc],VTSo).

/* test0(VTS) :- 
      vts_negate([['$VAR'(1)/integer,'$VAR'(2)/list(top)],
		['$VAR'(1)/list(integer)]],VTS).
*/

/* vts_proj_forall(VTSi,NVar,VTSo) VTSo approximates from below
   \forall Vars. VTSi where Vars are set of variables that have an
   identification number greater than NVar.  */

vts_proj_forall(VTSi,NVar,VTSo) :- 
    vts_negate(VTSi,VTSne), % VTSne=~VTSi
    reduce(VTSne,VTSneReduced), % remove redundency
    vts_restrict(VTSneReduced,NVar,VTSne1), % existential projection
    vts_negate(VTSne1,VTSo).


test1(VTS) :-  vts_proj_forall([['$VAR'(1)/list(top),'$VAR'(2)/top]],1,VTS).
test2(VTS) :-  
    vts_proj_forall([['$VAR'(1)/list(top),'$VAR'(2)/list(list(top))]],1,VTS).
test3(VTS) :-  vts_proj_forall([ 
		['$VAR'(1)/list(top),'$VAR'(2)/list(top)],
		['$VAR'(1)/list(top),'$VAR'(2)/not(list(top))]],1,VTS).


/* pseduo_complement(Left,Right,Complement) Complement = Left -> Right */

pseduo_complement(Left,Right,Complement) :-
    vts_negate(Left,LeftNe),          % negation
    append(Right,LeftNe,Complement1), % set union
    reduce(Complement1,Complement).   % remove redundency

/*
test4(X) :- pseduo_complement([['$VAR'(1)/integer]], [['$VAR'(1)/integer]], X).
*/

/* rtype_solve_bw(+EQS,+VTSright,-VTSleft): reverse abstract constraint
   solving */

rtype_solve_bw(EQS,VTSRt,VTSLt) :-
    rtype_solve_bw(VTSRt,EQS,[], VTSLt). % [] initial accumulator

rtype_solve_bw([],_EQS,VTSLt,VTSLt).
rtype_solve_bw([VTRt|VTSRt],EQS,VTSLtAcc,VTSLt) :- 
    findall(VTLt,rewrite_to_solved_form(EQS,VTRt,VTLt),VTSLt1),
    append(VTSLt1,VTSLtAcc,VTSLtAcc1),
    reduce(VTSLtAcc1,VTSLtAcc2), 
    rtype_solve_bw(VTSRt,EQS,VTSLtAcc2,VTSLt).

/* vt_split(+Vs,+VTi,-VTo1,-VTo2) VTo1=VTi_restricted_to_Vs and
   VTo2=VTi_restricted_to_other_variables. */

vt_split([],VT,[],VT).
vt_split([V|Vs],VTi,VTo1,VTo2) :- 
    vt_split_1(VTi,[V|Vs],VTo1,VTo2).
    

vt_split_1([],_,[],[]).
vt_split_1([V/T|VT],[V|Vs],[V/T|VTo1],VTo2) :- 
    !,
    vt_split(Vs,VT,VTo1,VTo2).
vt_split_1([Eq|VT],[V|Vs],VTo1,[Eq|VTo2]) :- 
    !,
    vt_split([V|Vs],VT,VTo1,VTo2).



/* rtype_bw_compile(+Call,-Compiled) compiles a bip call to its
   compiled form to factor out computation that is independent of
   abstract input state. Success of a builtin call is a conjunction
   (Sd and S~d) where Sd is implied by the demand of the builtin
   call. Compiled is set to S~d for most builtins. */

rtype_bw_compile(recorda(_Key,_Term,Ref), complement(VTS))    :-
    vts_UpApprox(integer,Ref,VTS).
rtype_bw_compile(recorded(_Key,_Term,Ref), complement(VTS))    :-
    vts_UpApprox(integer,Ref,VTS).

rtype_bw_compile(abolish(_),       copy).
rtype_bw_compile(assert(_),        copy).
rtype_bw_compile(retract(_),       copy).
rtype_bw_compile(erase(_I),        copy). 
rtype_bw_compile(format(_Fmt),     copy).
rtype_bw_compile(format(_Fmt,_),   copy).
rtype_bw_compile(format(_,_Fmt,_), copy).
rtype_bw_compile(put(_N),          copy).

rtype_bw_compile(keysort(_L,R), complement(VTS)) :- 
    %    vts_UpApprox(list(pair(top,top)), L,VTS1),
    %    vts_UpApprox(list(pair(top,top)), R,VTS2),
    %    vtsglb(VTS1,VTS2,VTS). 
    vts_UpApprox(list(pair(top,top)), R,VTS).

rtype_bw_compile(sort(_L,R), complement(VTS)) :- 
    %    vts_UpApprox(list(top), L,VTS1),
    %    vts_UpApprox(list(top), R,VTS2),
    %    vtsglb(VTS1,VTS2,VTS). 
    vts_UpApprox(list(top), R,VTS).

rtype_bw_compile(statistics(_N1,N2), complement(VTS)) :- 
     vts_UpApprox(list(integer),N2,VTS). 
 
rtype_bw_compile(compare(A,_,_), VTS) :- 
     vts_UpApprox(atom, A, VTS).


rtype_bw_compile(repeat,           copy).
rtype_bw_compile(listing,          copy).
rtype_bw_compile(listing(_),       copy).
rtype_bw_compile(compound(_),      copy).
rtype_bw_compile(portray_clause(_),copy).
rtype_bw_compile(ground(_),        copy).
rtype_bw_compile(!,	        copy).
rtype_bw_compile(display(_),	copy).
rtype_bw_compile(nl,            copy).
rtype_bw_compile(print(_),      copy).
rtype_bw_compile(write(_),      copy).
rtype_bw_compile(writeq(_),	copy).
rtype_bw_compile(@<(_,_),	copy). % comparing on term order
rtype_bw_compile(@>(_,_),	copy).
rtype_bw_compile(@=<(_,_),	copy).
rtype_bw_compile(@>=(_,_),      copy).
rtype_bw_compile(true,          copy).
rtype_bw_compile(nonvar(_),	copy).
rtype_bw_compile(read(_),	        copy).
rtype_bw_compile(\==(_,_),         copy).
rtype_bw_compile(\=(_,_),          copy).
rtype_bw_compile(=\=(_,_),	        copy).
rtype_bw_compile(var(_V), copy).


% these built-ins change type information as fail

%rtype_bw_compile(abort,            _) :- fail.
%rtype_bw_compile(fail,             _) :- fail.
%rtype_bw_compile(false,	         _) :- fail.

% these built-ins change type information as unification

rtype_bw_compile(=(Left,Right),    dsolve(E0))     :-
	mgu(Left,Right,E0).
rtype_bw_compile(=@=(Left,Right),  dsolve(E0))     :-
	mgu(Left,Right,E0).
        % structurally identical terms must unify
rtype_bw_compile(==(Left,Right),   dsolve(E0))     :-
	mgu(Left,Right,E0).
	% identical terms must unify
rtype_bw_compile(is(Left,Right),   dsolve(E0))     :-
	mgu(Left,Right,E0).
	% identical terms must unify

% these built-ins type check their arguments

rtype_bw_compile(float(Tm), complement(VTS))     :-
        vts_UpApprox(float,Tm,VTS).
rtype_bw_compile(string(Tm), complement(VTS))    :-
        vts_UpApprox(string,Tm,VTS).
rtype_bw_compile(atom(Tm),  complement(VTS))     :-
        vts_UpApprox(atom,Tm,VTS).
rtype_bw_compile(atomic(Tm), complement(VTS))    :-
        vts_UpApprox(atomic,Tm,VTS).
rtype_bw_compile(integer(Tm), complement(VTS))   :-
        vts_UpApprox(integer,Tm,VTS).
rtype_bw_compile(number(Tm), complement(VTS))    :-
        vts_UpApprox(number,Tm,VTS).
rtype_bw_compile(tab(Tm), complement(VTS))       :-
        vts_UpApprox(integer,Tm,VTS).

rtype_bw_compile(<(_,_), copy).
rtype_bw_compile(>(_,_), copy).
rtype_bw_compile(=<(_,_),copy).
rtype_bw_compile(>=(_,_),copy).

rtype_bw_compile(length(L,N),  complement(VTS))     :-
        vts_UpApprox(integer,N,VTS1),
        vts_UpApprox(list(top),L,VTS2),
        vtsglb(VTS1,VTS2,VTS). 
rtype_bw_compile(=..(_,L), complement(VTS)) :-
        vts_UpApprox(list(top),L,VTS).
rtype_bw_compile(functor(_,Nm,Ar),  complement(VTS))     :-
        vts_UpApprox(integer,Ar,VTS1),
        vts_UpApprox(atom,Nm,VTS2),
        vtsglb(VTS1,VTS2,VTS). 

rtype_bw_compile(arg(N,_,_), complement(VTS))     :- 
        vts_UpApprox(integer,N,VTS).
rtype_bw_compile(name(AI,S), complement(VTS))     :-
        vts_UpApprox(or(atom,integer),AI,VTS1),
        vts_UpApprox(string,S,VTS2),
        vtsglb(VTS1,VTS2,VTS). 
rtype_bw_compile(=:=(_,_), copy).



/* rtype_bw_bipai(+Call,+ASubi,-ASubo) ASubo describes the set of
   substitutions that satisfy ASubi and that are answers to Call.
*/

rtype_bw_bipai(copy,ASubi,ASubi). 
rtype_bw_bipai(complement(VTS),ASubi,ASubo)  :- 
    pseduo_complement(VTS,ASubi,ASubo). 
rtype_bw_bipai(dsolve(E0),ASubi,ASubo) :- 
    rtype_solve_bw(E0,ASubi,ASubo).



/* rewrite(+Eqsi,+VTi,-Eqso,-VTo) <Eqsi,VTi> rewrites
   into <Eqso,VTo> as specified in the paper */

rewrite(EQSi,VTi,EQSo,VTo) :- 
    select(Var/Tm, EQSi, EQSo), % non-deterministic
    once(vt_split([Var],VTi,_,Exists_Var_VTi)),
    vars(Tm,VarsInTm),
    once(vt_split(VarsInTm,VTi,Exists_VarsNotInTm_VTi,Exists_VarsInTm_VTi)),
    (
        (  %EQ-UP 
	       Exists_VarsNotInTm_VTi = [] 
	 -> 
	       VTo = VTi % VTi does not contrain type of Tm
	 ; 
	       type_LowApprox(Tm,VTi,TpTmLow),                
               ( 
		   TpTmLow = top 
		-> 
		   VTo = Exists_VarsInTm_VTi 
		;
	           vtglb([Var/TpTmLow],Exists_VarsInTm_VTi,VTo) 
	       )
	)
     ;
        (                                          % EQ-DOWN
	     mapsto(Var,VTi,TpVar) 
	 -> 
	     once(vts_LowApprox(TpVar,Tm,VTS)),
             member(VT,VTS), % non-deterministic
             vtglb(VT,Exists_Var_VTi,VTo)
	 ;   % TpVar = top by default
	     VTo=Exists_Var_VTi
	 )
     ;
         VTo = VTi                                     % EQ-REMOVE
    ).

/* rewrite_to_solved_form(+Eqsi,+VTi,-VTo) <Eqsi,VTi>
   rewrites to <[],VTo> in zero on more steps */

rewrite_to_solved_form([],VTi,VTi).
rewrite_to_solved_form([EQ|EQS],VTi,VTo) :- 
    rewrite([EQ|EQS],VTi,EQS1,VT1),
    rewrite_to_solved_form(EQS1,VT1,VTo).
        


current_time(MiniSec) :-  statistics(runtime,[MiniSec,_]).
