:- module(typedef, [
       typedef_add/1,
       typedef_clear/0,
       typedef_constructors/1,
       typedef_functors/1,
       typedef_add_default/1,
       typedef/3
   ]).

:- use_module(library(lists)).


:- dynamic user_typedef/3, 
           typedef_constructors/1, typedef_functors/1.

typedef_clear :-
    retractall(user_typedef(_,_,_)),
    ( retract(typedef_functors(_)),fail; true ),
    ( retract(typedef_constructors(_)),fail; true ).


is_constructor(CN/CA) :- 
    typedef(_,_,Type),
    functor(Type,CN,CA).
is_functor(Ar-Nm) :- 
    typedef(Nm/Ar,_,_). 


/* typedef_add_default/1 It also collects all type constructors used
   in type definition rules, adding top/0, and bot/0, and then putting
   these constructors */

typedef_add_default(FsInP) :- % collecting constructors
    initialise_user_typedef(FsInP), 
    ( setof(F,is_functor(F), FsDef) -> true ; FsDef=[] ),
    assert(typedef_functors(FsDef)),
    ( setof(C,is_constructor(C), Cs0) -> true ; Cs0=[] ),
    append(Cs0,[top/0,bot/0],Cs),
    assert(typedef_constructors(Cs)). 

typedef_add((Hx:-Bx)) :-
        Hx = type(Function,Tout),
        conversion(Bx,Tins),
        functor(Function,Func,Arity),
        assertz(user_typedef(Func/Arity,Tins,Tout)).

conversion(true,[]) :-
    !.
conversion((type(_Var,PT),B),[PT|B1]) :-
    !,
    conversion(B,B1).
conversion(type(_Var,PT),[PT]). 


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%:- index(typedef(1,1,1)).
%:- index(user_typedef(1,1,1)).
%:- index(default_typedef(1,1,1)).

typedef(NA,Tis,To) :- 
    user_typedef(NA,Tis,To).
typedef(NA,Tis,To) :- 
    default_typedef(NA,Tis,To). 


/* One is interpreted as the set of all terms including variables. 
   Remove the following clause if types are sets of ground terms. 
*/

%:- index(default_typedef(1,1,1)).

%default_typedef('$arbitrary_var_constant'/0,  [], '$$$$dummy_variable').

/* 
   
   The following default type definitions will be used in emptiness
   test. Primitive types at the leaves of the type hierarchy contains
   an arbitrary constant that is assumed not to appear in any user
   defined type rule. Thus, such primitive types should contain
   sufficient large amount of constants. If a primitive type contains
   only a small number of constants, then they can be enumerated by
   type rules.
*/
 


default_typedef('$arbitrary_integer_constant'/0, [], integer). 
default_typedef('$arbitrary_float_constant'/0,   [], float).
default_typedef('$arbitrary_atom_constant'/0,    [], atom). 


% The following will add arbitray constants for primitive types 
% that are not leaves in the type hierarchy. 
 
default_typedef(Func, [], number) :- 
    default_typedef(Func, [], integer). 
default_typedef(Func, [], number) :- 
    default_typedef(Func, [], float). 

default_typedef(Func, [], atomic) :- 
    default_typedef(Func, [], number). 
default_typedef(Func, [], atomic) :- 
    default_typedef(Func, [], atom). 

default_typedef(([])/0,[],list(_)).
default_typedef((.)/2,[A,list(A)],list(A)).




% adding default type definition for functions and/or operators

initialise_user_typedef(FsInP) :-
    initialise_user_typedef(
         [1-(-), 1-(+), 2-(+), 2-(-), 2-(*), 2-(/)], 
         FsInP). 

    initialise_user_typedef([],_). 
    initialise_user_typedef([F|Fs], FsInP) :- 
         ( 
             memberchk(F,FsInP)
         -> 
             add_user_typedef(F)
	  ;
             true
         ),
         initialise_user_typedef(Fs, FsInP).

add_user_typedef(1-(-)) :- 
    assert(user_typedef((-)/1, [float],   float)),   % 
    assert(user_typedef((-)/1, [integer], integer)),  
    assert(user_typedef((-)/1, [number],  number)). 
add_user_typedef(1-(+)) :- 
    assert(user_typedef((+)/1, [float],   float)),   % 
    assert(user_typedef((+)/1, [integer], integer)),  
    assert(user_typedef((+)/1, [number],  number)).
add_user_typedef(2-(-)) :- 
    assert(user_typedef((-)/2, [integer,integer], integer)), 
    assert(user_typedef((-)/2, [integer,float],   float)), 
    assert(user_typedef((-)/2, [integer,number],  number)),
    assert(user_typedef((-)/2, [float,integer],   float)), 
    assert(user_typedef((-)/2, [float,float],     float)), 
    assert(user_typedef((-)/2, [float,number],    number)), 
    assert(user_typedef((-)/2, [number,integer],  number)), 
    assert(user_typedef((-)/2, [number,float],    number)), 
    assert(user_typedef((-)/2, [number,number],   number)),
    assert(user_typedef((-)/2,[A,B],pair(A,B))). 
        % a pair is a term of the form T1-T2 
add_user_typedef(2-(+)) :- 
    assert(user_typedef((+)/2, [integer,integer], integer)), 
    assert(user_typedef((+)/2, [integer,float],   float)), 
    assert(user_typedef((+)/2, [integer,number],  number)),
    assert(user_typedef((+)/2, [float,integer],   float)), 
    assert(user_typedef((+)/2, [float,float],     float)), 
    assert(user_typedef((+)/2, [float,number],    number)), 
    assert(user_typedef((+)/2, [number,integer],  number)), 
    assert(user_typedef((+)/2, [number,float],    number)), 
    assert(user_typedef((+)/2, [number,number],   number)).
add_user_typedef(2-(*)) :- 
    assert(user_typedef((*)/2, [integer,integer], integer)), 
    assert(user_typedef((*)/2, [integer,float],   float)), 
    assert(user_typedef((*)/2, [integer,number],  number)),
    assert(user_typedef((*)/2, [float,integer],   float)), 
    assert(user_typedef((*)/2, [float,float],     float)), 
    assert(user_typedef((*)/2, [float,number],    number)), 
    assert(user_typedef((*)/2, [number,integer],  number)), 
    assert(user_typedef((*)/2, [number,float],    number)), 
    assert(user_typedef((*)/2, [number,number],   number)). 
add_user_typedef(2-(/)) :- 
    assert(user_typedef((/)/2, [integer,integer], number)), 
    assert(user_typedef((/)/2, [integer,float],   float)), 
    assert(user_typedef((/)/2, [integer,number],  number)),
    assert(user_typedef((/)/2, [float,integer],   float)), 
    assert(user_typedef((/)/2, [float,float],     float)), 
    assert(user_typedef((/)/2, [float,number],    float)), 
    assert(user_typedef((/)/2, [number,integer],  number)), 
    assert(user_typedef((/)/2, [number,float],    float)),  
    assert(user_typedef((/)/2, [number,number],   number)).









