/* This version of analysis framework records for each
   point an abstract substitution.
*/

:- module(asem, [
          asem_bu/6,
	  asem_td/7,
          asem_bw/8,
          asem_display/2
   ]).

:- use_module('Domain/domain').

:- use_module('Domain/basics', 
     [bip/1,
      mgu/3,
      sub_apply/3,
      tagterm/3,
      peel_meta/2,
      vars/2,
      variable/1,
      variable_id/2,
      variable_ids/2,
      variable_max_no/1,
      nameListToSub/3
    ]).

:- use_module(library(lists)).

:- op(1150,fx,type).
:- op(1125,xfx, ::=).
%:- op(1100,xfy,;).





:- dynamic invoked_predicate/1,
           defined_predicate/1,
           used_functor/1.

:- dynamic user_clause/9.

pg_LINES(S) :- next_Pt(NxtPt), S is NxtPt-1.

add_invoked(P) :- invoked_predicate(P) -> true ; assert(invoked_predicate(P)).
add_defined(P) :- defined_predicate(P) -> true ; assert(defined_predicate(P)).
add_used_functor(F) :- used_functor(F) -> true ; assert(used_functor(F)).



undef_preds(Ps) :- bagof(P,undef_pred(P),Ps) -> true ; Ps=[].
undef_pred(P) :- invoked_predicate(P), \+ defined_predicate(P), \+ isbip(P).

isbip(PN/PA) :- functor(A,PN,PA),bip(A).




:- dynamic bw_analysis/0. 

asem_bu(Alpha,Pg,InitLapse,BottomUpLapse,SIZE,Undefined) :-
        asub_init(Alpha),
        current_time(InitStart),
        init_bu_bw(Pg,Alpha),
        pg_LINES(SIZE),
        undef_preds(Undefined),
        current_time(InitEnd),
	!,
        lfp(Alpha),
        current_time(BottomUpEnd),
        InitLapse is InitEnd-InitStart,
        BottomUpLapse is BottomUpEnd-InitEnd.

:- dynamic select_cls/5,
           call_proc/4.


/* generating flow arcs for top-down analysis. The implementation can
    be improved by going through list of call_proc/4 tuples once
    instead of twice. */

gen_td_flow_arcs(Alpha) :- % arcs for all calls to user-defined predicates 
    call_proc(Pt0,PtAns,A,NVarCall),
    select_cls(PtAns,PtEntry,PtExit,Head,NVarCls),
    (   tagterm(A,NVarCls,ARenamed),
        mgu(ARenamed,Head,Sub) 
     -> 
	Pt1 is Pt0+1,
        asub_eq_constraints(Alpha,Sub,F),
        assert(ntt(Pt0, do_call_fw(Pt0,PtEntry,F,NVarCall,NVarCls))),
	tagterm(Head,NVarCall,HeadR),
	mgu(HeadR,A,Sub2),
	asub_eq_constraints(Alpha,Sub2,F2),
	assert(ntt(Pt0,  do_exit(PtExit,Pt0,Pt1,F2,NVarCls,NVarCall))),
	assert(ntt(PtExit,  do_exit(PtExit,Pt0,Pt1,F2,NVarCls,NVarCall)))
    ; 
        true 
    ),
    fail.
gen_td_flow_arcs(Alpha) :- % arcs for all calls to builtin predicates 
    call_proc(Pt0,_PtAns,A,NVarCall), 
    functor(A,PN,PA), 
    \+ defined_predicate(PN/PA),
    bip(A),
    del_ans_clause(PN,PA),
    asub_compile(Alpha,A,C),
    Pt1 is Pt0+1,
    assert(ntt(Pt0,do_bip(Pt0,Pt1,C,NVarCall))),
    fail.
gen_td_flow_arcs(_Alpha).


gen_bu_flow_arcs(Alpha) :- 
    call_proc(Pt0,PtAns,A,NVarCall),
    functor(A,PN,PA),
    defined_predicate(PN/PA),
    Pt1 is Pt0+1,
    ( 
      bw_analysis
    ->
      ans_pidPt(PN,PA,PtAns,BAns),
      tagterm(BAns,NVarCall,BAnsRenamed),
      mgu(BAnsRenamed,A,SUB),
      asub_eq_constraints(Alpha,SUB,EC),
      assert(ntt_bw(Pt1,do_exit_bw(PtAns,Pt1,Pt0,EC,PA,NVarCall))),
      assert(ntt_bw(PtAns,do_exit_bw(PtAns,Pt1,Pt0,EC,PA,NVarCall))),
      assert(ntt_bw(PtAns,do_call_bw(PtAns,Pt0,EC,PA,NVarCall)))
    ;
      true
    ),
    select_cls(PtAns,_PtEntry,PtExit,Head,NVarCls),
    ( 	tagterm(Head,NVarCall,HeadR),
	mgu(HeadR,A,Sub2)
     ->
	asub_eq_constraints(Alpha,Sub2,F2),
	assert(ntt(Pt0,  do_exit(PtExit,Pt0,Pt1,F2,NVarCls,NVarCall))),
	assert(ntt(PtExit,  do_exit(PtExit,Pt0,Pt1,F2,NVarCls,NVarCall)))
    ; 
        true 
    ),
    fail.
gen_bu_flow_arcs(Alpha) :- % arcs for all calls to builtin predicates 
    call_proc(Pt0,_PtAns,A,NVarCall), 
    functor(A,PN,PA), 
    \+ defined_predicate(PN/PA),
    bip(A),
    del_ans_clause(PN,PA),
    asub_compile(Alpha,A,Compiled),
    Pt1 is Pt0+1,
    assert(ntt(Pt0,do_bip(Pt0,Pt1,Compiled,NVarCall))),
    ( 
      bw_analysis, asub_bw_compile(Alpha,A,Compiled_bw)
    ->
      assert(ntt_bw(Pt1,do_bip_bw(Pt1,Pt0,Compiled_bw,NVarCall))),
      ( asub_bip_has_cond(Alpha,A,Cond) -> assert(demand(Pt0,Cond)) ; true ) 
    ;
      true
    ),
    fail.
gen_bu_flow_arcs(_Alpha).


asem_td(Alpha,Pg,(Goal,InputASubX,Tab),InitLapse,TDLapse,SIZE,Undefined) :-
    current_time(InitStart),
    asub_init(Alpha),
    init_td(Pg,(Goal,InputASubX,Tab),Alpha), 
    pg_LINES(SIZE),
    undef_preds(Undefined),
    current_time(InitEnd),
    !,
    lfp(Alpha),
    current_time(TDEnd),
    InitLapse is InitEnd-InitStart,
    TDLapse is TDEnd-InitEnd.


lfp(Alpha) :-  %only for top-down forward analysis 
        retract(do_call_fw(Q,P,Sub,NVarR,NVar)), 
        ( 
           sem(Q,Sigma) 
        -> 
           asub_unify_entry(Alpha,Sub,Sigma,NVarR,NVar,Zeta),
	   ( asub_bot(Alpha,NVar,Zeta) -> true
	   ; sem_update_lub(Alpha,P,Zeta)
	   )
        ;
           true
        ),
	!,
        lfp(Alpha).


lfp(Alpha) :-
        retract(do_bip(Q,P,A,NVar)),
        (  
           sem(Q,Theta)
        ->
           asub_bipai(Alpha,A,Theta,Zeta),
	   ( asub_bot(Alpha,NVar,Zeta) -> true
	   ; sem_update_lub(Alpha,P,Zeta)
	   )
        ;
           true
        ),
	!,
        lfp(Alpha).


lfp(Alpha) :-  
        retract(do_exit(Q,P_,P,Sub,NVarR,NVar)),
        ( 
           sem(Q,Sigma), sem(P_,Theta)
        ->
	   asub_unify_exit(Alpha,Sub,Sigma,NVarR,Theta,NVar,Zeta),
           ( asub_bot(Alpha,NVar,Zeta) -> true
	     ;
	     sem_update_lub(Alpha,P,Zeta)
	   )
        ;
           true
        ), 
	!,
        lfp(Alpha).


lfp(Alpha) :-
        retract(do_gather_ans(Q,P,Sub,NVarR,NVar)), 
        ( 
           sem(Q,Sigma) 
        -> 
           asub_unify_entry(Alpha,Sub,Sigma,NVarR,NVar,Zeta),
	   ( asub_bot(Alpha,NVar,Zeta) -> true
	   ; ans_update_lub(Alpha,P,Zeta)
	   )
        ;
           true
        ),
	!,
        lfp(Alpha).

lfp(_Alpha).


asem_bw(Alpha,Pg,InitLapse,BULapse,AnnLapse,BWLapse,SIZE,Undefined) :- 
        asub_init(Alpha),
        current_time(InitStart),
        assert(bw_analysis),
        init_bu_bw(Pg,Alpha),
        pg_LINES(SIZE),
        undef_preds(Undefined),
        retract(bw_analysis),
        current_time(InitEnd),
	!,
        lfp(Alpha),
        current_time(BottomUpEnd),
        InitLapse is InitEnd-InitStart,
        BULapse is BottomUpEnd-InitEnd,
%        write('Success set: '),nl,
%        asem_display(Alpha,answer),
        current_time(AnnotationStart),
        init_bw,
        current_time(AnnotationEnd),
        AnnLapse is AnnotationEnd-AnnotationStart,
%        write('Initial Demand: '),nl,
%        asem_display(Alpha,demand),
        current_time(BWStart),        
%trace,
        gfp(Alpha),
%notrace,
        current_time(BWEnd),
        BWLapse is BWEnd-BWStart
%        ,write('Derived Demand: '),nl,
%        asem_display(Alpha,demand)
        .

gfp(Alpha) :-
        retract(do_gather_demand(Q,P,CE,NVarR,NVar)),
        demand(Q,Sigma), 
        asub_bw_unify_entry(Alpha,CE,Sigma,NVarR,NVar,Eta),
        ( asub_top(Alpha,NVar,Eta) -> true
	; demand_update_glb(Alpha,P,Eta)
	),
	!,
        gfp(Alpha).

gfp(Alpha) :-
        retract(do_call_bw(Q,P,CE,NVarR,NVar)),
        demand(Q,Sigma), 
        asub_bw_unify_entry(Alpha,CE,Sigma,NVarR,NVar,Eta),
        ( asub_top(Alpha,NVar,Eta) -> true
	; demand_update_glb(Alpha,P,Eta)
	),
	!,
        gfp(Alpha).

gfp(Alpha) :-
        retract(do_exit_bw(Q,P,P_,CE,NVarR,NVar)),
        ( 
           ans(Q,Answer), demand(P,Success)
        ->
           asub_bw_unify_exit(Alpha,CE,Answer,NVarR,Success,NVar,F3),
           ( asub_top(Alpha,NVar,F3) -> true
	   ; demand_update_glb(Alpha,P_,F3)
	   )
        ;
           true
        ), 
	!,
        gfp(Alpha).

gfp(Alpha) :-
        retract(do_bip_bw(Q,P,A,NVar)),
        (  
           demand(Q,Theta)
        ->
           asub_bw_bipai(Alpha,A,Theta,Zeta),
           ( asub_top(Alpha,NVar,Zeta) -> true
	   ; demand_update_glb(Alpha,P,Zeta)
	   )
        ;
           true
        ),
	!,
        gfp(Alpha).

gfp(_Alpha).


init_bw :- demand(P,_),add_tasks_bw(P),fail.
init_bw.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


/* The bottom up abstract semantics is computed by
   using the same algorithm as the top-down abstract
   semantics. The only difference between them is the
   initialization process: (1) There is no call edge
   and (2) The abstract substitution at the entry point
   of each clause is initialized to be the abstract
   identity substitution which is then propagated to
   update abstract substitutions at other program
   points. 

   We mark the part of the initialization process that
   are different for the top-down and the bottom-up
   abstract semantics by comments containing %%BU

*/

:- dynamic proc_size/3. 

add_clause((H:-B),SYMBTAB,VL,NVar,PtEntry,BS,Alpha) :-
    functor(H,PN,PA),
    add_defined(PN/PA),
    ( retract(proc_size(PN,PA,Rk)) -> I is Rk+1; I is 1 ), 
    assert(proc_size(PN,PA,I)),
    next_Pt(PtEntry),
    add_ans_clause(PN,PA,PtAns,BAns),
    tagterm(H,PA,HRenamed),
    mgu(HRenamed,BAns,Sub),
    asub_eq_constraints(Alpha,Sub,EC),
%%%%%
    gen_prop_tab_goal(B,NVar,PtEntry,PtExit,Alpha), 
    BS is PtExit-PtEntry,
    assert(select_cls(PtAns,PtEntry,PtExit,H,NVar)),
    ( 
        bw_analysis 
     -> 
        assert(ntt_bw(PtEntry,do_gather_demand(PtEntry,PtAns,EC,NVar,PA)))
     ;
        true
    ),
    add_cls_cidPt(PN,PA,I,BS,PtEntry,NVar),
    assert(ntt(PtExit,do_gather_ans(PtExit,PtAns,EC,NVar,PA))),
    assert(user_clause(PN,PA,I,PtEntry,(H:-B),SYMBTAB,VL,NVar,BS)),
    ( type_analysis(Alpha) -> collect_used_functors((H:-B)) ; true ).



gen_prop_tab_goal((A,B),NVar,Pt0,Pt1,Alpha) :-
    !,
    gen_prop_tab_goal(A,NVar,Pt0,PtA,Alpha),
    gen_prop_tab_goal(B,NVar,PtA,Pt1,Alpha).

gen_prop_tab_goal((A;B),NVar,Pt0,Pt1,Alpha) :-
    !,
    gen_prop_tab_goal(A,NVar,Pt0,PtA,Alpha),
    PtB0 is PtA+1,
    gen_prop_tab_goal(B,NVar,PtB0,PtB1,Alpha),
    Pt1 is PtB1+1,
    asub_compile(Alpha,true,Compiled),
    assert(ntt(Pt0,do_bip(Pt0,PtB0,Compiled,NVar))), 
                      % PtB0 can be reached from Pt0
    assert(ntt(PtA,do_bip(PtA,Pt1,Compiled,NVar))),  
                      % Pt1 can be reached from PtA
    assert(ntt(PtB1,do_bip(PtB1,Pt1,Compiled,NVar))),
                      % Pt1 can be reached from PtB1
    ( 
        bw_analysis 
     -> 
        asub_bw_compile(Alpha,true,Compiled_bw),
        assert(ntt_bw(PtB0,do_bip_bw(PtB0,Pt0,Compiled_bw,NVar))), 
        assert(ntt_bw(Pt1,do_bip_bw(Pt1,PtA,Compiled_bw,NVar))), 
        assert(ntt_bw(Pt1,do_bip_bw(Pt1,PtB1,Compiled_bw,NVar)))
     ;
        true
    ).

gen_prop_tab_goal(\+(A),NVar,Pt0,Pt1,Alpha) :-
    !,
    gen_prop_tab_goal(A,NVar,Pt0,PtA,Alpha),
    /* Edges ending at PtA could be removed for efficiency. 
    */
    Pt1 is PtA+1,
    asub_compile(Alpha,true,Compiled),
    assert(ntt(Pt0,do_bip(Pt0,Pt1,Compiled,NVar))), 
                     % Pt1 can be reached from Pt0
    ( 
        bw_analysis 
     -> 
        asub_bw_compile(Alpha,true,Compiled_bw),
        assert(ntt_bw(Pt1,do_bip_bw(Pt1,Pt0,Compiled_bw,NVar)))
     ;
        true
    ).

gen_prop_tab_goal(not(A),NVar,Pt0,Pt1,Alpha) :-
    !,
    gen_prop_tab_goal(A,NVar,Pt0,PtA,Alpha),
    /* All edges ending at PtA could be removed to improve efficiency. */
    Pt1 is PtA+1,
    asub_compile(Alpha,true,Compiled),
    assert(ntt(Pt0,do_bip(Pt0,Pt1,Compiled,NVar))), % Pt1 can be reached from Pt0
    ( 
        bw_analysis 
     -> 
        asub_bw_compile(Alpha,true,Compiled_bw),
        assert(ntt_bw(Pt1,do_bip_bw(Pt1,Pt0,Compiled_bw,NVar)))
     ;
        true
    ).


gen_prop_tab_goal(findall(X,A,Xs),NVar,Pt0,Pt1,Alpha) :-
    !,
    gen_prop_tab_goal(A,NVar,Pt0,PtA,Alpha),
    Pt1 is PtA+1,
    ( variable(Xs) ->  true ; write(findall(X,A,Xs)),nl,halt),
    asub_compile(Alpha,Xs=[X],Compiled), %Xs must be a variable
    assert(ntt(PtA,do_bip(PtA,Pt1,Compiled,NVar))), 
                          % Pt1 can be reached from PtA
    ( 
        bw_analysis 
     -> 
        asub_bw_compile(Alpha,Xs=[X],Compiled_bw),
        assert(ntt_bw(Pt1,do_bip_bw(Pt1,PtA,Compiled_bw,NVar)))
     ;
        true
     ). 
      

gen_prop_tab_goal(setof(X,A,Xs),NVar,Pt0,Pt1,Alpha) :-
    !,
    gen_prop_tab_goal(A,NVar,Pt0,PtA,Alpha),
    Pt1 is PtA+1,
    ( variable(Xs) ->  true ; write(findall(X,A,Xs)),nl,halt),
    asub_compile(Alpha,Xs=[X],Compiled), %Xs must be a variable
    assert(ntt(PtA,do_bip(PtA,Pt1,Compiled,NVar))), 
                          % Pt1 can be reached from PtA
    ( 
        bw_analysis 
     -> 
        asub_bw_compile(Alpha,Xs=[X],Compiled_bw),
        assert(ntt_bw(Pt1,do_bip_bw(Pt1,PtA,Compiled_bw,NVar)))
     ;
        true
    ). 
    

gen_prop_tab_goal(bagof(X,A,Xs),NVar,Pt0,Pt1,Alpha) :-
    !,
    gen_prop_tab_goal(A,NVar,Pt0,PtA,Alpha),
    Pt1 is PtA+1,
    ( variable(Xs) ->  true ; write(findall(X,A,Xs)),nl,halt),
    asub_compile(Alpha,Xs=[X],Compiled), %Xs must be a variable
    assert(ntt(PtA,do_bip(PtA,Pt1,Compiled,NVar))),
    ( 
        bw_analysis 
     -> 
        asub_bw_compile(Alpha,Xs=[X],Compiled_bw), 
        assert(ntt_bw(Pt1,do_bip_bw(Pt1,PtA,Compiled_bw,NVar)))
     ;
        true
    ).

gen_prop_tab_goal((A->B),NVar,Pt0,Pt1,Alpha) :-
    !,
    gen_prop_tab_goal(A,NVar,Pt0,PtA,Alpha),
    PtB0 is PtA+1,
    gen_prop_tab_goal(B,NVar,PtB0,Pt1,Alpha),
    asub_compile(Alpha,true,Compiled),
    assert(ntt(PtA,do_bip(PtA,PtB0,Compiled,NVar))),
                     % PtB0 can be reached from PtA
    ( 
        bw_analysis 
     -> 
        asub_bw_compile(Alpha,true,Compiled_bw),
        assert(ntt_bw(PtB0,do_bip_bw(PtB0,PtA,Compiled_bw,NVar)))
     ;
        true
    ).


gen_prop_tab_goal(true,_NVar,Pt0,Pt0,_Alpha) :-
    !.

gen_prop_tab_goal(BIJRaw,NVar,Pt0,Pt1,_Alpha) :-
    peel_meta(BIJRaw,AIJ),
    Pt1 is Pt0+1,
    functor(AIJ,PN,PA),
    add_ans_clause(PN,PA,PtAns,_BAns),
    add_invoked(PN/PA),
    assert(call_proc(Pt0,PtAns,AIJ,NVar)).

:- dynamic do_bip/4, 
           do_exit/6, 
           do_call_fw/5,
           do_gather_ans/5.
:- dynamic do_bip_bw/4,
           do_exit_bw/6,
           do_gather_demand/5,
           do_call_bw/5.

:- dynamic next_Pt/1, 
           ans_next_Pt/1,
           cls_cidPt/6,
           ans_pidPt/4.

next_Pt(1). % dummy points removed
ans_next_Pt(100001).


add_cls_cidPt(PN,PA,I,BS,Pt,NVar) :- 
    (
       cls_cidPt(PN,PA,I,Pt,BS,NVar)
    -> 
       true
    ;
       retract(next_Pt(Pt)),
       Pt1 is Pt+BS+1, % dummy points removed
       assert(next_Pt(Pt1)),
       assert(cls_cidPt(PN,PA,I,Pt,BS,NVar))
    ).
    

del_ans_clause(PN,PA) :- 
    retractall(ans_pidPt(PN,PA,_Pt,_Body)).

add_ans_clause(PN,PA,Pt,Body) :- 
    ans_pidPt(PN,PA,Pt,Body),
    !.

add_ans_clause(PN,PA,Pt,Body) :- 
       functor(Body,PN,PA),
       numbervars(Body,1,_),
       retract(ans_next_Pt(Pt)),
       Pt1 is Pt+1, % only one point is needed
       assert(ans_next_Pt(Pt1)),
       assert(ans_pidPt(PN,PA,Pt,Body)).
       
init_bu_bw(Pg,Alpha) :-
             (
	        type_analysis(Alpha) 
	      -> 
	        ensure_loaded('Domain/typedef'),
	        typedef_clear
	      ; 
	        true
	     ),
             retractall(user_clause(_,_,_,_,_,_,_,_,_)),
             retractall(proc_size(_,_,_)),
             retractall(defined_predicate(_)), 
             retractall(invoked_predicate(_)),
             retractall(used_functor(_)),
             retractall(next_Pt(_)),
             assert(next_Pt(1)), % dummy points removed
             retractall(ans_next_Pt(_)),
             assert(ans_next_Pt(100001)), 
             retractall(cls_cidPt(_,_,_,_,_,_)),
             retractall(ans_pidPt(_,_,_,_)),
             ntt_new, sem_new, 
             load_program(Pg,BST,Alpha),
             (
	         type_analysis(Alpha)
	      ->
	         ( setof(F,used_functor(F),Fs) -> true ; Fs=[] ),
	         typedef_add_default(Fs)
	      ;
	         true
	     ),
             gen_bu_flow_arcs(Alpha),
             retractall(select_cls(_,_,_,_,_)),
	     retractall(call_proc(_,_,_,_)),
	     init_agenda(BST), 
             init_cnt_sem(BST,Alpha).


init_td(Pg,(Goal,InputASubX,Tab),Alpha) :-
             (
	        type_analysis(Alpha) 
	      -> 
	        ensure_loaded('Domain/typedef'),
       	        ensure_loaded(library(ordsets)),
	        typedef_clear
	      ; 
	        true
	     ),
             retractall(user_clause(_,_,_,_,_,_,_,_,_)),
             retractall(proc_size(_,_,_)),
             retractall(defined_predicate(_)), 
             retractall(used_functor(_)), 
             retractall(invoked_predicate(_)), 
             retractall(next_Pt(_)),
             assert(next_Pt(1)), % dummy points removed
             retractall(ans_next_Pt(_)),
             assert(ans_next_Pt(100001)), 
             retractall(cls_cidPt(_,_,_,_,_,_)),
             retractall(ans_pidPt(_,_,_,_)),
             ntt_new, sem_new, 
             load_program(Pg,_,Alpha),
             vars(Goal,Vs), length(Vs,NVar),
             add_clause(('$Goal':- Goal),Tab,Vs,NVar,PtStart,_BS,Alpha),
             (
	         type_analysis(Alpha)
	      ->
	         ( setof(Ar-Fn,(used_functor(Ar-Fn),Ar>0),Fs)->true;Fs=[]),
	         typedef_add_default(Fs),
	         typedef_functors(DFs),
                 ord_subtract(Fs,DFs,Untyped),
	         ( Untyped =[] -> true; write('Untyped: '), write(Untyped),nl )

	      ;
	         true
	     ),
             gen_td_flow_arcs(Alpha),
             retractall(select_cls(_,_,_,_,_)),
	     retractall(call_proc(_,_,_,_)),
             add_tasks(PtStart),
	     asub_internalize(Alpha,InputASubX,ASub),
	     assert(sem(PtStart,ASub)).


/* BU: init_agenda(+BST,+Alpha) initializes agenda for
   BU analysis. */

init_agenda([]).
init_agenda([(Pt,BS,_NVar)|BST]) :-
    (
      BS=:=0 -> add_tasks(Pt); 
      ntt(Pt,do_bip(P,Q,CE,NVar)) -> add_task(do_bip(P,Q,CE,NVar)) ;
      true
    ),
    init_agenda(BST).


/* %%BU: init_cnt_sem(+BST)
   initialise SEM. NTT is a table mapping a point into
   a set of new tasks generated by a change at that
   point. SEM is a table mapping a point into the
   bottom abstract substitution over the variables of
   the clause to which the point belongs. */

init_cnt_sem([],_Alpha).
init_cnt_sem([(Pt,_BS,NVar)|BST],Alpha) :-
    asub_ident(Alpha,NVar,Id),
    assert(sem(Pt,Id)),   
    init_cnt_sem(BST,Alpha). 


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

asem_display(Alpha,answer) :- 
    (
     ans_pidPt(_,_,Pt,Body), \+ Body = '$Goal', 
     Body =.. [_|VL],length(VL,NVar),
     ( ans(Pt,ASub) -> true; asub_bot(Alpha,NVar,ASub) ), 
     asub_ident(Alpha,NVar,Id),
     asub_externalize(Alpha,Id,IdX),
     asub_externalize(Alpha,ASub,ASubX),
     identity_symbol_tab(VL,Tab), % readable
     cls_ShowANNO( ('$answer'(Body):-Body), Tab, [(1,(Pt - IdX)),(2,ASubX)]),
     fail
    ;
     true
    ),nl,nl,
    display_user_clauses(Alpha,answer), nl,nl,
    asem_map(Alpha,answer).

asem_display(Alpha,demand) :- 
    (
     ans_pidPt(_,_,Pt,Body),
     Body =.. [_|VL], length(VL,NVar),
     asub_top(Alpha,NVar,Top),
     ( demand(Pt,ASub) -> true; ASub=Top ), 
     asub_externalize(Alpha,ASub,ASubX),
     asub_externalize(Alpha,Top,TopX),
     identity_symbol_tab(VL,Tab), % readable
     cls_ShowANNO(('$demand'(Body):-Body), Tab, [(1,(Pt - ASubX)),(2,TopX)]),
     fail
    ;
     true
    ),nl,nl,
    display_user_clauses(Alpha,demand), nl,nl,
    asem_map(Alpha,demand).


display_user_clauses(Alpha,Type) :- 
    user_clause(_PN,_PA,_I,PtEntry,(H:-B),SYMBTAB,_VL,NVar,BS),
    L1 is BS+1,
    integers(1,L1,Row),
    getRow(Row,PtEntry,ANNS,NVar,Alpha,Type),
    cls_ShowANNO((H:-B),SYMBTAB, ANNS),
    fail.
display_user_clauses(_Alpha,_Type).

integers(Min,Max,[]) :-
    Min > Max.
integers(Min,Max,[Min|Ints1]) :-
    Min =< Max,
    Min1 is Min+1,
    integers(Min1,Max,Ints1).

getRow([],_,[],_NVar,_Alpha,_Type).
getRow([PtY|Row],Pt,[(PtY,(Pt1-ASubX))|ANNS],NVar,Alpha,answer) :-
    Pt1 is Pt+PtY-1,
    ( sem(Pt1,ASub) -> true; asub_bot(Alpha,NVar,ASub) ),
    asub_externalize(Alpha,ASub,ASubX),
    getRow(Row,Pt,ANNS,NVar,Alpha,answer).
getRow([PtY|Row],Pt,[(PtY,(Pt1-ASubX))|ANNS],NVar,Alpha,demand) :-
    Pt1 is Pt+PtY-1,
    ( demand(Pt1,ASub) -> true; asub_top(Alpha,NVar,ASub)),
    asub_externalize(Alpha,ASub,ASubX),
    getRow(Row,Pt,ANNS,NVar,Alpha,demand).


identity_symbol_tab([],[]).
identity_symbol_tab([X|Xs],[X/X|Tab]) :- 
    identity_symbol_tab(Xs,Tab).


asem_map(Alpha,answer) :-
    !,
    ( setof(A, Pt^(sem(Pt,A);ans(Pt,A)), As) -> true ; As=[] ),
    asub_list_print(Alpha,As),nl.

asem_map(Alpha,demand) :-
    !,
    ( setof(A, Pt^demand(Pt,A), As) -> true ; As=[] ),
    asub_list_print(Alpha,As),nl.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


:- dynamic ntt/2, ntt_bw/2.

ntt_new :- retractall(ntt(_,_)), 
           retractall(ntt_bw(_,_)).



add_task(do_gather_ans(P,Q,E,NVarR,NVar)) :- 
    do_gather_ans(P,Q,_,_,_) -> true ; asserta(do_gather_ans(P,Q,E,NVarR,NVar)).
add_task(do_exit(P,Q,Q1,E,NVarR,NVar)) :- 
    do_exit(P,Q,_,_,_,_) -> true ; asserta(do_exit(P,Q,Q1,E,NVarR,NVar)).
add_task(do_bip(P,Q,E,NVar)) :- 
    do_bip(P,Q,_,_) -> true ; asserta(do_bip(P,Q,E,NVar)).

add_task(do_call_fw(P,Q,E,NVarR,NVar)) :- 
    do_call_fw(P,Q,_,_,_) -> true ; assertz(do_call_fw(P,Q,E,NVarR,NVar)).

add_tasks(P) :- 
    ntt(P,T), add_task(T),fail.
add_tasks(_).


add_task_bw(do_gather_demand(P,Q,E,NVarR,NVar)) :- 
    do_gather_demand(P,Q,_,_,_) -> true 
  ; asserta(do_gather_demand(P,Q,E,NVarR,NVar)).
add_task_bw(do_exit_bw(P,Q,Q1,E,NVarR,NVar)) :- 
    do_exit_bw(P,Q,_,_,_,_) -> true 
  ; asserta(do_exit_bw(P,Q,Q1,E,NVarR,NVar)).
add_task_bw(do_bip_bw(P,Q,E,NVar)) :- 
    do_bip_bw(P,Q,_,_) -> true 
  ; asserta(do_bip_bw(P,Q,E,NVar)).
add_task_bw(do_call_bw(P,Q,E,NVarR,NVar)) :- 
    do_call_bw(P,Q,_,_,_) -> true 
  ; asserta(do_call_bw(P,Q,E,NVarR,NVar)).

add_tasks_bw(P) :- 
    ntt_bw(P,T), add_task_bw(T),fail.
add_tasks_bw(_).




:- dynamic sem/2,ans/2,demand/2. 

sem_new :-  
    retractall(sem(_,_)),
    retractall(ans(_,_)),
    retractall(demand(_,_)).


sem_update_lub(Alpha,P,Zeta) :- 
        sem(P,Eta), % Eta is not FALSE
        !,
        asub_join(Alpha,Zeta,Eta,Eta1), 
        (
                asub_leq(Alpha,Eta1,Eta) 
	->
		true
	;
                retract(sem(P,_)), assert(sem(P,Eta1)),
		add_tasks(P)
	).
sem_update_lub(_Alpha,P,Zeta) :- % sem(P) was bottom 
    assert(sem(P,Zeta)),
    add_tasks(P).

ans_update_lub(Alpha,P,Zeta) :- 
        ans(P,Eta), % Eta is not FALSE
        !,
        asub_join(Alpha,Zeta,Eta,Eta1), 
        (
                asub_leq(Alpha,Eta1,Eta) 
	->
		true
	;
                retract(ans(P,_)), assert(ans(P,Eta1)),
		add_tasks(P)
	).
ans_update_lub(_Alpha,P,Zeta) :- % ans(P) was bottom 
    assert(ans(P,Zeta)),
    add_tasks(P).

demand_update_glb(Alpha,P,Zeta) :- 
        demand(P,Eta), % Eta is not TRUE
        !,
        asub_meet(Alpha,Zeta,Eta,Eta1), 
        (
                asub_leq(Alpha,Eta,Eta1) 
	->
		true
	;
                retract(demand(P,_)), assert(demand(P,Eta1)),
		add_tasks_bw(P)
	).
demand_update_glb(_Alpha,P,Zeta) :- % ans(P) was TRUE 
    assert(demand(P,Zeta)),
    add_tasks_bw(P).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


current_time(MiniSec) :-  statistics(runtime,[MiniSec,_]).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

/* load_program(+File,-BST,+Alpha) reads a program from
   +File into program the database. Type declarations
   are collected for type analyses. Default type
   definitions are added for unary and binary
   arithmetic operations. */



load_program(File,BST,Alpha) :-
    prolog_flag(fileerrors,Old,off), %turn off fileerrors
    ( 
        open(File,read,Stream)
    ->
        load_program_1(Stream,BST,Alpha),
        close(Stream)
    ;   
        nl,write(File),write(' does not exist!'),nl
    ),
    prolog_flag(fileerrors,_,Old).

load_program_1(Stream,BST,Alpha) :-
    ( 
         load_one_clause(Stream,PtEntry,BS,NVar,Alpha)
    -> 
         BST=[(PtEntry,BS,NVar)|BST1],
         load_program_1(Stream,BST1,Alpha)
    ;
         BST=[]
    ).


/* load_one_clause(+Stream,PtEntry,BS,NVar, Alpha)
   reads a clause from an input stream +Stream.  If the
   clause is a directive, execute it. Otherwise, add it
   into program database or type definition database
   accordingly. It fails when the end_of_file is
   reached */


load_one_clause(Stream,PtEntry,BS,NVar,Alpha) :-
    read_term(Stream,Term,[variable_names(NL)]),
    Term \= end_of_file,
    (
        Term =(:- type(::=(LHS,RHS))) 
    ->
        (type_analysis(Alpha) -> type_def_expansion(LHS,RHS); true),
        load_one_clause(Stream,PtEntry,BS,NVar,Alpha)
    ;
	Term = (:- Call)
    ->
	call(Call),
        load_one_clause(Stream,PtEntry,BS,NVar,Alpha)
    ;
        (
            Term = (H:-B) 
        -> 
            true
        ;
            H = Term,
            B = true
        ),
            numbervars(((H,B),NL),1,NextVar),
            NVar is NextVar-1,
            vars((H,B),VL),
            nameListToSub(VL,NL,Sub), 
            add_clause((H:-B),Sub,VL,NVar,PtEntry,BS,Alpha)
     ).

    
collect_used_functors((H:-B)) :- 
    collect_used_functors_in_atom(H),
    collect_used_functors_in_body(B).

collect_used_functors_in_atom(Atom) :-
    Atom =.. [_| Terms],
    collect_used_functors_in_term_list(Terms).

collect_used_functors_in_body((A,B)) :-
    !,
    collect_used_functors_in_body(A),
    collect_used_functors_in_body(B).
collect_used_functors_in_body((A->B)) :-
    !,
    collect_used_functors_in_body(A),
    collect_used_functors_in_body(B).
collect_used_functors_in_body((A;B)) :-
    !,
    collect_used_functors_in_body(A),
    collect_used_functors_in_body(B).
collect_used_functors_in_body(\+(A)) :-
    !,
    collect_used_functors_in_body(A).
collect_used_functors_in_body(not(A)) :-
    !,
    collect_used_functors_in_body(A).
collect_used_functors_in_body(findall(P,Call,Ps)) :-
    !,
    collect_used_functors_in_body(Call),
    collect_used_functors_in_atom((P=Ps)).
collect_used_functors_in_body(setof(P,Call,Ps)) :-
    !,
    collect_used_functors_in_body(Call),
    collect_used_functors_in_atom((P=Ps)).
collect_used_functors_in_body(bagof(P,Call,Ps)) :-
    !,
    collect_used_functors_in_body(Call),
    collect_used_functors_in_atom((P=Ps)).
collect_used_functors_in_body(A):-
    collect_used_functors_in_atom(A). 

collect_used_functors_in_term_list([]).
collect_used_functors_in_term_list([Term|Terms])	:-
    collect_used_functors_in_term(Term),
    collect_used_functors_in_term_list(Terms).

collect_used_functors_in_term(Term)	:-
    (
        variable(Term)
    -> 
        true
    ;  
        functor(Term,Functor,Arity),
	Term =.. [Functor|Subterms],
        add_used_functor(Arity-Functor),
	collect_used_functors_in_term_list(Subterms)
    ).


type_def_expansion(Tp, (RHS1;RHS)) :- 
   !,
    RHS1 =.. [FN|Tps],
    type_clause_body(Tps,Args,Body),
    Tm =.. [FN|Args],
    typedef_add((type(Tm,Tp):-Body)),
    type_def_expansion(Tp,RHS).
 
type_def_expansion(Tp,RHS1) :- 
   !,
    RHS1 =.. [FN|Tps],
    type_clause_body(Tps,Args,Body),
    Tm =.. [FN|Args],
    typedef_add((type(Tm,Tp):-Body)).
    
type_clause_body([], [], true) :-
    !.
type_clause_body([T],[X],type(X,T)) :-
    !.
type_clause_body([T,T2|Ts],[X|Xs], (type(X,T),Body)) :- 
    type_clause_body([T2|Ts],Xs,Body).






/* cls_ShowANNO(+HB,+Sub,+ANNS) displays the clause HB
   with +Sub providing external names for variables in
   HB and ANNS the annotations associated with program
   points in the clause. */

cls_ShowANNO(HB,Sub,ANNS) :-
    sub_apply(Sub,HB,HB1),
    sub_apply(Sub,ANNS,ANNS1),
    impl_anno_SHOW(HB1,ANNS1).


impl_anno_SHOW((A:-true),ANNS) :-
    !,
    tab(4),write(A),write('.'),nl,
    member((1,Inv),ANNS),
    tab(12), write('%'), write(Inv), nl.

impl_anno_SHOW((A:-B),ANNS) :-
    tab(4),write(A), write(' :-'),nl,
    member((1,Inv),ANNS),
    tab(12), write('%'), write(Inv), nl,
    body_anno_SHOW(B,1,_,ANNS,'.',8). 


body_anno_SHOW((A,B),K0,K1,ANNS,Terminator,Indent) :- 
    !, 
    body_anno_SHOW(A,K0,KA,ANNS,',',Indent),
    body_anno_SHOW(B,KA,K1,ANNS,Terminator,Indent).  

body_anno_SHOW((C->A;B),K0,K1,ANNS,Terminator,Indent) :- 
    !, 
    tab(Indent), write('('),nl,
    Indent1 is Indent+4,
    IndentInv1 is Indent1+4,
    IndentInv is Indent+4,
    body_anno_SHOW(C,K0,KC,ANNS,' ',Indent1),
    tab(Indent), write('->'),nl,
    KA0 is KC+1,
    (
        member((KA0,InvA0),ANNS) 
    ->
        tab(IndentInv1), write('%'), write(InvA0), nl
    ;
        true
    ),
    body_anno_SHOW(A,KA0,KA,ANNS,' ',Indent1),
    tab(Indent), write(';'),nl,
    KB0 is KA+1,
    (
        member((KB0,InvB0),ANNS) 
    ->
        tab(IndentInv1), write('%'), write(InvB0), nl
    ;
        true
    ),
    body_anno_SHOW(B,KB0,KB1,ANNS,' ',Indent1),
    tab(Indent), write(')'), write(Terminator), nl,
    K1 is KB1+1,
    (
        member((K1,Inv1),ANNS) 
    ->
        tab(IndentInv), write('%'), write(Inv1), nl
    ;
        true
    ).  

body_anno_SHOW((C->A),K0,K1,ANNS,Terminator,Indent) :- 
    !, 
    tab(Indent), write('('),nl,
    Indent1 is Indent+4,
    IndentInv1 is Indent1+4,
    body_anno_SHOW(C,K0,KC,ANNS,' ',Indent1),
    tab(Indent), write('->'),nl,
    KA0 is KC+1,
    (
        member((KA0,InvA0),ANNS) 
    ->
        tab(IndentInv1), write('%'), write(InvA0), nl
    ;
        true
    ),
    body_anno_SHOW(A,KA0,K1,ANNS,' ',Indent1),
    tab(Indent), write(')'), write(Terminator), nl.

body_anno_SHOW((A;B),K0,K1,ANNS,Terminator,Indent) :- 
    !, 
    tab(Indent),
    write('('),nl,
    Indent1 is Indent+4,
    IndentInv1 is Indent1+4,
    IndentInv is Indent+4,
    body_anno_SHOW(A,K0,KA,ANNS,' ',Indent1),
    tab(Indent), write(';'),nl,
    KB0 is KA+1,
    (
        member((KB0,Inv),ANNS) 
    ->
        tab(IndentInv1), write('%'),
        write(Inv), 
        nl
    ;
        true
    ),
    body_anno_SHOW(B,KB0,KB1,ANNS,' ',Indent1),
    tab(Indent), write(')'), write(Terminator), nl,
    K1 is KB1+1,
    (
        member((K1,Inv1),ANNS) 
    ->
        tab(IndentInv), write('%'),
        write(Inv1), 
        nl
    ;
        true
    ).  

body_anno_SHOW(\+(A),K0,K1,ANNS,Terminator,Indent) :- 
    !, 
    tab(Indent),
    write('\\+ ('),nl,
    Indent1 is Indent+4,
    IndentInv is Indent+4,
    body_anno_SHOW(A,K0,KA,ANNS,' ',Indent1),
    tab(Indent), write('   )'), write(Terminator), nl,
    K1 is KA+1,
    (
        member((K1,Inv1),ANNS) 
    ->
        tab(IndentInv), write('%'), write(Inv1), nl
    ;
        true
    ).  


body_anno_SHOW(not(A),K0,K1,ANNS,Terminator,Indent) :- 
    !, 
    tab(Indent),
    write('not('),nl,
    Indent1 is Indent+4,
    IndentInv is Indent+4,
    body_anno_SHOW(A,K0,KA,ANNS,' ',Indent1),
    tab(Indent), write('   )'), write(Terminator), nl,
    K1 is KA+1,
    (
        member((K1,Inv1),ANNS) 
    ->
        tab(IndentInv), write('%'), write(Inv1), nl
    ;
        true
    ).  


body_anno_SHOW(findall(P,A,Ps),K0,K1,ANNS,Terminator,Indent) :- 
    !, 
    tab(Indent),write('findall( '), write(P), write(','), nl,
    Indent1 is Indent+4,
    IndentInv is Indent+4,
    body_anno_SHOW(A,K0,KA,ANNS,',',Indent1),
    tab(Indent),write(Ps),write(')'), write(Terminator), nl,
    K1 is KA+1,
    (
        member((K1,Inv1),ANNS) 
    ->
        tab(IndentInv), write('%'), write(Inv1), nl
    ;
        true
    ).  


body_anno_SHOW(setof(P,A,Ps),K0,K1,ANNS,Terminator,Indent) :- 
    !, 
    tab(Indent),write('setof( '), write(P), write(','), nl,
    Indent1 is Indent+4,
    IndentInv is Indent+4,
    body_anno_SHOW(A,K0,KA,ANNS,',',Indent1),
    tab(Indent),write(Ps),write(')'), write(Terminator), nl,
    K1 is KA+1,
    (
        member((K1,Inv1),ANNS) 
    ->
        tab(IndentInv), write('%'), write(Inv1), nl
    ;
        true
    ).  


body_anno_SHOW(bagof(P,A,Ps),K0,K1,ANNS,Terminator,Indent) :- 
    !, 
    tab(Indent),write('bagof( '), write(P), write(','), nl,
    Indent1 is Indent+4,
    IndentInv is Indent+4,
    body_anno_SHOW(A,K0,KA,ANNS,',',Indent1),
    tab(Indent),write(Ps),write(')'), write(Terminator), nl,
    K1 is KA+1,
    (
        member((K1,Inv1),ANNS) 
    ->
        tab(IndentInv), write('%'), write(Inv1), nl
    ;
        true
    ).  


body_anno_SHOW(true,K,K,_ANNS, Terminator,Indent) :-
    !, 
    tab(Indent), write(true), write(Terminator), nl.

body_anno_SHOW(A,K0,K1,ANNS, Terminator,Indent) :- 
    tab(Indent), write(A), write(Terminator), nl,
    K1 is K0+1,
    (
        member((K1,Inv),ANNS) 
    ->
        IndentInv is Indent+4,
        tab(IndentInv),write('%'),
        write(Inv),nl
    ;
        true
    ).


/* type_analysis(+Alpha) = Alpha is a type abstraction. In a type */

type_analysis(rtype). 
type_analysis(monotype). 

    tab(0). %tab is missing from Sicstus 4.0.3!
    tab(I) :- I>0, write(' '), I1 is I-1, tab(I1).
