:- module(store,[   pg_LOAD/2,
	            pg_FUNCS/2,
        	    pg_PIDS/2,
        	    pg_PROC/3,
	            pg_AddPROC/3,
                    pg_LINES/2,
	            proc_SIZE/2,
	            proc_PID/2,
	            proc_CLS/3,
                    pg_Undefined/2,
	            cls_SIZE/2,
	            cls_IMPLICATION/2,
	            cls_SYMBOLTABLE/2,
		    cls_VL/2,
	            cls_ShowANNO/2,
        	    cls_CONS/3,
                    query_EMPTY/1,
                    query_CLS/2,
                    query_GOAL/2,
                    query_ASUB/2,
                    query_ALPHA/2,
                    query_CONS/4
         ]).

:- use_module(basics, [
	bagof0/3, 
	vars/2,
        bip/1,
        peel_meta/2,
	variable/1,
        intern/3, 
	sub_apply/3, 
	sub_inverse/2, 
	mgu/3,
        sub_domain/2,
        nameListToSub/3 ]).

:- use_module(typedef, [
        typedef_add/1,
        typedef_clear/0,
        typedef_add_default/1
   ]).

:- use_module('ordsets.pl', [ ord_add_element/3,ord_subtract/3 ]).

:- op(1150,fx,type).
:- op(1125,xfx, ::=).

lit_called_pids(L,Defined,PIDS0,PIDS) :-
    peel_meta(L,Atom),
    functor(Atom,PN,PA),
    (
        \+ member(PN/PA,Defined),bip(Atom)
    ->
        PIDS=PIDS0
    ;
        PIDS=[PN/PA|PIDS0]
    ).

body_called_pids((A,B),Defined,PIDSAcc,PIDS) :-
    !,
    body_called_pids(A,Defined,PIDSAcc,PIDSAcc1),
    body_called_pids(B,Defined,PIDSAcc1,PIDS).
body_called_pids((A->B),Defined,PIDSAcc,PIDS) :-
    !,
    body_called_pids(A,Defined,PIDSAcc,PIDSAcc1),
    body_called_pids(B,Defined,PIDSAcc1,PIDS).
body_called_pids((A;B),Defined,PIDSAcc,PIDS) :-
    !,
    body_called_pids(A,Defined,PIDSAcc,PIDSAcc1),
    body_called_pids(B,Defined,PIDSAcc1,PIDS).
body_called_pids(\+(A),Defined,PIDSAcc,PIDS) :-
    !,
    body_called_pids(A,Defined,PIDSAcc,PIDS).
body_called_pids(not(A),Defined,PIDSAcc,PIDS) :-
    !,
    body_called_pids(A,Defined,PIDSAcc,PIDS).
body_called_pids(findall(_,A,_),Defined,PIDSAcc,PIDS) :-
    !,
    body_called_pids(A,Defined,PIDSAcc,PIDS).
body_called_pids(setof(_,A,_),Defined,PIDSAcc,PIDS) :-
    !,
    body_called_pids(A,Defined,PIDSAcc,PIDS).
body_called_pids(bagof(_,A,_),Defined,PIDSAcc,PIDS) :-
    !,
    body_called_pids(A,Defined,PIDSAcc,PIDS).
body_called_pids(A,Defined,PIDSAcc,PIDS) :-
    !,
    lit_called_pids(A,Defined,PIDSAcc,PIDS).


cls_called_pids(C,Defined,PIDS0,PIDS) :-
    cls_IMPLICATION(C,(_:-B)),
    body_called_pids(B,Defined,PIDS0,PIDS).

proc_called_pids(proc(_,_,CLS),Defined,PIDS) :-
    proc_called_pids_1(CLS,Defined,[],PIDS).

proc_called_pids_1([],_Defined,PIDS,PIDS).
proc_called_pids_1([CL|CLS],Defined,PIDS0,PIDS) :-
    cls_called_pids(CL,Defined,PIDS0,PIDS1),
    proc_called_pids_1(CLS,Defined,PIDS1,PIDS).

proc_CallGraph(Proc,Defined,Graph) :-
    proc_PID(Proc,Caller),
    proc_called_pids(Proc,Defined,Callees),
    bagof0(invokes(Caller,Callee),member(Callee,Callees),Graph).

pg_CallGraph(prog(Procs),Defined,Graph) :-
    pg_CallGraph_1(Procs,Defined,Graph).

pg_CallGraph_1([],_,[]).
pg_CallGraph_1([Proc|Procs],Defined,Graph) :-
    proc_CallGraph(Proc,Defined,Graph0),
    append(Graph0,Graph1,Graph),
    pg_CallGraph_1(Procs,Defined,Graph1).


pg_Undefined(Pg,PIDS) :-
    pg_PIDS(Pg,Callers),
    sort(Callers,Defined),
    pg_CallGraph(Pg,Defined,Graph),
    bagof0(Callee,Y^member(invokes(Y,Callee),Graph),Callees),
    sort(Callees,Called),
    ord_subtract(Called,Defined,PIDS).


pg_LINES(prog(Procs),Lines) :-
    pg_LINES_1(Procs,0,Lines).

pg_LINES_1([],Lines,Lines).
pg_LINES_1([Proc|Procs],Lines0,Lines) :-
    proc_LINES(Proc,Size),
    Lines1 is Lines0 + Size +1,
    pg_LINES_1(Procs,Lines1,Lines).



/*pg_PIDS(+Pg,-DescriptorsOfDefinedPredicates) */

pg_PIDS(prog(ProcLst),DesciptorLst) :-
    pg_PIDS_1(ProcLst,DesciptorLst).
pg_PIDS_1([],[]).
pg_PIDS_1([proc(PN,PA,_)|ProcLst],[PN/PA|DesciptorLst]) :-
    pg_PIDS_1(ProcLst,DesciptorLst).
    

% predicates for object "pred"

proc_SIZE(proc(_,_,CLS),Size) :-
    length(CLS,Size).

proc_LINES(proc(_,_,CLS),Lines) :-
    proc_LINES_1(CLS,0,Lines).

proc_LINES_1([],Lines,Lines).
proc_LINES_1([CL|CLS],Lines0,Lines) :-
    cls_SIZE(CL,Size),
    Lines1 is Lines0 + Size +1,
    proc_LINES_1(CLS,Lines1,Lines).



proc_PID(proc(PN,PA,_),PN/PA).

proc_CLS(proc(_,_,CLS),Rk,Cl) :-
    proc_CLS(CLS,1,Rk,Cl).

proc_CLS([Cl|_],Rk,Rk,Cl). 
proc_CLS([_|CLS],N,Rk,Cl) :-
    N < Rk,
    N1 is N+1,
    proc_CLS(CLS,N1,Rk,Cl).

% predicates for object "Cl"

cls_IMPLICATION('$cl'(IMPL,_,_,_),IMPL).
cls_SYMBOLTABLE('$cl'(_,SYMBTAB,_,_),SYMBTAB).
cls_SIZE('$cl'(_,_,_,SIZE),SIZE).
cls_VL('$cl'(_,_,VL,_),VL).

cls_CONS((H:-B),SYMBTAB,'$cl'((H:-B),SYMBTAB,VL,SIZE)) :-
    sub_domain(SYMBTAB,VL),
    body_size(B,SIZE).
cls_CONS((H:-B),SYMBTAB,VL,'$cl'((H:-B),SYMBTAB,VL,SIZE)) :-
    body_size(B,SIZE).

/*cls_ShowANNO(+Cl,+ANNOS) prints clause Cl in its mnemonic form*/
 
cls_ShowANNO(C,ANNS) :-
    cls_IMPLICATION(C,HB),
    cls_SYMBOLTABLE(C,Sub),
    sub_apply(Sub,HB,HB1),
    sub_apply(Sub,ANNS,ANNS1),
    impl_anno_SHOW(HB1,ANNS1).


impl_anno_SHOW((A:-true),ANNS) :-
    !,
    tab(4),write(A),write('.'),nl,
    member((1,Inv),ANNS),
    tab(12), write('%'), write(Inv), nl.

impl_anno_SHOW((A:-B),ANNS) :-
    tab(4),write(A), write(' :-'),nl,
    member((1,Inv),ANNS),
    tab(12), write('%'), write(Inv), nl,
    body_anno_SHOW(B,1,_,ANNS,'.',8). 


body_anno_SHOW((A,B),K0,K1,ANNS,Terminator,Indent) :- 
    !, 
    body_anno_SHOW(A,K0,KA,ANNS,',',Indent),
    body_anno_SHOW(B,KA,K1,ANNS,Terminator,Indent).  

body_anno_SHOW((C->A;B),K0,K1,ANNS,Terminator,Indent) :- 
    !, 
    tab(Indent), write('('),nl,
    Indent1 is Indent+4,
    IndentInv1 is Indent1+4,
    IndentInv is Indent+4,
    body_anno_SHOW(C,K0,KC,ANNS,' ',Indent1),
    tab(Indent), write('->'),nl,
    KA0 is KC+1,
    (
        member((KA0,InvA0),ANNS) 
    ->
        tab(IndentInv1), write('%'), write(InvA0), nl
    ;
        true
    ),
    body_anno_SHOW(A,KA0,KA,ANNS,' ',Indent1),
    tab(Indent), write(';'),nl,
    KB0 is KA+1,
    (
        member((KB0,InvB0),ANNS) 
    ->
        tab(IndentInv1), write('%'), write(InvB0), nl
    ;
        true
    ),
    body_anno_SHOW(B,KB0,KB1,ANNS,' ',Indent1),
    tab(Indent), write(')'), write(Terminator), nl,
    K1 is KB1+1,
    (
        member((K1,Inv1),ANNS) 
    ->
        tab(IndentInv), write('%'), write(Inv1), nl
    ;
        true
    ).  

body_anno_SHOW((C->A),K0,K1,ANNS,Terminator,Indent) :- 
    !, 
    tab(Indent), write('('),nl,
    Indent1 is Indent+4,
    IndentInv1 is Indent1+4,
    body_anno_SHOW(C,K0,KC,ANNS,' ',Indent1),
    tab(Indent), write('->'),nl,
    KA0 is KC+1,
    (
        member((KA0,InvA0),ANNS) 
    ->
        tab(IndentInv1), write('%'), write(InvA0), nl
    ;
        true
    ),
    body_anno_SHOW(A,KA0,K1,ANNS,' ',Indent1),
    tab(Indent), write(')'), write(Terminator), nl.

body_anno_SHOW((A;B),K0,K1,ANNS,Terminator,Indent) :- 
    !, 
    tab(Indent),
    write('('),nl,
    Indent1 is Indent+4,
    IndentInv1 is Indent1+4,
    IndentInv is Indent+4,
    body_anno_SHOW(A,K0,KA,ANNS,' ',Indent1),
    tab(Indent), write(';'),nl,
    KB0 is KA+1,
    (
        member((KB0,Inv),ANNS) 
    ->
        tab(IndentInv1), write('%'),
        write(Inv), 
        nl
    ;
        true
    ),
    body_anno_SHOW(B,KB0,KB1,ANNS,' ',Indent1),
    tab(Indent), write(')'), write(Terminator), nl,
    K1 is KB1+1,
    (
        member((K1,Inv1),ANNS) 
    ->
        tab(IndentInv), write('%'),
        write(Inv1), 
        nl
    ;
        true
    ).  

body_anno_SHOW(\+(A),K0,K1,ANNS,Terminator,Indent) :- 
    !, 
    tab(Indent),
    write('\\+ ('),nl,
    Indent1 is Indent+4,
    IndentInv is Indent+4,
    body_anno_SHOW(A,K0,KA,ANNS,' ',Indent1),
    tab(Indent), write('   )'), write(Terminator), nl,
    K1 is KA+1,
    (
        member((K1,Inv1),ANNS) 
    ->
        tab(IndentInv), write('%'), write(Inv1), nl
    ;
        true
    ).  


body_anno_SHOW(not(A),K0,K1,ANNS,Terminator,Indent) :- 
    !, 
    tab(Indent),
    write('not('),nl,
    Indent1 is Indent+4,
    IndentInv is Indent+4,
    body_anno_SHOW(A,K0,KA,ANNS,' ',Indent1),
    tab(Indent), write('   )'), write(Terminator), nl,
    K1 is KA+1,
    (
        member((K1,Inv1),ANNS) 
    ->
        tab(IndentInv), write('%'), write(Inv1), nl
    ;
        true
    ).  


body_anno_SHOW(findall(P,A,Ps),K0,K1,ANNS,Terminator,Indent) :- 
    !, 
    tab(Indent),write('findall( '), write(P), write(','), nl,
    Indent1 is Indent+4,
    IndentInv is Indent+4,
    body_anno_SHOW(A,K0,KA,ANNS,',',Indent1),
    tab(Indent),write(Ps),write(')'), write(Terminator), nl,
    K1 is KA+1,
    (
        member((K1,Inv1),ANNS) 
    ->
        tab(IndentInv), write('%'), write(Inv1), nl
    ;
        true
    ).  


body_anno_SHOW(setof(P,A,Ps),K0,K1,ANNS,Terminator,Indent) :- 
    !, 
    tab(Indent),write('setof( '), write(P), write(','), nl,
    Indent1 is Indent+4,
    IndentInv is Indent+4,
    body_anno_SHOW(A,K0,KA,ANNS,',',Indent1),
    tab(Indent),write(Ps),write(')'), write(Terminator), nl,
    K1 is KA+1,
    (
        member((K1,Inv1),ANNS) 
    ->
        tab(IndentInv), write('%'), write(Inv1), nl
    ;
        true
    ).  


body_anno_SHOW(bagof(P,A,Ps),K0,K1,ANNS,Terminator,Indent) :- 
    !, 
    tab(Indent),write('bagof( '), write(P), write(','), nl,
    Indent1 is Indent+4,
    IndentInv is Indent+4,
    body_anno_SHOW(A,K0,KA,ANNS,',',Indent1),
    tab(Indent),write(Ps),write(')'), write(Terminator), nl,
    K1 is KA+1,
    (
        member((K1,Inv1),ANNS) 
    ->
        tab(IndentInv), write('%'), write(Inv1), nl
    ;
        true
    ).  




body_anno_SHOW(A,K0,K1,ANNS, Terminator,Indent) :- 
    tab(Indent), write(A), write(Terminator), nl,
    K1 is K0+1,
    (
        member((K1,Inv),ANNS) 
    ->
        IndentInv is Indent+4,
        tab(IndentInv),write('%'),
        write(Inv),nl
    ;
        true
    ).


body_size((A,B),Size) :-
    !,
    body_size(A,Sa),
    body_size(B,Sb),
    Size is Sa+Sb.

body_size((A;B),Size) :-
    !,
    body_size(A,Sa),
    body_size(B,Sb),
    Size is Sa+Sb+2. % A point precedes B and follows outside disjunction

body_size(\+(A),Size) :-
    !,
    body_size(A,Sa),
    Size is Sa+1.    % A point follows A inside negation

body_size(not(A),Size) :-
    !,
    body_size(A,Sa),
    Size is Sa+1.    % A point follows A inside negation

body_size(findall(_,A,_),Size) :-
    !,
    body_size(A,Sa),
    Size is Sa+1.    % A point follows A inside findall

body_size(setof(_,A,_),Size) :-
    !,
    body_size(A,Sa),
    Size is Sa+1.    % A point follows A inside findall

body_size(bagof(_,A,_),Size) :-
    !,
    body_size(A,Sa),
    Size is Sa+1.    % A point follows A inside findall


body_size((A->B),Size) :-
    !,
    body_size(A,Sa),
    body_size(B,Sb),
    Size is Sa+Sb+1. % A point precedes B and another follows A

body_size(_Atom,1).


/*cls_READ(+Stream,-Clause)  reads a  clause -Clause  from an input
stream +Stream. If the end of the stream is encountered, Clause is set
to end_of_file*/

cls_READ(Stream,Clause) :-
    read_variables(Stream,TermX,NLX),
    (
        TermX = end_of_file
    -> 
        Clause = end_of_file
    ;
        TermX =(:- type(::=(LHS,RHS)))
    ->
	type_def_expansion(LHS,RHS),
        cls_READ(Stream,Clause)
    ;
	TermX = (:- Call)
    ->
	call(Call),
        cls_READ(Stream,Clause)
    ;
        (
            TermX = (Hx:-Bx) 
        -> 
            true
        ;
            Hx = TermX,
            Bx = true
        ),
            intern(((Hx,Bx),NLX),((H,B),NL),_Map),
            vars((H,B),VL),
            nameListToSub(VL,NL,Sub),
            cls_CONS((H:-B),Sub,VL,Clause)
     ).

cl_getPD(Cl,PN/PA) :-
    cls_IMPLICATION(Cl,(H:-_B)),
    functor(H,PN,PA).


/*A  prolog  program   is  represented   as  a   term   of  the   form
program(ListOfProcedures)  where  ListOfProcedures  is a list  of  the
representations of a set of procedres. A procedre is represented as  a
term  of  the  form proc(Name,Arity,ListOfClauses)  where  Name is the
name of  the  procedure,  Arity the  arity  of the  procedre  and  the
ListOfClauses is an ordered  list of the representation of the clauses
of the  procedure. A  clause  is represented  as  a  term of  the form
cl((Head:-Body),Sub)  where  (head:-Body)  is the  program  clause
in the form of internal representation of  the PROLOG system, that is,
the  variables are in  the  form of storage addresses,  where  Sub  is
maps the variable into their name */

pg_EMPTY(prog([])).

/*pg_LOAD(+File,-Program) reads -Program from +File */

pg_LOAD(File,Prog) :-
    typedef_clear,
    op(1150,fx,type),
    op(1125,xfx, ::=),
    fileerrors(Old,off), %turn off fileerrors
    ( 
        open(File,read,Stream)
    ->
        pg_LOAD_1(Stream,Prog,FUNCS),
        close(Stream)
    ;   
        nl,write(File),write(' does not exist!'),nl,
        FUNCS = [],
        pg_EMPTY(Prog)
    ),
    fileerrors(_,Old),  %restore fileerrors
    typedef_add_default(FUNCS).

pg_LOAD_1(Stream,Prog,FUNCS) :-
    pg_LOAD_1(Stream,prog([]),Prog,[],FUNCS).
pg_LOAD_1(Stream,Pg0,Pg,FUNCSAcc,FUNCS) :-
    cls_READ(Stream,Cl),
    (
        Cl = end_of_file
    ->
        Pg = Pg0, FUNCS = FUNCSAcc
    ;
        pg_AddCLS(Pg0,Cl,Pg1),
        clss_FUNC([Cl],FUNCSAcc,FUNCSAcc1),
        pg_LOAD_1(Stream,Pg1,Pg,FUNCSAcc1,FUNCS)
    ).

pg_AddCLS(OldProg,Cl,NewProg) :-
    cl_getPD(Cl,PN/PA),
    (
        pg_PROC_pg(OldProg,PN/PA,OldProc,OldProg1)
    ->
        cls_proc_ADD_proc(Cl,OldProc,NewProc),
        pg_AddPROC(NewProc,OldProg1,NewProg)
    ;
        pg_AddPROC(proc(PN,PA,[Cl]),
                           OldProg,NewProg)
    ).

/*pg_AddPROC(+Proc,+ProgI,-ProgO) ProgO is ProgI plus Proc.*/

pg_AddPROC(Proc,prog(ProcList),prog([Proc|ProcList])).

/*cls_proc_ADD_proc(+Cl,+ProcI,-ProcO) procedure ProcO is procedure
ProcI plus clause Cl.*/

cls_proc_ADD_proc(Cl,proc(PN,PA,Cls0),proc(PN,PA,Cls1)) :-
    append(Cls0,[Cl],Cls1).

/*pg_PROC_pg(+ProgI,+PN/+PA,-Proc,ProgO) succeeds if ProgI
contains clauses for procedure PN/PA.  If so,  Proc is the  definition
of  PN/PA  and  ProgO  contains definitions  for other  procedures  in
ProgI.*/

pg_PROC_pg(prog([Proc|ProcList]),PN/PA,Proc,prog(ProcList)) :-
    Proc = proc(PN,PA,_),
    !.
pg_PROC_pg(prog([Proc0|ProcList0]),PN/PA,Proc,prog([Proc0|ProcList1])) :-
    pg_PROC_pg(prog(ProcList0),PN/PA,Proc,prog(ProcList1)).

/*pg_PROC(+PgI,+(PN/PA),-PrO)*/

pg_PROC(PgI,Pn/Pa,PrO) :- 
    pg_PROC_pg(PgI,Pn/Pa,PrO,_).

pg_FUNCS(Pg,Fs) :-
    pg_PIDS(Pg,Pids),
    pids_pg_FUNCS(Pids,Pg,[],Fs).

pids_pg_FUNCS([],_,Fs,Fs).
pids_pg_FUNCS([Pid|Pids],Pg,Fs0,Fs1) :-
    pg_PROC(Pg,Pid,proc(_,_,Cs)),
    clss_FUNC(Cs,Fs0,Fs2),
    pids_pg_FUNCS(Pids,Pg,Fs2,Fs1).

clss_FUNC([],Fs,Fs).
clss_FUNC([C|Cs],Fs0,Fs1) :-
    cls_IMPLICATION(C,(H:-B)),
    atom_FUNCS(H,Fs0,Fs2),
    body_FUNCS(B,Fs2,Fs3),
    clss_FUNC(Cs,Fs3,Fs1).
    
atom_FUNCS(Atom,FUNCSAcc,FUNCS) :-
    Atom =.. [_| Terms],
    terms_FUNCS(Terms,FUNCSAcc,FUNCS).

body_FUNCS((A,B),FUNCSAcc,FUNCS) :-
    !,
    body_FUNCS(A,FUNCSAcc,NewFUNCSAcc),
    body_FUNCS(B,NewFUNCSAcc,FUNCS).
body_FUNCS((A->B),FUNCSAcc,FUNCS) :-
    !,
    body_FUNCS(A,FUNCSAcc,NewFUNCSAcc),
    body_FUNCS(B,NewFUNCSAcc,FUNCS).
body_FUNCS((A;B),FUNCSAcc,FUNCS) :-
    !,
    body_FUNCS(A,FUNCSAcc,NewFUNCSAcc),
    body_FUNCS(B,NewFUNCSAcc,FUNCS).
body_FUNCS(\+(A),FUNCSAcc,FUNCS) :-
    !,
    body_FUNCS(A,FUNCSAcc,FUNCS).
body_FUNCS(not(A),FUNCSAcc,FUNCS) :-
    !,
    body_FUNCS(A,FUNCSAcc,FUNCS).
body_FUNCS(findall(P,Call,Ps),FUNCSAcc,FUNCS) :-
    !,
    body_FUNCS(Call,FUNCSAcc,FUNCSAcc1),
    atom_FUNCS((P=Ps),FUNCSAcc1,FUNCS).
body_FUNCS(setof(P,Call,Ps),FUNCSAcc,FUNCS) :-
    !,
    body_FUNCS(Call,FUNCSAcc,FUNCSAcc1),
    atom_FUNCS((P=Ps),FUNCSAcc1,FUNCS).
body_FUNCS(bagof(P,Call,Ps),FUNCSAcc,FUNCS) :-
    !,
    body_FUNCS(Call,FUNCSAcc,FUNCSAcc1),
    atom_FUNCS((P=Ps),FUNCSAcc1,FUNCS).
body_FUNCS(A,FUNCSAcc,FUNCS):-
    atom_FUNCS(A,FUNCSAcc,FUNCS). 

/*   terms_FUNCS(Terms,+FunSetAcc,-FunSet)   FunSet   is
FunSetAcc plus functors in Terms */

terms_FUNCS([],FunSetAcc,FunSetAcc).
terms_FUNCS([Term|Terms],FunSetAcc,FunSet)	:-
    term_FUNCS(Term,FunSetAcc,NewFunSetAcc),
    terms_FUNCS(Terms,NewFunSetAcc,FunSet).

/*    term_FUNCS(Term,+FunSetAcc,-FunSet)    FunSet    =
FunSetAcc plus functors in Term.*/
   
term_FUNCS(Term,FunSetAcc,FunSet)	:-
    (
        variable(Term)
    -> 
        FunSet = FunSetAcc
    ;  
        functor(Term,Functor,Arity),
	Term =.. [Functor|Subterms],
        ord_add_element(FunSetAcc,Functor/Arity,NewFunSetAcc),
	terms_FUNCS(Subterms,NewFunSetAcc,FunSet)
    ).


type_def_expansion(Tp, RHS1|RHS) :- 
   !,
    RHS1 =.. [FN|Tps],
    type_clause_body(Tps,Args,Body),
    Tm =.. [FN|Args],
    typedef_add((type(Tm,Tp):-Body)),
    type_def_expansion(Tp,RHS).
 
type_def_expansion(Tp,RHS1) :- 
   !,
    RHS1 =.. [FN|Tps],
    type_clause_body(Tps,Args,Body),
    Tm =.. [FN|Args],
    typedef_add(type(Tm,Tp):-Body).
   
    
type_clause_body([], [], true) :-
    !.
type_clause_body([T],[X],type(X,T)) :-
    !.
type_clause_body([T,T2|Ts],[X|Xs], (type(X,T),Body)) :- 
    type_clause_body([T2|Ts],Xs,Body).




query_EMPTY(empty_query).

query_CLS(query(_Alpha,_ASub,Cls),Cls).

query_GOAL(query(_Alpha,_ASub,Cls),Body) :-
    cls_IMPLICATION(Cls,(_Head:-Body)).

query_ASUB(query(_Alpha,ASub,_Cls),ASub).

query_ALPHA(query(Alpha,_ASub,_Cls),Alpha).

query_CONS(Alpha,ASub,Cls,query(Alpha,ASub,Cls)).

/*

test :- pg_LOAD('test_progs/insert.pl',P),
        write(P),nl,
        pg_Undefined(P,FunSet),
	write(FunSet),nl.

:- trace, test.
*/












